theory CCau
imports "~~/src/HOL/Nominal/Nominal" "Common"
(*imports "Nominal/Nominal2"*)

begin

(* The Coherent Calculus of Audited Units, or \<lambda>hc *)

(* The types and codes of CCAU *)
nominal_datatype ty = PTy
                    | FunTy "ty" "ty"
                    | BoxTy "code" "ty"
and code = Var "var"
                      | UVar "uvar"
                      | Lam "ty" "\<guillemotleft>var\<guillemotright>code"
                      | App "code" "code"
                      | Box "code"
                      | LetBox "ty" "code" "\<guillemotleft>uvar\<guillemotright>code"
                      | OInsp "code_repl"
                      | IInsp "code" "code_repl"
and code_repl = CrNil "code"
              | CrCons "trail_label" "code" "code_repl"

(* trails *)
nominal_datatype trail = Refl "code"
                       | Trans "trail" "trail"
                       | Beta "ty" "\<guillemotleft>var\<guillemotright>code" "code"
                       | BetaBox "ty" "code" "\<guillemotleft>uvar\<guillemotright>code"
                       | ITI "trail" "code_repl"
                       | OTI "trail" "code_repl"
                       | QLam "ty" "\<guillemotleft>var\<guillemotright>trail"
                       | QApp "trail" "trail"
                       | QLet "ty" "trail" "\<guillemotleft>uvar\<guillemotright>trail"
                       | ITB "trail" "trail_repl"
                       | OTB "trail_repl"
and trail_repl = QrNil "trail" (* represents the default branch *)
               | QrCons "trail_label" "trail" "trail_repl"

nominal_primrec skel_of_trail :: "trail \<Rightarrow> trailskel"
  and skel_of_trail_repl :: "trail_repl \<Rightarrow> trailskel_repl"
where
  "skel_of_trail (Refl _) = SRefl"
| "skel_of_trail (Trans q1 q2) = STrans (skel_of_trail q1) (skel_of_trail q2)"
| "a \<sharp> (\<sigma>,N) \<Longrightarrow> skel_of_trail (Beta \<sigma> a M N) = SBeta"
| "u \<sharp> (\<sigma>,M) \<Longrightarrow> skel_of_trail (BetaBox \<sigma> M u N) = SBetaBox"
| "skel_of_trail (ITI _ _) = SITI"
| "skel_of_trail (OTI _ _) = SOTI"
| "a \<sharp> \<sigma> \<Longrightarrow> skel_of_trail (QLam \<sigma> a q) = SLam (skel_of_trail q)"
| "skel_of_trail (QApp q1 q2) = SApp (skel_of_trail q1) (skel_of_trail q2)"
| "u \<sharp> (\<sigma>,q1) \<Longrightarrow> skel_of_trail (QLet \<sigma> q1 u q2) = SLet (skel_of_trail q1) (skel_of_trail q2)"
| "skel_of_trail (ITB q qs) = SITB (skel_of_trail q) (skel_of_trail_repl qs)"
| "skel_of_trail (OTB qs) = SOTB (skel_of_trail_repl qs)"
| "skel_of_trail_repl (QrNil q) = SrNil (skel_of_trail q)"
| "skel_of_trail_repl (QrCons l q qs) = SrCons l (skel_of_trail q) (skel_of_trail_repl qs)"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto simp:fresh_trailskel)
  apply(fresh_guess add:fresh_trail_label)+
done
  

lemma skel_of_trail_eqvt[eqvt] :
  fixes \<pi> :: "var prm"
  fixes \<pi>':: "uvar prm"
  shows "\<pi> \<bullet> (skel_of_trail q) = skel_of_trail (\<pi> \<bullet> q)"
  and   "\<pi>' \<bullet> (skel_of_trail q) = skel_of_trail (\<pi>' \<bullet> q)"
  and "\<pi> \<bullet> (skel_of_trail_repl qs) = skel_of_trail_repl (\<pi> \<bullet> qs)"
  and   "\<pi>' \<bullet> (skel_of_trail_repl qs) = skel_of_trail_repl (\<pi>' \<bullet> qs)"
apply (nominal_induct q and qs rule:trail_trail_repl.strong_inducts)
apply (auto simp add:perm_bij fresh_bij)
done

type_synonym  ctx = "(var * ty) list"

lemma fresh_var_not_in:
  fixes x::var
  and \<Gamma>::ctx
  shows "(x,\<tau>) \<in> set \<Gamma> \<Longrightarrow> x \<sharp> \<Gamma> \<Longrightarrow>  False"
  apply(induct \<Gamma>)
  by(auto simp add:fresh_list_cons fresh_prod fresh_atm)

(* FALSE
lemma uvar_fresh_ctx[simp]:
  fixes u::uvar 
  fixes \<Gamma> :: ctx
  shows "u \<sharp> \<Gamma>"
  apply(induct \<Gamma>)
  apply(auto simp add:fresh_list_nil fresh_list_cons fresh_prod fresh_atm fresh_ty)
*)

type_synonym  uctx = "(uvar * ty) list"

lemma fresh_uvar_not_in:
  fixes u::uvar
  and \<Delta>::uctx
  shows "(u,\<tau>) \<in> set \<Delta> \<Longrightarrow> u \<sharp> \<Delta> \<Longrightarrow>  False"
  apply(induct \<Delta>)
  by(auto simp add:fresh_list_cons fresh_prod fresh_atm)  

(* FALSE
lemma var_fresh_uctx[simp]: 
  fixes x::var
  fixes \<Delta>::uctx
  shows "x \<sharp> \<Delta>"
  apply(induct \<Delta>)
  apply(auto simp add:fresh_list_nil fresh_list_cons fresh_prod fresh_atm fresh_ty)
  sorry
*)

(* Terms *)
nominal_datatype trm = TVar "var"
                     | TUVar "uvar"
                     | TLam "ty" "\<guillemotleft>var\<guillemotright>trm"
                     | TApp "trm" "trm"
                     | TBox "trail" "trm"
                     | TLetBox "ty" "trm" "\<guillemotleft>uvar\<guillemotright>trm"
                     | TOInsp "trm_repl"
                     | TIInsp "trm" "trm_repl"
and trm_repl = TrNil "trm"
             | TrCons "trail_label" "trm" "trm_repl"

fun mkcr :: "(trail_label \<times> code) list \<Rightarrow> code \<Rightarrow> code_repl" where
 "mkcr [] bdef = CrNil bdef"
|"mkcr ((x,M)#l) bdef = CrCons x M (mkcr l bdef)"

fun mkqr :: "(trail_label \<times> trail) list \<Rightarrow> trail \<Rightarrow> trail_repl" where
 "mkqr [] bdef = QrNil bdef"
|"mkqr ((x,q)#l) bdef = QrCons x q (mkqr l bdef)"

fun mktr :: "(trail_label \<times> trm) list \<Rightarrow> trm \<Rightarrow> trm_repl" where
 "mktr [] bdef = TrNil bdef"
|"mktr ((x,M)#l) bdef = TrCons x M (mktr l bdef)"

nominal_primrec
  bcdom :: "code_repl \<Rightarrow> trail_label list"
  and bcdaux :: "code \<Rightarrow> unit"
  and bcdaux2 :: "ty \<Rightarrow> unit"
where
  "bcdom (CrNil Mdef) = []"
| "bcdom (CrCons x M tr) = x#bcdom tr"
| "bcdaux (Var _) = ()"
| "bcdaux (UVar _) = ()"
| "a \<sharp> \<tau> \<Longrightarrow> bcdaux (Lam \<tau> a M) = ()"
| "u \<sharp> (\<tau>,M) \<Longrightarrow> bcdaux (LetBox \<tau> M u N) = ()"
| "bcdaux (App _ _) = ()"
| "bcdaux (Box _) = ()"
| "bcdaux (OInsp _) = ()"
| "bcdaux (IInsp _ _) = ()"
| "bcdaux2 PTy = ()"
| "bcdaux2 (FunTy _ _) = ()"
| "bcdaux2 (BoxTy _ _) = ()"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto)
  apply(fresh_guess)+
done

nominal_primrec
  bqdefault :: "trail_repl \<Rightarrow> trail"
  and bqdefaux :: "trail \<Rightarrow> unit"
where
  "bqdefault (QrNil q) = q"
| "bqdefault (QrCons _ _ qs) = bqdefault qs"
| "bqdefaux (Refl _) = ()"
| "bqdefaux (Trans _ _) = ()"
| "a \<sharp> (\<sigma>,N) \<Longrightarrow> bqdefaux (Beta \<sigma> a M N) = ()"
| "u \<sharp> (\<sigma>,M) \<Longrightarrow> bqdefaux (BetaBox \<sigma> M u N) = ()"
| "bqdefaux (ITI _ _) = ()"
| "bqdefaux (OTI _ _) = ()"
| "a \<sharp> \<sigma> \<Longrightarrow> bqdefaux (QLam \<sigma> a _) = ()"
| "bqdefaux (QApp _ _) = ()"
| "u \<sharp> (\<sigma>,q1) \<Longrightarrow> bqdefaux (QLet \<sigma> q1 u q2) = ()"
| "bqdefaux (ITB _ _) = ()"
| "bqdefaux (OTB _) = ()"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto)
  apply(fresh_guess add:fresh_trail_label)+
done

nominal_primrec
  bqlookup :: "trail_repl \<Rightarrow> trail_label \<Rightarrow> trail option"
  and bqlaux :: "trail \<Rightarrow> trail_label \<Rightarrow> unit"
where
  "bqlookup (QrNil _) y = None"
| "bqlookup (QrCons x q tr) y = (if x=y then Some q else bqlookup tr y)"
| "bqlaux (Refl _) y = ()"
| "bqlaux (Trans _ _) y = ()"
| "a \<sharp> (\<sigma>,N) \<Longrightarrow> bqlaux (Beta \<sigma> a M N) y = ()"
| "u \<sharp> (\<sigma>,M) \<Longrightarrow> bqlaux (BetaBox \<sigma> M u N) y = ()"
| "bqlaux (ITI _ _) y = ()"
| "bqlaux (OTI _ _) y = ()"
| "a \<sharp> \<sigma> \<Longrightarrow> bqlaux (QLam \<sigma> a _) y = ()"
| "bqlaux (QApp _ _) y = ()"
| "u \<sharp> (\<sigma>,q1) \<Longrightarrow> bqlaux (QLet \<sigma> q1 u q2) y = ()"
| "bqlaux (ITB _ _) y = ()"
| "bqlaux (OTB _) y = ()"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto)
  apply(fresh_guess add:fresh_trail_label)+
done

nominal_primrec
  bqdom :: "trail_repl \<Rightarrow> trail_label list"
  and bqdaux :: "trail \<Rightarrow> unit"
where
  "bqdom (QrNil qdef) = []"
| "bqdom (QrCons x q tr) = x#bqdom tr"
| "bqdaux (Refl _) = ()"
| "bqdaux (Trans _ _) = ()"
| "a \<sharp> (\<sigma>,N) \<Longrightarrow> bqdaux (Beta \<sigma> a M N) = ()"
| "u \<sharp> (\<sigma>,M) \<Longrightarrow> bqdaux (BetaBox \<sigma> M u N) = ()"
| "bqdaux (ITI _ _) = ()"
| "bqdaux (OTI _ _) = ()"
| "a \<sharp> \<sigma> \<Longrightarrow> bqdaux (QLam \<sigma> a _) = ()"
| "bqdaux (QApp _ _) = ()"
| "u \<sharp> (\<sigma>,q1) \<Longrightarrow> bqdaux (QLet \<sigma> q1 u q2) = ()"
| "bqdaux (ITB _ _) = ()"
| "bqdaux (OTB _) = ()"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto)
  apply(fresh_guess)+
done

nominal_primrec
  btdom :: "trm_repl \<Rightarrow> trail_label list"
  and btdaux :: "trm \<Rightarrow> unit"
where
  "btdom (TrNil Mdef) = []"
| "btdom (TrCons x M tr) = x#btdom tr"
| "btdaux (TVar _) = ()"
| "btdaux (TUVar _) = ()"
| "a \<sharp> \<tau> \<Longrightarrow> btdaux (TLam \<tau> a M) = ()"
| "u \<sharp> (\<tau>,M) \<Longrightarrow> btdaux (TLetBox \<tau> M u N) = ()"
| "btdaux (TApp _ _) = ()"
| "btdaux (TBox _ _) = ()"
| "btdaux (TOInsp _) = ()"
| "btdaux (TIInsp _ _) = ()"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto)
  apply(fresh_guess)+
done

nominal_primrec
  bcdefault :: "code_repl \<Rightarrow> code"
  and bcdefaux :: "code \<Rightarrow> unit"
  and bcdefaux2 :: "ty \<Rightarrow> unit"
where
  "bcdefault (CrNil M) = M"
| "bcdefault (CrCons _ _ Ms) = bcdefault Ms"
| "bcdefaux (Var _) = ()"
| "bcdefaux (UVar _) = ()"
| "a \<sharp> \<tau> \<Longrightarrow> bcdefaux (Lam \<tau> a M)= ()"
| "u \<sharp> (\<tau>,M) \<Longrightarrow> bcdefaux (LetBox \<tau> M u N) = ()"
| "bcdefaux (App _ _) = ()"
| "bcdefaux (Box _) = ()"
| "bcdefaux (OInsp _) = ()"
| "bcdefaux (IInsp _ _) = ()"
| "bcdefaux2 PTy = ()"
| "bcdefaux2 (FunTy _ _) = ()"
| "bcdefaux2 (BoxTy _ _) = ()"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto)
  apply(fresh_guess add:fresh_trail_label)+
done

nominal_primrec
  bclookup :: "code_repl \<Rightarrow> trail_label \<Rightarrow> code option"
  and bclaux :: "code \<Rightarrow> trail_label \<Rightarrow> unit"
  and bclaux2 :: "ty \<Rightarrow> trail_label \<Rightarrow> unit"
where
  "bclookup (CrNil _) y = None"
| "bclookup (CrCons x M tr) y = (if x=y then Some M else bclookup tr y)"
| "bclaux (Var _) y = ()"
| "bclaux (UVar _) y = ()"
| "a \<sharp> \<tau> \<Longrightarrow> bclaux (Lam \<tau> a M) y = ()"
| "u \<sharp> (\<tau>,M) \<Longrightarrow> bclaux (LetBox \<tau> M u N) y = ()"
| "bclaux (App _ _) y = ()"
| "bclaux (Box _) y = ()"
| "bclaux (OInsp _) y = ()"
| "bclaux (IInsp _ _) y = ()"
| "bclaux2 PTy y = ()"
| "bclaux2 (FunTy _ _) y = ()"
| "bclaux2 (BoxTy _ _) y = ()"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto)
  apply(fresh_guess add:fresh_trail_label)+
done

lemma bcdefault_eqvt[eqvt] :
  fixes \<pi>::"var prm"
  fixes \<pi>'::"uvar prm"
  shows "\<pi> \<bullet> bcdefault \<theta> = bcdefault (\<pi> \<bullet> \<theta>)"
  and "\<pi>' \<bullet> bcdefault \<theta> = bcdefault (\<pi>' \<bullet> \<theta>)"
  apply (nominal_induct \<theta> rule:ty_code_code_repl.strong_inducts(3))
  apply (simp) 
  apply (auto)
done

lemma bcdefault_fresh :
  fixes a::var
  fixes u::uvar
  shows "a \<sharp> \<theta> \<Longrightarrow> a \<sharp> bcdefault \<theta>"
  and "u \<sharp> \<theta> \<Longrightarrow> u \<sharp> bcdefault \<theta>"
by (nominal_induct \<theta> rule:ty_code_code_repl.strong_inducts(3),auto)

lemma bclookup_eqvt[eqvt] :
  fixes \<pi>::"var prm"
  fixes \<pi>'::"uvar prm"
  shows "\<pi> \<bullet> bclookup \<theta> l = bclookup (\<pi> \<bullet> \<theta>) (\<pi> \<bullet> l)"
  and "\<pi>' \<bullet> bclookup \<theta> l = bclookup (\<pi>' \<bullet> \<theta>) (\<pi>' \<bullet> l)"
  apply (nominal_induct \<theta> rule:ty_code_code_repl.strong_inducts(3)[of _ "\<lambda>_ _. True"])
  apply (simp) 
  apply (auto)
  apply (subgoal_tac "trail_label = l")
  apply (simp)
  apply (simp only:perm_bij)
  apply (subgoal_tac "trail_label = l")
  apply (simp)
  apply (simp only:perm_bij)
done

lemma bclookup_fresh[simp] :
  fixes a::var
  fixes u::uvar
  shows "a \<sharp> \<theta> \<Longrightarrow> a \<sharp> bclookup \<theta> l"
  and "u \<sharp> \<theta> \<Longrightarrow> u \<sharp> bclookup \<theta> l"
by (nominal_induct \<theta> rule:ty_code_code_repl.strong_inducts(3)[of _ "\<lambda>_ _. True"],auto simp:fresh_some fresh_none)

primrec codebind :: "code option \<Rightarrow> (code \<Rightarrow> code) \<Rightarrow> code \<Rightarrow> code" where
  "codebind None f b = b"
| "codebind (Some a) f b = f a"

lemma codebind_eqvt[eqvt] :
  fixes \<pi> :: "var prm"
  fixes \<pi>':: "uvar prm"
  shows "\<pi> \<bullet> (codebind M f b) = codebind (\<pi> \<bullet> M) (\<pi> \<bullet> f) (\<pi> \<bullet> b)"
  and   "\<pi>' \<bullet> (codebind M f b) = codebind (\<pi>' \<bullet> M) (\<pi>' \<bullet> f) (\<pi>' \<bullet> b)"
  by (cases M,auto simp add:perm_fun_def perm_pi_simp)+

lemma codebind_fresh_None :
  fixes x::var
  fixes u::uvar
  shows "x \<sharp> b \<Longrightarrow> x \<sharp> codebind None f b"
  and "u \<sharp> b \<Longrightarrow> u \<sharp> codebind None f b"
  by auto

lemma codebind_fresh_Some :
  fixes x::var
  fixes u::uvar
  shows "x \<sharp> f a \<Longrightarrow> x \<sharp> codebind (Some a) f b"
  and "u \<sharp> f a \<Longrightarrow> u \<sharp> codebind (Some a) f b"
  by auto

lemma codebind_fresh2 :
  fixes x::var
  fixes u::uvar
  shows "(M = None \<Longrightarrow> x \<sharp> b) \<Longrightarrow> (\<And>a. M = Some a \<Longrightarrow> x \<sharp> f a) \<Longrightarrow> x \<sharp> codebind M f b"
  and "(M = None \<Longrightarrow> u \<sharp> b) \<Longrightarrow> (\<And>a. M = Some a \<Longrightarrow> u \<sharp> f a) \<Longrightarrow> u \<sharp> codebind M f b"
  by (cases M, auto simp: codebind_fresh_None codebind_fresh_Some)+

lemma codebind_fresh[simp] :
  fixes x::var
  fixes u::uvar
  shows "x \<sharp> codebind M f b \<Longrightarrow> (M = None \<and> x \<sharp> b) \<or> (\<exists>a. M = Some a \<and> x \<sharp> f a)"
  and "u \<sharp> codebind M f b \<Longrightarrow> (M = None \<and> u \<sharp> b) \<or> (\<exists>a. M = Some a \<and> u \<sharp> f a)"
  by (cases M,simp_all)+

definition qcfbranch :: "code_repl \<Rightarrow> trail_label \<Rightarrow> (code \<Rightarrow> code) \<Rightarrow> code" where
  "qcfbranch \<theta> l f = codebind (bclookup \<theta> l) f (bcdefault \<theta>)"

lemma qcfbranch_eqvt[eqvt] :
  fixes \<pi> :: "var prm"
  fixes \<pi>':: "uvar prm"
  shows "\<pi> \<bullet> (qcfbranch \<theta> l f) = qcfbranch (\<pi> \<bullet> \<theta>) (\<pi> \<bullet> l) (\<pi> \<bullet> f)"
  and   "\<pi>' \<bullet> (qcfbranch \<theta> l f) = qcfbranch (\<pi>' \<bullet> \<theta>) (\<pi>' \<bullet> l) (\<pi>' \<bullet> f)"
by (auto simp add:qcfbranch_def eqvts)

lemma qcfbranch_fresh:
  fixes x::var
  fixes u::uvar
  shows "x \<sharp> bcdefault \<theta> \<Longrightarrow> (\<And>l M. bclookup \<theta> l = Some M \<Longrightarrow> x \<sharp> f M) \<Longrightarrow> x \<sharp> qcfbranch \<theta> l f" 
  and "u \<sharp> bcdefault \<theta> \<Longrightarrow> (\<And>l M. bclookup \<theta> l = Some M \<Longrightarrow> u \<sharp> f M) \<Longrightarrow> u \<sharp> qcfbranch \<theta> l f"
apply (simp_all add:qcfbranch_def)
apply (cases "bclookup \<theta> l",simp_all)+
done

lemma qcfbranch_fresh2 :
  fixes x::var
  fixes u::uvar
  shows "x \<sharp> qcfbranch \<theta> l f \<Longrightarrow> (bclookup \<theta> l = None \<and> x \<sharp> bcdefault \<theta>) \<or> (\<exists>M. bclookup \<theta> l = Some M \<and> x \<sharp> f M)"
  and "u \<sharp> qcfbranch \<theta> l f \<Longrightarrow> (bclookup \<theta> l = None \<and> u \<sharp> bcdefault \<theta>) \<or> (\<exists>M. bclookup \<theta> l = Some M \<and> u \<sharp> f M)"
  by (simp_all add:qcfbranch_def)

nominal_primrec
  scfold :: "trailskel \<Rightarrow> code_repl \<Rightarrow> code"
  and scfaux :: "trailskel_repl \<Rightarrow> code_repl \<Rightarrow> code \<Rightarrow> trail_label \<Rightarrow> code"
where
  "scfold SRefl \<theta> = qcfbranch \<theta> LRefl (\<lambda>M. M)"
| "scfold (STrans q1 q2) \<theta> = qcfbranch \<theta> LTrans (\<lambda>M. App (App M (scfold q1 \<theta>)) (scfold q2 \<theta>))"
| "scfold SBeta \<theta> = qcfbranch \<theta> LBeta (\<lambda>M. M)"
| "scfold SBetaBox \<theta> = qcfbranch \<theta> LBetaBox (\<lambda>M. M)"
| "scfold SITI \<theta> = qcfbranch \<theta> LITI (\<lambda>M. M)"
| "scfold SOTI \<theta> = qcfbranch \<theta> LOTI (\<lambda>M. M)"
| "scfold (SLam q) \<theta> = qcfbranch \<theta> LQLam (\<lambda>M. App M (scfold q \<theta>))"
| "scfold (SApp q1 q2) \<theta> = qcfbranch \<theta> LQApp (\<lambda>M. App (App M (scfold q1 \<theta>)) (scfold q2 \<theta>))"
| "scfold (SLet q1 q2) \<theta> = qcfbranch \<theta> LQLet (\<lambda>M. App (App M (scfold q1 \<theta>)) (scfold q2 \<theta>))"
| "scfold (SITB q qs) \<theta> = qcfbranch \<theta> LITBc (\<lambda>M. qcfbranch \<theta> LITBn (\<lambda>N. scfaux qs \<theta> (App (App M N) (scfold q \<theta>)) LITBc))"
| "scfold (SOTB qs) \<theta> = qcfbranch \<theta> LOTBn (\<lambda>M. scfaux qs \<theta> M LOTBc)"
| "scfaux (SrNil qdef) \<theta> = (\<lambda>Mnil Lcons. qcfbranch \<theta> Lcons (\<lambda>M. App (App M (scfold qdef \<theta>)) Mnil))"
| "scfaux (SrCons x q qs) \<theta> = (\<lambda>Mnil Lcons. qcfbranch \<theta> Lcons (\<lambda>M. App (App M (scfold q \<theta>)) (scfaux qs \<theta> Mnil Lcons)))"
by auto

lemma scfold_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  shows "pi\<bullet>(scfold q \<theta>) = scfold (pi\<bullet>q) (pi\<bullet>\<theta>)"
  and "pi'\<bullet>(scfold q \<theta>) = scfold (pi'\<bullet>q) (pi'\<bullet>\<theta>)"
  and "\<And>M l. pi\<bullet>(scfaux qs \<theta> M l) = scfaux (pi\<bullet>qs) (pi\<bullet>\<theta>) (pi\<bullet>M) (pi\<bullet>l)"
  and "\<And>M l. pi'\<bullet>(scfaux qs \<theta> M l) = scfaux (pi'\<bullet>qs) (pi'\<bullet>\<theta>) (pi'\<bullet>M) (pi'\<bullet>l)"
apply(nominal_induct q and qs rule:trailskel_trailskel_repl.strong_inducts,auto)
apply(simp add:qcfbranch_eqvt, rule arg_cong[of _ _ "qcfbranch (_ \<bullet> \<theta>) _"],
      simp add: perm_fun_def perm_pi_simp)+
apply(auto simp add:fun_eq_iff, simp add:qcfbranch_eqvt, rule arg_cong[of _ _ "qcfbranch (_ \<bullet> \<theta>) _"],
      simp add:perm_fun_def perm_pi_simp)
apply(simp add:qcfbranch_eqvt, rule arg_cong[of _ _ "qcfbranch (_ \<bullet> \<theta>) _"],
      simp add: perm_fun_def perm_pi_simp qcfbranch_eqvt)+
done  

lemma scfold_fresh :
  fixes a :: var
  fixes u :: uvar
  shows "a \<sharp> \<theta> \<Longrightarrow> a \<sharp> scfold s \<theta>"
  and "u \<sharp> \<theta> \<Longrightarrow> u \<sharp> scfold s \<theta>"
  and "a \<sharp> M \<Longrightarrow> a \<sharp> \<theta> \<Longrightarrow> a \<sharp> scfaux ss \<theta> M l"
  and "u \<sharp> M \<Longrightarrow> u \<sharp> \<theta> \<Longrightarrow> u \<sharp> scfaux ss \<theta> M l"
apply (nominal_induct s and ss avoiding:\<theta> M l rule:trailskel_trailskel_repl.strong_inducts, simp_all)
apply (rule qcfbranch_fresh, rule bcdefault_fresh, simp)
apply (subgoal_tac "a \<sharp> bclookup \<theta> l", simp add:fresh_some)
apply (rule bclookup_fresh,simp)
apply (rule qcfbranch_fresh, rule bcdefault_fresh, simp)
apply (subgoal_tac "u \<sharp> bclookup \<theta> l", simp add:fresh_some)
apply (rule bclookup_fresh,simp)
apply (rule qcfbranch_fresh, rule bcdefault_fresh, simp)
apply (subgoal_tac "a \<sharp> bclookup \<theta> l", simp add:fresh_some)
apply (rule bclookup_fresh,simp)
apply (rule qcfbranch_fresh, rule bcdefault_fresh, simp_all)
apply (subgoal_tac "u \<sharp> bclookup \<theta> l", simp add:fresh_some)
apply (rule bclookup_fresh,simp)
apply (rule qcfbranch_fresh, rule bcdefault_fresh, simp,
       subgoal_tac "a \<sharp> bclookup \<theta> l", simp add:fresh_some,
       rule bclookup_fresh,simp,
       rule qcfbranch_fresh, rule bcdefault_fresh, simp,
       subgoal_tac "u \<sharp> bclookup \<theta> l", simp add:fresh_some,
       rule bclookup_fresh,simp)+
apply (rule qcfbranch_fresh, rule bcdefault_fresh, simp_all)+
apply (subgoal_tac "a \<sharp> App (App M Ma) (scfold trailskel \<theta>)", auto)
apply (subgoal_tac "a \<sharp> bclookup \<theta> l", simp add:fresh_some, rule bclookup_fresh, simp)
apply (subgoal_tac "a \<sharp> bclookup \<theta> la", simp add:fresh_some, rule bclookup_fresh, simp)
apply (rule qcfbranch_fresh, rule bcdefault_fresh, simp_all)+
apply (subgoal_tac "u \<sharp> App (App M Ma) (scfold trailskel \<theta>)", auto)
apply (subgoal_tac "u \<sharp> bclookup \<theta> l", simp add:fresh_some, rule bclookup_fresh, simp)
apply (subgoal_tac "u \<sharp> bclookup \<theta> la", simp add:fresh_some, rule bclookup_fresh, simp)
apply (rule qcfbranch_fresh, rule bcdefault_fresh, simp)
apply (subgoal_tac "a \<sharp> bclookup \<theta> l", simp add:fresh_some)
apply (rule bclookup_fresh,simp)
apply (rule qcfbranch_fresh, rule bcdefault_fresh, simp_all)
apply (subgoal_tac "u \<sharp> bclookup \<theta> l", simp add:fresh_some)
apply (rule bclookup_fresh,simp)
apply (rule qcfbranch_fresh, rule bcdefault_fresh, simp)
apply (subgoal_tac "a \<sharp> bclookup \<theta> la", simp add:fresh_some)
apply (rule bclookup_fresh,simp)
apply (rule qcfbranch_fresh, rule bcdefault_fresh, simp_all)
apply (subgoal_tac "u \<sharp> bclookup \<theta> la", simp add:fresh_some)
apply (rule bclookup_fresh,simp)
apply (rule qcfbranch_fresh, rule bcdefault_fresh, simp)
apply (subgoal_tac "a \<sharp> bclookup \<theta> la", simp add:fresh_some)
apply (rule bclookup_fresh,simp)
apply (rule qcfbranch_fresh, rule bcdefault_fresh, simp_all)
apply (subgoal_tac "u \<sharp> bclookup \<theta> la", simp add:fresh_some)
apply (rule bclookup_fresh,simp)
done

(* Structural recursion on trails (Def.1) used to produce a code *)
definition qcfold :: "trail \<Rightarrow> code_repl \<Rightarrow> code"
where "qcfold q \<theta> = scfold (skel_of_trail q) \<theta>"

definition qcfaux :: "trail_repl \<Rightarrow> code_repl \<Rightarrow> code \<Rightarrow> trail_label \<Rightarrow> code"
where "qcfaux qs \<theta> = scfaux (skel_of_trail_repl qs) \<theta>"

lemma qcfold_simps[simp]:
  "qcfold (Refl M) \<theta> = qcfbranch \<theta> LRefl (\<lambda>M. M)"
  "qcfold (Trans q1 q2) \<theta> = qcfbranch \<theta> LTrans (\<lambda>M. App (App M (qcfold q1 \<theta>)) (qcfold q2 \<theta>))"
  "x \<sharp> (\<tau>,M2) \<Longrightarrow> qcfold (Beta \<tau> x M1 M2) \<theta> = qcfbranch \<theta> LBeta (\<lambda>M. M)"
  "u \<sharp> (\<tau>,M1) \<Longrightarrow> qcfold (BetaBox \<tau> M1 u M2) \<theta> = qcfbranch \<theta> LBetaBox (\<lambda>M. M)"
  "qcfold (ITI q \<theta>') \<theta> = qcfbranch \<theta> LITI (\<lambda>M. M)"
  "qcfold (OTI q \<theta>') \<theta> = qcfbranch \<theta> LOTI (\<lambda>M. M)"
  "x \<sharp> \<tau> \<Longrightarrow> qcfold (QLam \<tau> x q) \<theta> = qcfbranch \<theta> LQLam (\<lambda>M. App M (qcfold q \<theta>))"
  "qcfold (QApp q1 q2) \<theta> = qcfbranch \<theta> LQApp (\<lambda>M. App (App M (qcfold q1 \<theta>)) (qcfold q2 \<theta>))"
  "u \<sharp> (\<tau>,q1) \<Longrightarrow> qcfold (QLet \<tau> q1 u q2) \<theta> = qcfbranch \<theta> LQLet (\<lambda>M. App (App M (qcfold q1 \<theta>)) (qcfold q2 \<theta>))"
  "qcfold (ITB q qs) \<theta> = qcfbranch \<theta> LITBc (\<lambda>M. qcfbranch \<theta> LITBn (\<lambda>N. qcfaux qs \<theta> (App (App M N) (qcfold q \<theta>)) LITBc))"
  "qcfold (OTB qs) \<theta> = qcfbranch \<theta> LOTBn (\<lambda>M. qcfaux qs \<theta> M LOTBc)"
  "qcfaux (QrNil qdef) \<theta> = (\<lambda>Mnil Lcons. qcfbranch \<theta> Lcons (\<lambda>M. App (App M (qcfold qdef \<theta>)) Mnil))"
  "qcfaux (QrCons l q qs) \<theta> = (\<lambda>Mnil Lcons. qcfbranch \<theta> Lcons (\<lambda>M. App (App M (qcfold q \<theta>)) (qcfaux qs \<theta> Mnil Lcons)))"
by (simp_all add:qcfold_def qcfaux_def)

lemma qcfold_eqvt[eqvt] :
  fixes \<pi> :: "var prm"
  fixes \<pi>' :: "uvar prm"
  shows "\<pi> \<bullet> (qcfold q \<theta>) = qcfold (\<pi> \<bullet> q) (\<pi> \<bullet> \<theta>)"
  and "\<pi> \<bullet> (qcfaux qs \<theta> Mnil Lcons) = qcfaux (\<pi> \<bullet> qs) (\<pi> \<bullet> \<theta>) (\<pi>  \<bullet> Mnil) (\<pi> \<bullet> Lcons)"
  and "\<pi>' \<bullet> (qcfold q \<theta>) = qcfold (\<pi>' \<bullet> q) (\<pi>' \<bullet> \<theta>)"
  and "\<pi>' \<bullet> (qcfaux qs \<theta> Mnil Lcons) = qcfaux (\<pi>' \<bullet> qs) (\<pi>' \<bullet> \<theta>) (\<pi>'  \<bullet> Mnil) (\<pi>' \<bullet> Lcons)"
  by (simp_all only: qcfold_def qcfaux_def eqvts)

lemma qcfold_fresh :
  fixes a :: var
  fixes u :: uvar
  shows "a \<sharp> \<theta> \<Longrightarrow> a \<sharp> qcfold q \<theta>"
  and "u \<sharp> \<theta> \<Longrightarrow> u \<sharp> qcfold q \<theta>"
  and "a \<sharp> M \<Longrightarrow> a \<sharp> \<theta> \<Longrightarrow> a \<sharp> qcfaux qs \<theta> M l"
  and "u \<sharp> M \<Longrightarrow> u \<sharp> \<theta> \<Longrightarrow> u \<sharp> qcfaux qs \<theta> M l"
by (simp_all add:qcfold_def qcfaux_def scfold_fresh)

nominal_primrec
  btdefault :: "trm_repl \<Rightarrow> trm"
  and btdefaux :: "trm \<Rightarrow> unit"
where
  "btdefault (TrNil M) = M"
| "btdefault (TrCons _ _ Ms) = btdefault Ms"
| "btdefaux (TVar _) = ()"
| "btdefaux (TUVar _) = ()"
| "a \<sharp> \<tau> \<Longrightarrow> btdefaux (TLam \<tau> a M)= ()"
| "u \<sharp> (\<tau>,M) \<Longrightarrow> btdefaux (TLetBox \<tau> M u N) = ()"
| "btdefaux (TApp _ _) = ()"
| "btdefaux (TBox _ _) = ()"
| "btdefaux (TOInsp _) = ()"
| "btdefaux (TIInsp _ _) = ()"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto)
  apply(fresh_guess add:fresh_trail_label)+
done

nominal_primrec
  btlookup :: "trm_repl \<Rightarrow> trail_label \<Rightarrow> trm option"
  and btlaux :: "trm \<Rightarrow> trail_label \<Rightarrow> unit"
where
  "btlookup (TrNil _) y = None"
| "btlookup (TrCons x M tr) y = (if x=y then Some M else btlookup tr y)"
| "btlaux (TVar _) y = ()"
| "btlaux (TUVar _) y = ()"
| "a \<sharp> \<tau> \<Longrightarrow> btlaux (TLam \<tau> a M) y = ()"
| "u \<sharp> (\<tau>,M) \<Longrightarrow> btlaux (TLetBox \<tau> M u N) y = ()"
| "btlaux (TApp _ _) y = ()"
| "btlaux (TBox _ _) y = ()"
| "btlaux (TOInsp _) y = ()"
| "btlaux (TIInsp _ _) y = ()"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto)
  apply(fresh_guess add:fresh_trail_label)+
done

lemma btdefault_eqvt[eqvt] :
  fixes \<pi>::"var prm"
  fixes \<pi>'::"uvar prm"
  shows "\<pi> \<bullet> btdefault \<theta> = btdefault (\<pi> \<bullet> \<theta>)"
  and "\<pi>' \<bullet> btdefault \<theta> = btdefault (\<pi>' \<bullet> \<theta>)"
  apply (nominal_induct \<theta> rule:trm_trm_repl.strong_inducts(2))
  apply (simp) 
  apply (auto)
done

lemma btdefault_fresh :
  fixes a::var
  fixes u::uvar
  shows "a \<sharp> \<theta> \<Longrightarrow> a \<sharp> btdefault \<theta>"
  and "u \<sharp> \<theta> \<Longrightarrow> u \<sharp> btdefault \<theta>"
by (nominal_induct \<theta> rule:trm_trm_repl.strong_inducts(2),auto)

lemma btlookup_eqvt[eqvt] :
  fixes \<pi>::"var prm"
  fixes \<pi>'::"uvar prm"
  shows "\<pi> \<bullet> btlookup \<theta> l = btlookup (\<pi> \<bullet> \<theta>) (\<pi> \<bullet> l)"
  and "\<pi>' \<bullet> btlookup \<theta> l = btlookup (\<pi>' \<bullet> \<theta>) (\<pi>' \<bullet> l)"
  apply (nominal_induct \<theta> rule:trm_trm_repl.strong_inducts(2))
  apply (simp) 
  apply (auto)
  apply (subgoal_tac "trail_label = l")
  apply (simp)
  apply (simp only:perm_bij)
  apply (subgoal_tac "trail_label = l")
  apply (simp)
  apply (simp only:perm_bij)
done

lemma btlookup_fresh[simp] :
  fixes a::var
  fixes u::uvar
  shows "a \<sharp> \<theta> \<Longrightarrow> a \<sharp> btlookup \<theta> l"
  and "u \<sharp> \<theta> \<Longrightarrow> u \<sharp> btlookup \<theta> l"
by (nominal_induct \<theta> rule:trm_trm_repl.strong_inducts(2)[of "\<lambda>_ _. True"],auto simp:fresh_some fresh_none)

primrec trmbind :: "trm option \<Rightarrow> (trm \<Rightarrow> trm) \<Rightarrow> trm \<Rightarrow> trm" where
  "trmbind None f b = b"
| "trmbind (Some a) f b = f a"

lemma trmbind_eqvt[eqvt] :
  fixes \<pi> :: "var prm"
  fixes \<pi>':: "uvar prm"
  shows "\<pi> \<bullet> (trmbind M f b) = trmbind (\<pi> \<bullet> M) (\<pi> \<bullet> f) (\<pi> \<bullet> b)"
  and   "\<pi>' \<bullet> (trmbind M f b) = trmbind (\<pi>' \<bullet> M) (\<pi>' \<bullet> f) (\<pi>' \<bullet> b)"
  by (cases M,auto simp add:perm_fun_def perm_pi_simp)+

lemma trmbind_fresh_None :
  fixes x::var
  fixes u::uvar
  shows "x \<sharp> b \<Longrightarrow> x \<sharp> trmbind None f b"
  and "u \<sharp> b \<Longrightarrow> u \<sharp> trmbind None f b"
  by auto

lemma trmbind_fresh_Some :
  fixes x::var
  fixes u::uvar
  shows "x \<sharp> f a \<Longrightarrow> x \<sharp> trmbind (Some a) f b"
  and "u \<sharp> f a \<Longrightarrow> u \<sharp> trmbind (Some a) f b"
  by auto

lemma trmbind_fresh2 :
  fixes x::var
  fixes u::uvar
  shows "(M = None \<Longrightarrow> x \<sharp> b) \<Longrightarrow> (\<And>a. M = Some a \<Longrightarrow> x \<sharp> f a) \<Longrightarrow> x \<sharp> trmbind M f b"
  and "(M = None \<Longrightarrow> u \<sharp> b) \<Longrightarrow> (\<And>a. M = Some a \<Longrightarrow> u \<sharp> f a) \<Longrightarrow> u \<sharp> trmbind M f b"
  by (cases M, auto simp: trmbind_fresh_None trmbind_fresh_Some)+

lemma trmbind_fresh[simp] :
  fixes x::var
  fixes u::uvar
  shows "x \<sharp> trmbind M f b \<Longrightarrow> (M = None \<and> x \<sharp> b) \<or> (\<exists>a. M = Some a \<and> x \<sharp> f a)"
  and "u \<sharp> trmbind M f b \<Longrightarrow> (M = None \<and> u \<sharp> b) \<or> (\<exists>a. M = Some a \<and> u \<sharp> f a)"
  by (cases M,simp_all)+

definition qtfbranch :: "trm_repl \<Rightarrow> trail_label \<Rightarrow> (trm \<Rightarrow> trm) \<Rightarrow> trm" where
  "qtfbranch \<theta> l f = trmbind (btlookup \<theta> l) f (btdefault \<theta>)"

lemma qtfbranch_eqvt[eqvt] :
  fixes \<pi> :: "var prm"
  fixes \<pi>':: "uvar prm"
  shows "\<pi> \<bullet> (qtfbranch \<theta> l f) = qtfbranch (\<pi> \<bullet> \<theta>) (\<pi> \<bullet> l) (\<pi> \<bullet> f)"
  and   "\<pi>' \<bullet> (qtfbranch \<theta> l f) = qtfbranch (\<pi>' \<bullet> \<theta>) (\<pi>' \<bullet> l) (\<pi>' \<bullet> f)"
by (auto simp add:qtfbranch_def eqvts)

lemma qtfbranch_fresh:
  fixes x::var
  fixes u::uvar
  shows "x \<sharp> btdefault \<theta> \<Longrightarrow> (\<And>l M. btlookup \<theta> l = Some M \<Longrightarrow> x \<sharp> f M) \<Longrightarrow> x \<sharp> qtfbranch \<theta> l f" 
  and "u \<sharp> btdefault \<theta> \<Longrightarrow> (\<And>l M. btlookup \<theta> l = Some M \<Longrightarrow> u \<sharp> f M) \<Longrightarrow> u \<sharp> qtfbranch \<theta> l f"
apply (simp_all add:qtfbranch_def)
apply (cases "btlookup \<theta> l",simp_all)+
done

lemma qtfbranch_fresh2 :
  fixes x::var
  fixes u::uvar
  shows "x \<sharp> qtfbranch \<theta> l f \<Longrightarrow> (btlookup \<theta> l = None \<and> x \<sharp> btdefault \<theta>) \<or> (\<exists>M. btlookup \<theta> l = Some M \<and> x \<sharp> f M)"
  and "u \<sharp> qtfbranch \<theta> l f \<Longrightarrow> (btlookup \<theta> l = None \<and> u \<sharp> btdefault \<theta>) \<or> (\<exists>M. btlookup \<theta> l = Some M \<and> u \<sharp> f M)"
  by (simp_all add:qtfbranch_def)

nominal_primrec
  stfold :: "trailskel \<Rightarrow> trm_repl \<Rightarrow> trm"
  and stfaux :: "trailskel_repl \<Rightarrow> trm_repl \<Rightarrow> trm \<Rightarrow> trail_label \<Rightarrow> trm"
where
  "stfold SRefl \<theta> = qtfbranch \<theta> LRefl (\<lambda>M. M)"
| "stfold (STrans q1 q2) \<theta> = qtfbranch \<theta> LTrans (\<lambda>M. TApp (TApp M (stfold q1 \<theta>)) (stfold q2 \<theta>))"
| "stfold SBeta \<theta> = qtfbranch \<theta> LBeta (\<lambda>M. M)"
| "stfold SBetaBox \<theta> = qtfbranch \<theta> LBetaBox (\<lambda>M. M)"
| "stfold SITI \<theta> = qtfbranch \<theta> LITI (\<lambda>M. M)"
| "stfold SOTI \<theta> = qtfbranch \<theta> LOTI (\<lambda>M. M)"
| "stfold (SLam q) \<theta> = qtfbranch \<theta> LQLam (\<lambda>M. TApp M (stfold q \<theta>))"
| "stfold (SApp q1 q2) \<theta> = qtfbranch \<theta> LQApp (\<lambda>M. TApp (TApp M (stfold q1 \<theta>)) (stfold q2 \<theta>))"
| "stfold (SLet q1 q2) \<theta> = qtfbranch \<theta> LQLet (\<lambda>M. TApp (TApp M (stfold q1 \<theta>)) (stfold q2 \<theta>))"
| "stfold (SITB q qs) \<theta> = qtfbranch \<theta> LITBc (\<lambda>M. qtfbranch \<theta> LITBn (\<lambda>N. stfaux qs \<theta> (TApp (TApp M N) (stfold q \<theta>)) LITBc))"
| "stfold (SOTB qs) \<theta> = qtfbranch \<theta> LOTBn (\<lambda>M. stfaux qs \<theta> M LOTBc)"
| "stfaux (SrNil qdef) \<theta> = (\<lambda>Mnil Lcons. qtfbranch \<theta> Lcons (\<lambda>M. TApp (TApp M (stfold qdef \<theta>)) Mnil))"
| "stfaux (SrCons x q qs) \<theta> = (\<lambda>Mnil Lcons. qtfbranch \<theta> Lcons (\<lambda>M. TApp (TApp M (stfold q \<theta>)) (stfaux qs \<theta> Mnil Lcons)))"
by auto

lemma stfold_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  shows "pi\<bullet>(stfold q \<theta>) = stfold (pi\<bullet>q) (pi\<bullet>\<theta>)"
  and "pi'\<bullet>(stfold q \<theta>) = stfold (pi'\<bullet>q) (pi'\<bullet>\<theta>)"
  and "\<And>M l. pi\<bullet>(stfaux qs \<theta> M l) = stfaux (pi\<bullet>qs) (pi\<bullet>\<theta>) (pi\<bullet>M) (pi\<bullet>l)"
  and "\<And>M l. pi'\<bullet>(stfaux qs \<theta> M l) = stfaux (pi'\<bullet>qs) (pi'\<bullet>\<theta>) (pi'\<bullet>M) (pi'\<bullet>l)"
apply(nominal_induct q and qs rule:trailskel_trailskel_repl.strong_inducts,auto)
apply(simp add:qtfbranch_eqvt, rule arg_cong[of _ _ "qtfbranch (_ \<bullet> \<theta>) _"],
      simp add: perm_fun_def perm_pi_simp)+
apply(auto simp add:fun_eq_iff, simp add:qtfbranch_eqvt, rule arg_cong[of _ _ "qtfbranch (_ \<bullet> \<theta>) _"],
      simp add:perm_fun_def perm_pi_simp)
apply(simp add:qtfbranch_eqvt, rule arg_cong[of _ _ "qtfbranch (_ \<bullet> \<theta>) _"],
      simp add: perm_fun_def perm_pi_simp qtfbranch_eqvt)+
done  

lemma stfold_fresh :
  fixes a :: var
  fixes u :: uvar
  shows "a \<sharp> \<theta> \<Longrightarrow> a \<sharp> stfold s \<theta>"
  and "u \<sharp> \<theta> \<Longrightarrow> u \<sharp> stfold s \<theta>"
  and "a \<sharp> M \<Longrightarrow> a \<sharp> \<theta> \<Longrightarrow> a \<sharp> stfaux ss \<theta> M l"
  and "u \<sharp> M \<Longrightarrow> u \<sharp> \<theta> \<Longrightarrow> u \<sharp> stfaux ss \<theta> M l"
apply (nominal_induct s and ss avoiding:\<theta> M l rule:trailskel_trailskel_repl.strong_inducts, simp_all)
apply (rule qtfbranch_fresh, rule btdefault_fresh, simp)
apply (subgoal_tac "a \<sharp> btlookup \<theta> l", simp add:fresh_some)
apply (rule btlookup_fresh,simp)
apply (rule qtfbranch_fresh, rule btdefault_fresh, simp)
apply (subgoal_tac "u \<sharp> btlookup \<theta> l", simp add:fresh_some)
apply (rule btlookup_fresh,simp)
apply (rule qtfbranch_fresh, rule btdefault_fresh, simp)
apply (subgoal_tac "a \<sharp> btlookup \<theta> l", simp add:fresh_some)
apply (rule btlookup_fresh,simp)
apply (rule qtfbranch_fresh, rule btdefault_fresh, simp_all)
apply (subgoal_tac "u \<sharp> btlookup \<theta> l", simp add:fresh_some)
apply (rule btlookup_fresh,simp)
apply (rule qtfbranch_fresh, rule btdefault_fresh, simp,
       subgoal_tac "a \<sharp> btlookup \<theta> l", simp add:fresh_some,
       rule btlookup_fresh,simp,
       rule qtfbranch_fresh, rule btdefault_fresh, simp,
       subgoal_tac "u \<sharp> btlookup \<theta> l", simp add:fresh_some,
       rule btlookup_fresh,simp)+
apply (rule qtfbranch_fresh, rule btdefault_fresh, simp_all)+
apply (subgoal_tac "a \<sharp> TApp (TApp M Ma) (stfold trailskel \<theta>)", auto)
apply (subgoal_tac "a \<sharp> btlookup \<theta> l", simp add:fresh_some, rule btlookup_fresh, simp)
apply (subgoal_tac "a \<sharp> btlookup \<theta> la", simp add:fresh_some, rule btlookup_fresh, simp)
apply (rule qtfbranch_fresh, rule btdefault_fresh, simp_all)+
apply (subgoal_tac "u \<sharp> TApp (TApp M Ma) (stfold trailskel \<theta>)", auto)
apply (subgoal_tac "u \<sharp> btlookup \<theta> l", simp add:fresh_some, rule btlookup_fresh, simp)
apply (subgoal_tac "u \<sharp> btlookup \<theta> la", simp add:fresh_some, rule btlookup_fresh, simp)
apply (rule qtfbranch_fresh, rule btdefault_fresh, simp)
apply (subgoal_tac "a \<sharp> btlookup \<theta> l", simp add:fresh_some)
apply (rule btlookup_fresh,simp)
apply (rule qtfbranch_fresh, rule btdefault_fresh, simp_all)
apply (subgoal_tac "u \<sharp> btlookup \<theta> l", simp add:fresh_some)
apply (rule btlookup_fresh,simp)
apply (rule qtfbranch_fresh, rule btdefault_fresh, simp)
apply (subgoal_tac "a \<sharp> btlookup \<theta> la", simp add:fresh_some)
apply (rule btlookup_fresh,simp)
apply (rule qtfbranch_fresh, rule btdefault_fresh, simp_all)
apply (subgoal_tac "u \<sharp> btlookup \<theta> la", simp add:fresh_some)
apply (rule btlookup_fresh,simp)
apply (rule qtfbranch_fresh, rule btdefault_fresh, simp)
apply (subgoal_tac "a \<sharp> btlookup \<theta> la", simp add:fresh_some)
apply (rule btlookup_fresh,simp)
apply (rule qtfbranch_fresh, rule btdefault_fresh, simp_all)
apply (subgoal_tac "u \<sharp> btlookup \<theta> la", simp add:fresh_some)
apply (rule btlookup_fresh,simp)
done

(* Structural recursion on trails (Def.1) used to produce a term *)
definition qtfold :: "trail \<Rightarrow> trm_repl \<Rightarrow> trm"
where "qtfold q \<theta> = stfold (skel_of_trail q) \<theta>"

definition qtfaux :: "trail_repl \<Rightarrow> trm_repl \<Rightarrow> trm \<Rightarrow> trail_label \<Rightarrow> trm"
where "qtfaux qs \<theta> = stfaux (skel_of_trail_repl qs) \<theta>"

lemma qtfold_simps[simp]:
  "qtfold (Refl M) \<theta> = qtfbranch \<theta> LRefl (\<lambda>M. M)"
  "qtfold (Trans q1 q2) \<theta> = qtfbranch \<theta> LTrans (\<lambda>M. TApp (TApp M (qtfold q1 \<theta>)) (qtfold q2 \<theta>))"
  "x \<sharp> (\<tau>,M2) \<Longrightarrow> qtfold (Beta \<tau> x M1 M2) \<theta> = qtfbranch \<theta> LBeta (\<lambda>M. M)"
  "u \<sharp> (\<tau>,M1) \<Longrightarrow> qtfold (BetaBox \<tau> M1 u M2) \<theta> = qtfbranch \<theta> LBetaBox (\<lambda>M. M)"
  "qtfold (ITI q \<theta>') \<theta> = qtfbranch \<theta> LITI (\<lambda>M. M)"
  "qtfold (OTI q \<theta>') \<theta> = qtfbranch \<theta> LOTI (\<lambda>M. M)"
  "x \<sharp> \<tau> \<Longrightarrow> qtfold (QLam \<tau> x q) \<theta> = qtfbranch \<theta> LQLam (\<lambda>M. TApp M (qtfold q \<theta>))"
  "qtfold (QApp q1 q2) \<theta> = qtfbranch \<theta> LQApp (\<lambda>M. TApp (TApp M (qtfold q1 \<theta>)) (qtfold q2 \<theta>))"
  "u \<sharp> (\<tau>,q1) \<Longrightarrow> qtfold (QLet \<tau> q1 u q2) \<theta> = qtfbranch \<theta> LQLet (\<lambda>M. TApp (TApp M (qtfold q1 \<theta>)) (qtfold q2 \<theta>))"
  "qtfold (ITB q qs) \<theta> = qtfbranch \<theta> LITBc (\<lambda>M. qtfbranch \<theta> LITBn (\<lambda>N. qtfaux qs \<theta> (TApp (TApp M N) (qtfold q \<theta>)) LITBc))"
  "qtfold (OTB qs) \<theta> = qtfbranch \<theta> LOTBn (\<lambda>M. qtfaux qs \<theta> M LOTBc)"
  "qtfaux (QrNil qdef) \<theta> = (\<lambda>Mnil Lcons. qtfbranch \<theta> Lcons (\<lambda>M. TApp (TApp M (qtfold qdef \<theta>)) Mnil))"
  "qtfaux (QrCons l q qs) \<theta> = (\<lambda>Mnil Lcons. qtfbranch \<theta> Lcons (\<lambda>M. TApp (TApp M (qtfold q \<theta>)) (qtfaux qs \<theta> Mnil Lcons)))"
by (simp_all add:qtfold_def qtfaux_def)

lemma qtfold_eqvt[eqvt] :
  fixes \<pi> :: "var prm"
  shows "\<pi> \<bullet> (qtfold q \<theta>) = qtfold (\<pi> \<bullet> q) (\<pi> \<bullet> \<theta>)"
  and "\<pi> \<bullet> (qtfaux qs \<theta> Mnil Lcons) = qtfaux (\<pi> \<bullet> qs) (\<pi> \<bullet> \<theta>) (\<pi>  \<bullet> Mnil) (\<pi> \<bullet> Lcons)"
  by (simp_all only: qtfold_def qtfaux_def eqvts)

lemma qtfold_fresh :
  fixes a :: var
  fixes u :: uvar
  shows "a \<sharp> \<theta> \<Longrightarrow> a \<sharp> qtfold q \<theta>"
  and "u \<sharp> \<theta> \<Longrightarrow> u \<sharp> qtfold q \<theta>"
  and "a \<sharp> M \<Longrightarrow> a \<sharp> \<theta> \<Longrightarrow> a \<sharp> qtfaux qs \<theta> M l"
  and "u \<sharp> M \<Longrightarrow> u \<sharp> \<theta> \<Longrightarrow> u \<sharp> qtfaux qs \<theta> M l"
by (simp_all add:qtfold_def qtfaux_def stfold_fresh)

(* substitutions *)

type_synonym subst = "(var * code) list"

fun
  lookup :: "subst \<Rightarrow> var \<Rightarrow> code"   
where
  "lookup [] x        = Var x"
| "lookup ((y,M)#\<sigma>) x = (if x=y then M else lookup \<sigma> x)"

lemma lookup_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  shows "pi\<bullet>(lookup \<theta> x) = lookup (pi\<bullet>\<theta>) (pi\<bullet>x)"
  and "pi'\<bullet>(lookup \<theta> x) = lookup (pi'\<bullet>\<theta>) (pi'\<bullet>x)"
by (induct \<theta>) (auto simp add: perm_bij)

lemma lookup_fresh:
  fixes x ::"var" 
  fixes u ::"uvar"
  shows "x \<sharp> \<theta> \<Longrightarrow> x \<sharp> y \<Longrightarrow> x \<sharp> lookup \<theta> y"
  and "u \<sharp> \<theta> \<Longrightarrow>  u \<sharp> lookup \<theta> y"
apply (induct \<theta>)
by (auto simp add: fresh_atm fresh_prod fresh_list_cons)

lemma lookup_default:
  "x \<sharp> \<gamma> \<Longrightarrow>  lookup \<gamma> x = Var x"
  apply (induct \<gamma>)
  by (auto simp add: fresh_atm fresh_prod fresh_list_cons)

type_synonym usubst = "(uvar * code) list"

fun
  ulookup :: "usubst \<Rightarrow> uvar \<Rightarrow> code"   
where
  "ulookup [] x        = UVar x"
| "ulookup ((y,M)#\<sigma>) x = (if x=y then M else ulookup \<sigma> x)"

lemma ulookup_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  shows "pi\<bullet>(ulookup \<theta> x) = ulookup (pi\<bullet>\<theta>) (pi\<bullet>x)"
  and "pi'\<bullet>(ulookup \<theta> x) = ulookup (pi'\<bullet>\<theta>) (pi'\<bullet>x)"
by (induct \<theta>) (auto simp add: perm_bij)

lemma ulookup_fresh:
  fixes x ::"var" 
  fixes u::"uvar"
  shows "x \<sharp> \<theta> \<Longrightarrow> x \<sharp> ulookup \<theta> v"
  and "u \<sharp> \<theta> \<Longrightarrow> u \<sharp> v \<Longrightarrow>  u \<sharp> ulookup \<theta> v"
apply (induct \<theta>)
by (auto simp add: fresh_atm fresh_prod fresh_list_cons)

lemma ulookup_default:
  "u \<sharp> \<theta> \<Longrightarrow>  ulookup \<theta> u = UVar u"
  apply (induct \<theta>)
  by (auto simp add: fresh_atm fresh_prod fresh_list_cons)

(* Substitution on codes (Def.2) *)
nominal_primrec subst :: "code \<Rightarrow> usubst \<Rightarrow> subst \<Rightarrow> code" ("_<_,_>" [100,60] 100)
 and subst_repl :: "code_repl \<Rightarrow> usubst \<Rightarrow> subst \<Rightarrow> code_repl" ("_<_,_>" [100,60] 100)
 and subst_ty :: "ty \<Rightarrow> usubst \<Rightarrow> subst \<Rightarrow> ty" ("_<_,_>" [100,60] 100)
where 
  "(Var y)<S,T> = (lookup T y)"
| "(UVar u)<S,T> = (ulookup S u)"
| "y \<sharp> (\<tau>,S,T) \<Longrightarrow> (Lam \<tau> y M)<S,T> = Lam (\<tau><S,T>) y (M<S,T>)"
| "(App M1 M2)<S,T> = App (M1<S,T>) (M2<S,T>)"
| "(Box M)<S,T> = Box (M<S,T>)"
| "v \<sharp> (\<tau>,M1,S,T) \<Longrightarrow> (LetBox \<tau> M1 v M2)<S,T> = LetBox (\<tau><S,T>) (M1<S,T>) v (M2<S,T>)"
| "(OInsp er)<S,T> = OInsp (er<S,T>)"
| "(IInsp M er)<S,T> = IInsp (M<S,T>) (er<S,T>)"
| "(CrNil M)<S,T> = CrNil (M<S,T>)"
| "(CrCons l M er)<S,T> = CrCons l (M<S,T>) (er<S,T>)"
| "PTy<S,T> = PTy"
| "(FunTy \<sigma> \<tau>)<S,T> = FunTy (\<sigma><S,T>) (\<tau><S,T>)"
| "(BoxTy M \<sigma>)<S,T> = BoxTy (M<S,T>) (\<sigma><S,T>)"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(simp_all add: abs_fresh)+
  apply(fresh_guess)+
done

lemma subst_eqvt[eqvt]:
  fixes \<pi>::"var prm"
  fixes \<pi>'::"uvar prm"
  fixes \<sigma> :: ty
  fixes M::code
  fixes Ms :: code_repl
  fixes x::var
  shows "\<pi> \<bullet> (\<sigma><S,T>) = (\<pi>\<bullet>\<sigma>)<\<pi>\<bullet>S,\<pi>\<bullet>T>" "\<pi>' \<bullet> (\<sigma><S,T>) = (\<pi>'\<bullet>\<sigma>)<\<pi>'\<bullet>S,\<pi>'\<bullet>T>"
  and "\<pi> \<bullet> (M<S,T>) = (\<pi>\<bullet>M)<\<pi>\<bullet>S,\<pi>\<bullet>T>" "\<pi>' \<bullet> (M<S,T>) = (\<pi>'\<bullet>M)<\<pi>'\<bullet>S,\<pi>'\<bullet>T>"
  and "\<pi> \<bullet> (Ms<S,T>) = (\<pi>\<bullet>Ms)<\<pi>\<bullet>S,\<pi>\<bullet>T>" "\<pi>' \<bullet> (Ms<S,T>) = (\<pi>'\<bullet>Ms)<\<pi>'\<bullet>S,\<pi>'\<bullet>T>"
  apply(nominal_induct \<sigma> and M and Ms avoiding: S T rule:ty_code_code_repl.strong_inducts)
  apply(auto simp add:perm_bij fresh_bij lookup_eqvt ulookup_eqvt)
done

lemma subst2_fresh: 
  fixes \<sigma> ::ty
  fixes M ::code
  fixes Ms ::code_repl
  fixes x::var
  fixes u::uvar
  shows "x \<sharp> \<sigma> \<Longrightarrow> x \<sharp> S \<Longrightarrow> x \<sharp> T \<Longrightarrow> x \<sharp> \<sigma><S,T>" "u \<sharp> \<sigma> \<Longrightarrow> u \<sharp> S \<Longrightarrow> u \<sharp> T \<Longrightarrow> u \<sharp> \<sigma><S,T> "
  and "x \<sharp> M \<Longrightarrow> x \<sharp> S \<Longrightarrow> x \<sharp> T \<Longrightarrow> x \<sharp> M<S,T> " "u \<sharp> M \<Longrightarrow> u \<sharp> S \<Longrightarrow> u \<sharp> T \<Longrightarrow> u \<sharp> M<S,T> "
  and "x \<sharp> Ms \<Longrightarrow> x \<sharp> S \<Longrightarrow> x \<sharp> T \<Longrightarrow> x \<sharp> Ms<S,T> " "u \<sharp> Ms \<Longrightarrow> u \<sharp> S \<Longrightarrow> u \<sharp> T \<Longrightarrow> u \<sharp> Ms<S,T> "    
  apply(nominal_induct \<sigma> and M and Ms avoiding: S T rule:ty_code_code_repl.strong_inducts)
  apply(auto simp add: abs_fresh fresh_ty fresh_list_cons fresh_list_nil lookup_fresh ulookup_fresh)
done

lemma var_id:
  fixes \<sigma> :: ty
  fixes M :: code
  fixes Ms :: code_repl
  shows "x \<sharp> \<gamma> \<Longrightarrow> \<sigma><\<delta>,(x,Var x)#\<gamma>> = \<sigma><\<delta>,\<gamma>>"
  and "x \<sharp> \<gamma> \<Longrightarrow> M<\<delta>,(x,Var x)#\<gamma>> = M<\<delta>,\<gamma>>"
  and "x \<sharp> \<gamma> \<Longrightarrow> Ms<\<delta>,(x,Var x)#\<gamma>> = Ms<\<delta>,\<gamma>>"
  apply(nominal_induct \<sigma> and M and Ms avoiding: \<gamma> \<delta> x rule:ty_code_code_repl.strong_inducts)
  apply(auto simp add: fresh_list_cons fresh_prod lookup_default)
done

lemma uvar_id:
  fixes \<sigma> :: ty
  fixes M :: code
  fixes Ms :: code_repl
  shows"u \<sharp> \<delta> \<Longrightarrow>\<sigma><(u,UVar u)#\<delta>,\<gamma>> = \<sigma><\<delta>,\<gamma>>"
  and "u \<sharp> \<delta> \<Longrightarrow>M<(u,UVar u)#\<delta>,\<gamma>> = M<\<delta>,\<gamma>>"
  and "u \<sharp> \<delta> \<Longrightarrow>Ms<(u,UVar u)#\<delta>,\<gamma>> = Ms<\<delta>,\<gamma>>"
  apply(nominal_induct \<sigma> and M and Ms avoiding: \<gamma> \<delta> u rule:ty_code_code_repl.strong_inducts)
  apply(auto simp add: fresh_list_cons fresh_prod ulookup_default)
done

abbreviation
  subst1 :: "code \<Rightarrow> var \<Rightarrow> code \<Rightarrow> code" ("_[_\<mapsto>_]" [60,60,60] 100) 
where
  "M[x\<mapsto>N] \<equiv>  M<[],[(x,N)]>"

abbreviation
  subst1_repl :: "code_repl \<Rightarrow> var \<Rightarrow> code \<Rightarrow> code_repl" ("_[_\<mapsto>_]" [60,60,60] 100) 
where
  "subst1_repl Ms x N \<equiv>  Ms<[],[(x,N)]>"

abbreviation
  subst1_ty :: "ty \<Rightarrow> var \<Rightarrow> code \<Rightarrow> ty" ("_[_\<mapsto>_]" [60,60,60] 100) 
where
  "subst1_ty \<sigma> x N \<equiv>  \<sigma><[],[(x,N)]>"

lemma subst_lam_simp[simp]:
  "x\<sharp> (y,ty,N) \<Longrightarrow> (Lam ty x M)[y\<mapsto>N] = Lam (ty[y\<mapsto>N]) x (M[y\<mapsto>N])"  
  by(simp add: fresh_list_cons fresh_list_nil)
  
lemma subst_letbox_simp[simp]:
  "u \<sharp> (\<tau>,M,Q) \<Longrightarrow> (LetBox \<tau> M u N)[y\<mapsto>Q] = LetBox (\<tau>[y\<mapsto>Q]) (M[y\<mapsto> Q]) u (N[y\<mapsto>Q])"  
  by(simp add: fresh_list_cons fresh_list_nil fresh_atm)

lemma subst_fresh_t: 
  fixes \<sigma>::ty
  fixes M::code
  fixes Ms::code_repl
  fixes x::var
  shows "x \<sharp> [y].\<sigma> \<Longrightarrow> x \<sharp> N \<Longrightarrow> x \<sharp> (\<sigma>[y\<mapsto>N])"
  and "x \<sharp> [y].M \<Longrightarrow> x \<sharp> N \<Longrightarrow> x \<sharp> (M[y\<mapsto>N])"
  and "x \<sharp> [y].Ms \<Longrightarrow> x \<sharp> N \<Longrightarrow> x \<sharp> (Ms[y\<mapsto>N])"
  apply(nominal_induct \<sigma> and M and Ms avoiding: y N rule:ty_code_code_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_trail_label)
done

lemma var_subst_fresh: 
  fixes \<sigma>::ty
  fixes M::code
  fixes Ms::code_repl
  fixes x::var
  shows "x \<sharp> N \<Longrightarrow> x \<sharp> (\<sigma>[x\<mapsto>N])"  
  and "x \<sharp> N \<Longrightarrow> x \<sharp> (M[x\<mapsto>N])"  
  and "x \<sharp> N \<Longrightarrow> x \<sharp> (Ms[x\<mapsto>N])"  
  apply(nominal_induct \<sigma> and M and Ms avoiding: x N rule:ty_code_code_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_trail_label)
done

lemma var_subst_id:
  fixes \<sigma>::ty
  fixes M::code
  fixes Ms::code_repl
  shows "x \<sharp> \<sigma> \<Longrightarrow> \<sigma>[x\<mapsto>N] = \<sigma>"
  and "x \<sharp> M \<Longrightarrow> M[x\<mapsto>N] = M"
  and "x \<sharp> Ms \<Longrightarrow> Ms[x\<mapsto>N] = Ms"
  apply(nominal_induct \<sigma> and M and Ms avoiding: x N rule:ty_code_code_repl.strong_inducts)
  apply(auto simp add: fresh_atm abs_fresh)
done

lemma subst1_merge:
  fixes \<sigma>::ty
  fixes M::code
  fixes Ms::code_repl
  shows "x \<sharp> \<delta> \<Longrightarrow> x \<sharp> \<gamma> \<Longrightarrow> \<sigma><\<delta>,(x, N) # \<gamma>> = \<sigma><\<delta>, \<gamma>>[x\<mapsto>N]"
  and  "x \<sharp> \<delta> \<Longrightarrow> x \<sharp> \<gamma> \<Longrightarrow> M<\<delta>,(x, N) # \<gamma>> = M<\<delta>, \<gamma>>[x\<mapsto>N]"
  and "x \<sharp> \<delta> \<Longrightarrow> x \<sharp> \<gamma> \<Longrightarrow> Ms<\<delta>,(x, N) # \<gamma>> = Ms<\<delta>, \<gamma>>[x\<mapsto>N]"
  apply(nominal_induct \<sigma> and M and Ms avoiding: \<gamma> x \<delta> N rule:ty_code_code_repl.strong_inducts)
  apply(auto simp add:lookup_default fresh_atm var_subst_id lookup_fresh)
  apply(auto simp add:ulookup_default var_subst_id ulookup_fresh)
  apply(auto simp add: fresh_list_cons subst2_fresh fresh_prod fresh_atm fresh_list_nil)
  done

abbreviation
  usubst1 :: "code \<Rightarrow> uvar \<Rightarrow> code \<Rightarrow> code" ("_{_\<mapsto>_}" [100,60,60] 100) 
where
  "M{u\<mapsto>N} \<equiv>  M<[(u,N)],[]>"

abbreviation
  usubst1_repl :: "code_repl \<Rightarrow> uvar \<Rightarrow> code \<Rightarrow> code_repl" ("_{_\<mapsto>_}" [100,60,60] 100) 
where
  "usubst1_repl Ms u N \<equiv>  Ms<[(u,N)],[]>"

abbreviation
  usubst1_ty :: "ty \<Rightarrow> uvar \<Rightarrow> code \<Rightarrow> ty" ("_{_\<mapsto>_}" [100,60,60] 100) 
where
  "usubst1_ty \<sigma> u N \<equiv>  \<sigma><[(u,N)],[]>"

lemma usubst_lam_simp[simp]:
  "x\<sharp>(ty,N) \<Longrightarrow> (Lam ty x M){u\<mapsto>N} = Lam (ty{u\<mapsto>N}) x (M{u\<mapsto>N})"  
  by(simp add: fresh_list_cons fresh_list_nil fresh_atm)
  
lemma usubst_letbox_simp[simp]:
  "v \<sharp> (\<tau>,M,u,Q) \<Longrightarrow> (LetBox \<tau> M v N){u\<mapsto>Q} = LetBox (\<tau>{u\<mapsto>Q}) (M{u\<mapsto>Q}) v (N{u\<mapsto>Q})"  
  by(simp add: fresh_list_cons fresh_list_nil fresh_atm fresh_prod)
  
lemma uvar_usubst_fresh: 
  fixes \<sigma>::ty
  fixes M::code
  fixes Ms::code_repl
  fixes u::uvar
  shows "u \<sharp> N \<Longrightarrow> u \<sharp> (\<sigma>{u\<mapsto>N})"  
  and "u \<sharp> N \<Longrightarrow> u \<sharp> (M{u\<mapsto>N})"
  and "u \<sharp> N \<Longrightarrow> u \<sharp> (Ms{u\<mapsto>N})"  
  apply(nominal_induct \<sigma> and M and Ms avoiding: u N rule:ty_code_code_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_trail_label)
done

lemma usubst_fresh_u: 
  fixes \<sigma>::ty
  fixes M::code
  fixes Ms::code_repl
  fixes u::uvar
  shows "u \<sharp> [v].\<sigma> \<Longrightarrow> u \<sharp> N \<Longrightarrow> u \<sharp> (\<sigma>{v\<mapsto>N})"  
  and "u \<sharp> [v].M \<Longrightarrow> u \<sharp> N \<Longrightarrow> u \<sharp> (M{v\<mapsto>N})"  
  and "u \<sharp> [v].Ms \<Longrightarrow> u \<sharp> N \<Longrightarrow> u \<sharp> (Ms{v\<mapsto>N})"  
  apply(nominal_induct \<sigma> and M and Ms avoiding: v N rule:ty_code_code_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_ty fresh_trail_label)
done

lemma subst_fresh_u: 
  fixes \<sigma>::ty
  fixes M::code
  fixes Ms::code_repl
  fixes u::uvar
  fixes a::var
  shows "u \<sharp> [a].\<sigma> \<Longrightarrow> u \<sharp> N \<Longrightarrow> u \<sharp> (subst1_ty \<sigma> a N)" 
  and "u \<sharp> [a].M \<Longrightarrow> u \<sharp> N \<Longrightarrow> u \<sharp> (subst1 M a N)" 
  and "u \<sharp> [a].Ms \<Longrightarrow> u \<sharp> N \<Longrightarrow> u \<sharp> (subst1_repl Ms a N)" 
  apply(nominal_induct \<sigma> and M and Ms avoiding: a N rule:ty_code_code_repl.strong_inducts)
  apply(auto simp add:code.fresh abs_fresh fresh_ty)
done

lemmas subst_fresh = subst_fresh_t subst_fresh_u

lemma usubst_fresh_t: 
  fixes \<sigma>::ty
  fixes M::code
  fixes Ms::code_repl
  fixes a::var
  shows "a \<sharp> [v].\<sigma> \<Longrightarrow> a \<sharp> N \<Longrightarrow> a \<sharp> (\<sigma>{v\<mapsto>N})"  
  and "a \<sharp> [v].M \<Longrightarrow> a \<sharp> N \<Longrightarrow> a \<sharp> (M{v\<mapsto>N})"  
  and "a \<sharp> [v].Ms \<Longrightarrow> a \<sharp> N \<Longrightarrow> a \<sharp> (Ms{v\<mapsto>N})"  
  apply(nominal_induct \<sigma> and M and Ms avoiding: v N rule:ty_code_code_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_ty )
done

lemmas usubst_fresh = usubst_fresh_t usubst_fresh_u

lemma uvar_subst_id:
  fixes \<sigma>::ty
  fixes M::code
  fixes Ms::code_repl
  shows "u \<sharp> \<sigma> \<Longrightarrow> \<sigma>{u\<mapsto>N} = \<sigma>"
  and "u \<sharp> M \<Longrightarrow> M{u\<mapsto>N} = M"
  and "u \<sharp> Ms \<Longrightarrow> Ms{u\<mapsto>N} = Ms"
  apply(nominal_induct \<sigma> and M and Ms avoiding: u N rule:ty_code_code_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh) 
done

lemma usubst1_merge:
  fixes \<sigma>::ty
  fixes M::code
  fixes Ms::code_repl
  shows "u \<sharp> \<delta> \<Longrightarrow> u \<sharp> \<gamma> \<Longrightarrow> \<sigma><(u,N)#\<delta>, \<gamma>> = \<sigma><\<delta>, \<gamma>>{u\<mapsto>N}"
  and "u \<sharp> \<delta> \<Longrightarrow> u \<sharp> \<gamma> \<Longrightarrow> M<(u,N)#\<delta>, \<gamma>> = M<\<delta>, \<gamma>>{u\<mapsto>N}"
  and "u \<sharp> \<delta> \<Longrightarrow> u \<sharp> \<gamma> \<Longrightarrow> Ms<(u,N)#\<delta>, \<gamma>> = Ms<\<delta>, \<gamma>>{u\<mapsto>N}"
  apply(nominal_induct \<sigma> and M and Ms avoiding: \<gamma> u \<delta> N rule:ty_code_code_repl.strong_inducts)
  apply(auto simp add:lookup_default fresh_atm uvar_subst_id lookup_fresh)
  apply(auto simp add:ulookup_default uvar_subst_id ulookup_fresh fresh_atm)
  apply(auto simp add: fresh_list_cons subst2_fresh fresh_prod fresh_atm fresh_list_nil)
  done

(* Source and target of a trail (Def.3) *)
nominal_primrec src_tgt :: "trail \<Rightarrow> code \<times> code"
 and src_tgt_repl :: "trail_repl \<Rightarrow> code_repl \<times> code_repl"
where 
  "src_tgt (Refl M) = (M,M)"
| "src_tgt (Trans q1 q2) = (fst (src_tgt q1), snd (src_tgt q2))"
| "b \<sharp> (\<sigma>,M2) \<Longrightarrow> src_tgt (Beta \<sigma> b M1 M2) = (App (Lam \<sigma> b M1) M2, M1[b\<mapsto>M2])"
| "u \<sharp> (\<sigma>,M1) \<Longrightarrow> src_tgt (BetaBox \<sigma> M1 u M2) = (LetBox \<sigma> (Box M1) u M2, M2{u\<mapsto>M1})"
| "src_tgt (ITI q \<theta>) = (IInsp (Box (snd (src_tgt q))) \<theta>, qcfold q \<theta>)"
| "src_tgt (OTI q \<theta>) = (OInsp \<theta>, qcfold q \<theta>)"
| "b \<sharp> \<sigma> \<Longrightarrow> src_tgt (QLam \<sigma> b q) = (Lam \<sigma> b (fst (src_tgt q)), Lam \<sigma> b (snd (src_tgt q)))"
| "src_tgt (QApp q1 q2) 
    = (App (fst (src_tgt q1)) (fst (src_tgt q2)), App (snd (src_tgt q1)) (snd (src_tgt q2)))"
| "u \<sharp> (\<sigma>,q1) \<Longrightarrow> src_tgt (QLet \<sigma> q1 u q2)
    = (LetBox \<sigma> (fst (src_tgt q1)) u (fst (src_tgt q2)), LetBox \<sigma> (snd (src_tgt q1)) u (snd (src_tgt q2)))"
| "src_tgt (ITB q \<zeta>) = (IInsp (fst (src_tgt q)) (fst (src_tgt_repl \<zeta>)), IInsp (snd (src_tgt q)) (snd (src_tgt_repl \<zeta>)))"
| "src_tgt (OTB \<zeta>) = (OInsp (fst (src_tgt_repl \<zeta>)), OInsp (snd (src_tgt_repl \<zeta>)))"
| "src_tgt_repl (QrNil q) = (CrNil (fst (src_tgt q)), CrNil (snd (src_tgt q)))"
| "src_tgt_repl (QrCons l q \<zeta>) 
    = (CrCons l (fst (src_tgt q)) (fst (src_tgt_repl \<zeta>)), CrCons l (snd (src_tgt q)) (snd (src_tgt_repl \<zeta>)))"
  apply(finite_guess add: qcfold_def)+
  apply(rule TrueI)+
  apply(auto simp: abs_fresh subst_fresh usubst_fresh)
  apply(fresh_guess add:qcfold_def)+
done

lemma src_tgt_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  shows "pi\<bullet>(src_tgt q) = src_tgt (pi\<bullet>q)"
  and "pi'\<bullet>(src_tgt q) = src_tgt (pi'\<bullet>q)"
  and "pi\<bullet>(src_tgt_repl qs) = src_tgt_repl (pi\<bullet>qs)"
  and "pi'\<bullet>(src_tgt_repl qs) = src_tgt_repl (pi'\<bullet>qs)"
apply (nominal_induct q and qs rule:trail_trail_repl.strong_inducts)
apply (auto simp: fst_eqvt snd_eqvt subst_eqvt qcfold_eqvt perm_bij fresh_bij)
done

lemma fst_snd_fresh[simp]:
  fixes a::var
  fixes u::uvar
  shows "a \<sharp> x \<Longrightarrow> a \<sharp> fst x"
  and "u \<sharp> x \<Longrightarrow> u \<sharp> fst x"
  and "a \<sharp> x \<Longrightarrow> a \<sharp> snd x"
  and "u \<sharp> x \<Longrightarrow> u \<sharp> snd x"
by (cases x,auto)+

lemma src_tgt_fresh[simp]:
  fixes q::trail
  fixes qs::trail_repl
  fixes a::var
  fixes u::uvar
  shows "a \<sharp> q \<Longrightarrow> a \<sharp> src_tgt q"
  and "u \<sharp> q \<Longrightarrow> u \<sharp> src_tgt q"
  and "a \<sharp> qs \<Longrightarrow> a \<sharp> src_tgt_repl qs"
  and "u \<sharp> qs \<Longrightarrow> u \<sharp> src_tgt_repl qs"
apply (nominal_induct q and qs rule:trail_trail_repl.strong_inducts)
apply (auto simp: subst_fresh qcfold_fresh fresh_prod abs_fresh usubst_fresh)
done

definition src :: "trail \<Rightarrow> code" 
where "src q = fst (src_tgt q)"

definition tgt :: "trail \<Rightarrow> code" 
where "tgt q = snd (src_tgt q)"

lemma src_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  shows "pi\<bullet>(src q) = src (pi\<bullet>q)"
  and "pi'\<bullet>(src q) = src (pi'\<bullet>q)"
by (simp_all add:src_def fst_eqvt src_tgt_eqvt)

lemma tgt_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  shows "pi\<bullet>(tgt q) = tgt (pi\<bullet>q)"
  and "pi'\<bullet>(tgt q) = tgt (pi'\<bullet>q)"
by (simp_all add:tgt_def snd_eqvt src_tgt_eqvt)

lemma src_fresh[simp]:
  fixes a::var
  fixes u::uvar
  fixes q::trail
  shows "a \<sharp> q \<Longrightarrow> a \<sharp> src q"
  and "u \<sharp> q \<Longrightarrow> u \<sharp> src q"
by (simp_all add:src_def)

lemma tgt_fresh[simp]:
  fixes a::var
  fixes u::uvar
  fixes q::trail
  shows "a \<sharp> q \<Longrightarrow> a \<sharp> tgt q"
  and "u \<sharp> q \<Longrightarrow> u \<sharp> tgt q"
by (simp_all add:tgt_def)

(* term to code (Def.5) *)
nominal_primrec t2c :: "trm \<Rightarrow> code"
and t2c_repl :: "trm_repl \<Rightarrow> code_repl"
where
  "t2c (TVar a) = Var a"
| "t2c (TUVar u) = UVar u"
| "a \<sharp> \<sigma> \<Longrightarrow> t2c (TLam \<sigma> a M) = Lam \<sigma> a (t2c M)"
| "t2c (TApp M N) = App (t2c M) (t2c N)"
| "u \<sharp> (\<sigma>,M) \<Longrightarrow> t2c (TLetBox \<sigma> M u N) = LetBox \<sigma> (t2c M) u (t2c N)"
| "t2c (TBox q M) = Box (src q)"
| "t2c (TOInsp \<theta>) = OInsp (t2c_repl \<theta>)"
| "t2c (TIInsp M \<theta>) = IInsp (t2c M) (t2c_repl \<theta>)"
| "t2c_repl (TrNil M) = CrNil (t2c M)"
| "t2c_repl (TrCons l M \<theta>) = CrCons l (t2c M) (t2c_repl \<theta>)"
  apply(finite_guess)+
  apply(rule TrueI)+
  defer
  defer
  apply(fresh_guess)+
  apply(simp add: abs_fresh)+
done

lemma t2c_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  shows "pi\<bullet>(t2c M) = t2c (pi\<bullet>M)"
  and "pi'\<bullet>(t2c M) = t2c (pi'\<bullet>M)"
  and "pi\<bullet>(t2c_repl Ms) = t2c_repl (pi\<bullet>Ms)"
  and "pi'\<bullet>(t2c_repl Ms) = t2c_repl (pi'\<bullet>Ms)"
by (nominal_induct M and Ms rule:trm_trm_repl.strong_inducts,auto simp:src_eqvt perm_bij fresh_bij)

lemma t2c_fresh[simp]:
  fixes M::trm
  and a::var
  and u::uvar
  shows "a \<sharp> M \<Longrightarrow> a \<sharp> t2c M"
  and "u \<sharp> M \<Longrightarrow> u \<sharp> t2c M"
  and "a \<sharp> Ms \<Longrightarrow> a \<sharp> t2c_repl Ms"
  and "u \<sharp> Ms \<Longrightarrow> u \<sharp> t2c_repl Ms"
by (nominal_induct M and Ms rule:trm_trm_repl.strong_inducts,auto simp:abs_fresh)

(* the subst. \<delta> of Def.2 is lifted to trails as explained in the paper *)
nominal_primrec (freshness_context : "(u::uvar,N::code)")
 qusubst :: "trail \<Rightarrow> uvar \<Rightarrow> code \<Rightarrow> trail" ("_{_\<mapsto>_}" [100,60] 100)
 and qusubst_repl :: "trail_repl \<Rightarrow> uvar \<Rightarrow> code \<Rightarrow> trail_repl" ("_{_\<mapsto>_}" [100,60] 100)
where 
  "(Refl M){u\<mapsto>N} = Refl (M{u\<mapsto>N})"
| "(Trans q1 q2){u\<mapsto>N} = Trans (q1{u\<mapsto>N}) (q2{u\<mapsto>N})"
| "b \<sharp> (\<sigma>,M2,N) \<Longrightarrow> (Beta \<sigma> b M1 M2){u\<mapsto>N} = Beta (\<sigma>{u\<mapsto>N}) b (M1{u\<mapsto>N}) (M2{u\<mapsto>N})"
| "v \<sharp> (\<sigma>,M1,u,N) \<Longrightarrow> (BetaBox \<sigma> M1 v M2){u\<mapsto>N} = BetaBox (\<sigma>{u\<mapsto>N}) (M1{u\<mapsto>N}) v (M2{u\<mapsto>N})"
| "(ITI q \<theta>){u\<mapsto>N} = ITI (q{u\<mapsto>N}) (\<theta>{u\<mapsto>N})"
| "(OTI q \<theta>){u\<mapsto>N} = OTI (q{u\<mapsto>N}) (\<theta>{u\<mapsto>N})"
| "b \<sharp> (\<sigma>,N) \<Longrightarrow> (QLam \<sigma> b q){u\<mapsto>N} = QLam (\<sigma>{u\<mapsto>N}) b (q{u\<mapsto>N})"
| "(QApp q1 q2){u\<mapsto>N} = QApp (q1{u\<mapsto>N}) (q2{u\<mapsto>N})"
| "v \<sharp> (\<sigma>,q1,u,N) \<Longrightarrow> (QLet \<sigma> q1 v q2){u\<mapsto>N} = QLet (\<sigma>{u\<mapsto>N}) (q1{u\<mapsto>N}) v (q2{u\<mapsto>N})"
| "(ITB q \<zeta>){u\<mapsto>N} = ITB (q{u\<mapsto>N}) (\<zeta>{u\<mapsto>N})"
| "(OTB \<zeta>){u\<mapsto>N} = OTB (\<zeta>{u\<mapsto>N})"
| "(QrNil q){u\<mapsto>N} = QrNil (q{u\<mapsto>N})"
| "(QrCons l q \<zeta>){u\<mapsto>N} = QrCons l (q{u\<mapsto>N}) (\<zeta>{u\<mapsto>N})"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto simp:abs_fresh usubst_fresh)
  apply(fresh_guess)+
done

lemma qusubst_eqvt[eqvt]:
  fixes \<pi>::"var prm"
  fixes \<pi>'::"uvar prm"
  fixes q::trail
  fixes qs :: trail_repl
  fixes x::uvar
  shows "\<pi> \<bullet> (q{x\<mapsto>N}) = (\<pi>\<bullet>q){\<pi>\<bullet>x\<mapsto>\<pi>\<bullet>N}"
  and "\<pi>' \<bullet> (q{x\<mapsto>N}) = (\<pi>'\<bullet>q){\<pi>'\<bullet>x\<mapsto>\<pi>'\<bullet>N}"
  and "\<pi> \<bullet> (qs{x\<mapsto>N}) = (\<pi>\<bullet>qs){\<pi>\<bullet>x\<mapsto>\<pi>\<bullet>N}"
  and "\<pi>' \<bullet> (qs{x\<mapsto>N}) = (\<pi>'\<bullet>qs){\<pi>'\<bullet>x\<mapsto>\<pi>'\<bullet>N}"
apply (nominal_induct q and qs avoiding:x N rule:trail_trail_repl.strong_inducts)
apply (auto simp:subst_eqvt fresh_bij perm_bij)
done

lemma qusubst_fresh_t[simp]: 
  fixes q::trail
  fixes qs::trail_repl
  fixes N::code
  fixes x::var
  shows "x \<sharp> [y].q \<Longrightarrow> x \<sharp> N \<Longrightarrow> x \<sharp> (q{y\<mapsto>N})"
  and "x \<sharp> [y].qs \<Longrightarrow> x \<sharp> N \<Longrightarrow> x \<sharp> (qs{y\<mapsto>N})"
  apply(nominal_induct q and qs avoiding: y N rule:trail_trail_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_trail_label subst_fresh usubst_fresh)
done

lemma qusubst_fresh_u[simp]: 
  fixes q::trail
  fixes qs::trail_repl
  fixes N::code
  fixes x::uvar
  shows "x \<sharp> [y].q \<Longrightarrow> x \<sharp> N \<Longrightarrow> x \<sharp> (q{y\<mapsto>N})"
  and "x \<sharp> [y].qs \<Longrightarrow> x \<sharp> N \<Longrightarrow> x \<sharp> (qs{y\<mapsto>N})"
  apply(nominal_induct q and qs avoiding: y N rule:trail_trail_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_trail_label subst_fresh usubst_fresh)
done

lemmas qusubst_fresh = qusubst_fresh_t qusubst_fresh_u

(* the subst. \<gamma> of Def.2 is lifted to trails as explained in the paper *)
nominal_primrec (freshness_context: "(a::var,N::code)")
     qsubst :: "trail \<Rightarrow> var \<Rightarrow> code \<Rightarrow> trail" ("_[_\<mapsto>_]" [100,60] 100)
 and qsubst_repl :: "trail_repl \<Rightarrow> var \<Rightarrow> code \<Rightarrow> trail_repl" ("_[_\<mapsto>_]" [100,60] 100)
where 
  "(Refl M)[a\<mapsto>N] = Refl (M[a\<mapsto>N])"
| "(Trans q1 q2)[a\<mapsto>N] = Trans (q1[a\<mapsto>N]) (q2[a\<mapsto>N])"
| "b \<sharp> (\<sigma>,M2,a,N) \<Longrightarrow> (Beta \<sigma> b M1 M2)[a\<mapsto>N] = Beta (\<sigma>[a\<mapsto>N]) b (M1[a\<mapsto>N]) (M2[a\<mapsto>N])"
| "u \<sharp> (\<sigma>,M1,N) \<Longrightarrow> (BetaBox \<sigma> M1 u M2)[a\<mapsto>N] = BetaBox (\<sigma>[a\<mapsto>N]) (M1[a\<mapsto>N]) u (M2[a\<mapsto>N])"
| "(ITI q \<theta>)[a\<mapsto>N] = ITI (q[a\<mapsto>N]) (\<theta>[a\<mapsto>N])" (*** subst on \<theta> not needed but harmless (?) ***)
| "(OTI q \<theta>)[a\<mapsto>N] = OTI (q[a\<mapsto>N]) (\<theta>[a\<mapsto>N])"
| "b \<sharp> (\<sigma>,a,N) \<Longrightarrow> (QLam \<sigma> b q)[a\<mapsto>N] = QLam (\<sigma>[a\<mapsto>N]) b (q[a\<mapsto>N])"
| "(QApp q1 q2)[a\<mapsto>N] = QApp (q1[a\<mapsto>N]) (q2[a\<mapsto>N])"
| "u \<sharp> (\<sigma>,q1,N) \<Longrightarrow> (QLet \<sigma> q1 u q2)[a\<mapsto>N] = QLet (\<sigma>[a\<mapsto>N]) (q1[a\<mapsto>N]) u (q2[a\<mapsto>N])"
| "(ITB q \<zeta>)[a\<mapsto>N] = ITB (q[a\<mapsto>N]) (\<zeta>[a\<mapsto>N])" (*** subst on \<zeta> not needed but harmless (?) ***)
| "(OTB \<zeta>)[a\<mapsto>N] = OTB (\<zeta>[a\<mapsto>N])"
| "(QrNil q)[a\<mapsto>N] = QrNil (q[a\<mapsto>N])"
| "(QrCons l q \<zeta>)[a\<mapsto>N] = QrCons l (q[a\<mapsto>N]) (\<zeta>[a\<mapsto>N])"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto simp add: abs_fresh fresh_prod code.inject fresh_ty subst_fresh)
  apply(fresh_guess)+
done

lemma qsubst_fresh_t[simp]: 
  fixes q::trail
  fixes qs::trail_repl
  fixes x::var
  shows "x \<sharp> [y].q \<Longrightarrow> x \<sharp> N \<Longrightarrow> x \<sharp> (q[y\<mapsto>N])"
  and "x \<sharp> [y].qs \<Longrightarrow> x \<sharp> N \<Longrightarrow> x \<sharp> (qs[y\<mapsto>N])"
  apply(nominal_induct q and qs avoiding: y N rule:trail_trail_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_trail_label subst_fresh)
done

lemma qsubst_fresh_u[simp]: 
  fixes q::trail
  fixes qs::trail_repl
  fixes x::uvar
  shows "x \<sharp> [y].q \<Longrightarrow> x \<sharp> N \<Longrightarrow> x \<sharp> (q[y\<mapsto>N])"
  and "x \<sharp> [y].qs \<Longrightarrow> x \<sharp> N \<Longrightarrow> x \<sharp> (qs[y\<mapsto>N])"
  apply(nominal_induct q and qs avoiding: y N rule:trail_trail_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_trail_label subst_fresh)
done

lemmas qsubst_fresh = qsubst_fresh_t qsubst_fresh_u

lemma qsubst_eqvt[eqvt]:
  fixes \<pi>::"var prm"
  fixes \<pi>'::"uvar prm"
  fixes q::trail
  fixes qs :: trail_repl
  fixes x::var
  shows "\<pi> \<bullet> (q[x\<mapsto>N]) = (\<pi>\<bullet>q)[\<pi>\<bullet>x\<mapsto>\<pi>\<bullet>N]"
  and "\<pi>' \<bullet> (q[x\<mapsto>N]) = (\<pi>'\<bullet>q)[\<pi>'\<bullet>x\<mapsto>\<pi>'\<bullet>N]"
  and "\<pi> \<bullet> (qs[x\<mapsto>N]) = (\<pi>\<bullet>qs)[\<pi>\<bullet>x\<mapsto>\<pi>\<bullet>N]"
  and "\<pi>' \<bullet> (qs[x\<mapsto>N]) = (\<pi>'\<bullet>qs)[\<pi>'\<bullet>x\<mapsto>\<pi>'\<bullet>N]"
apply (nominal_induct q and qs avoiding:x N rule:trail_trail_repl.strong_inducts)
apply (auto simp:perm_bij fresh_bij subst_eqvt)
done

(* the subst. \<gamma> of Def.1 is lifted to terms as explained in the paper *)
nominal_primrec  (freshness_context : "(a::var,R::trm)")
 tsubst :: "trm \<Rightarrow> var \<Rightarrow> trm \<Rightarrow> trm" ("_[_\<mapsto>_]" [100,60] 100)
 and tsubst_repl :: "trm_repl \<Rightarrow> var \<Rightarrow> trm \<Rightarrow> trm_repl" ("_[_\<mapsto>_]" [100,60] 100)
where 
  "(TVar b)[a\<mapsto>R] = (if a = b then R else TVar b)"
| "(TUVar u)[a\<mapsto>R] = TUVar u"
| "y \<sharp> (\<tau>,a,R) \<Longrightarrow> (TLam \<tau> y M)[a\<mapsto>R] = TLam (\<tau>[a\<mapsto>t2c R]) y (M[a\<mapsto>R])"
| "(TApp M1 M2)[a\<mapsto>R] = TApp (M1[a\<mapsto>R]) (M2[a\<mapsto>R])"
| "(TBox q M)[a\<mapsto>R] = TBox (q[a\<mapsto>t2c R]) (M[a\<mapsto>R])" (*** no need to subst inside Box, but makes proofs of freshness simpler ***)
| "v \<sharp> (\<tau>,M1,R) \<Longrightarrow> (TLetBox \<tau> M1 v M2)[a\<mapsto>R] = TLetBox (\<tau>[a\<mapsto>t2c R]) (M1[a\<mapsto>R]) v (M2[a\<mapsto>R])"
| "(TOInsp \<theta>)[a\<mapsto>R] = TOInsp (\<theta>[a\<mapsto>R])"
| "(TIInsp M \<theta>)[a\<mapsto>R] = TIInsp (M[a\<mapsto>R]) (\<theta>[a\<mapsto>R])"
| "(TrNil M)[a\<mapsto>R] = TrNil (M[a\<mapsto>R])"
| "(TrCons l M \<theta>)[a\<mapsto>R] = TrCons l (M[a\<mapsto>R]) (\<theta>[a\<mapsto>R])"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto simp add: abs_fresh fresh_prod)+
  apply(fresh_guess)+
done

definition uvar_qsubst :: "uvar \<Rightarrow> trail \<Rightarrow> uvar \<Rightarrow> trail"
where
  "uvar_qsubst v q u = (if u = v then q else Refl (UVar u))"

lemma uvar_qsubst_fresh_t[simp] :
  fixes u v::uvar
  fixes q::trail
  fixes x::var
  shows "x \<sharp> u \<Longrightarrow> x \<sharp> q \<Longrightarrow> x \<sharp> uvar_qsubst v q u"
by (simp add:uvar_qsubst_def abs_fresh)

lemma uvar_qsubst_fresh_u[simp] :
  fixes u v::uvar
  fixes q::trail
  fixes x::uvar
  shows "x \<sharp> [v].u \<Longrightarrow> x \<sharp> q \<Longrightarrow> x \<sharp> uvar_qsubst v q u"
apply (simp add:uvar_qsubst_def abs_fresh)
apply (cases "x=v",auto simp:at_fresh at_uvar_inst)
done

lemma uvar_qsubst_fresh_u'[simp] :
  fixes u v::uvar
  fixes q::trail
  fixes x::uvar
  shows "x \<sharp> u \<Longrightarrow> x \<sharp> q \<Longrightarrow> x \<sharp> uvar_qsubst v q u"
by (simp add:uvar_qsubst_def abs_fresh)

lemmas uvar_qsubst_fresh[simp] = uvar_qsubst_fresh_t uvar_qsubst_fresh_u uvar_qsubst_fresh_u'

lemma uvar_qsubst_eqvt[eqvt]:
  fixes \<pi>::"var prm"
  fixes \<pi>'::"uvar prm"
  fixes u v::uvar
  fixes q :: trail
  shows "\<pi> \<bullet> (uvar_qsubst v q u) = uvar_qsubst (\<pi>\<bullet>v) (\<pi>\<bullet>q) (\<pi>\<bullet>u)"
  and "\<pi>' \<bullet> (uvar_qsubst v q u) = uvar_qsubst (\<pi>'\<bullet>v) (\<pi>'\<bullet>q) (\<pi>'\<bullet>u)"
by (simp_all add:uvar_qsubst_def perm_bij)

(* This is the \<times>-substitution of Def.4 *)
nominal_primrec (freshness_context : "(v::uvar,r'::code)")
 tusubst :: "trm \<Rightarrow> uvar \<Rightarrow> trm \<Rightarrow> trail \<Rightarrow> code \<Rightarrow> trail" ("_\<lbrace>_\<mapsto>_,_,_\<rbrace>" [100,60,60,60] 100)
 and tusubst_repl :: "trm_repl \<Rightarrow> uvar \<Rightarrow> trm \<Rightarrow> trail \<Rightarrow> code \<Rightarrow> trail_repl" ("_\<lbrace>_\<mapsto>_,_,_\<rbrace>" [100,60,60,60] 100)
where 
  "(TVar b)\<lbrace>v\<mapsto>R,q,r'\<rbrace> = Refl (Var b)"
| "(TUVar u)\<lbrace>v\<mapsto>R,q,r'\<rbrace> = uvar_qsubst v q u"
| "y \<sharp> (\<tau>,R,q,r') \<Longrightarrow> (TLam \<tau> y M)\<lbrace>v\<mapsto>R,q,r'\<rbrace> = QLam (\<tau>{v\<mapsto>r'}) y (M\<lbrace>v\<mapsto>R,q,r'\<rbrace>)"
| "(TApp M1 M2)\<lbrace>v\<mapsto>R,q,r'\<rbrace> = QApp (M1\<lbrace>v\<mapsto>R,q,r'\<rbrace>) (M2\<lbrace>v\<mapsto>R,q,r'\<rbrace>)"
| "(TBox q0 M)\<lbrace>v\<mapsto>R,q,r'\<rbrace> = Refl ((src q0){v\<mapsto>r'})"
| "u \<sharp> (\<tau>,v,M1,R,q,r') \<Longrightarrow> (TLetBox \<tau> M1 u M2)\<lbrace>v\<mapsto>R,q,r'\<rbrace> = QLet (\<tau>{v\<mapsto>r'}) (M1\<lbrace>v\<mapsto>R,q,r'\<rbrace>) u (M2\<lbrace>v\<mapsto>R,q,r'\<rbrace>)"
| "(TOInsp \<theta>)\<lbrace>v\<mapsto>R,q,r'\<rbrace> = OTB (\<theta>\<lbrace>v\<mapsto>R,q,r'\<rbrace>)"
| "(TIInsp M \<theta>)\<lbrace>v\<mapsto>R,q,r'\<rbrace> = ITB (M\<lbrace>v\<mapsto>R,q,r'\<rbrace>) (\<theta>\<lbrace>v\<mapsto>R,q,r'\<rbrace>)"
| "(TrNil M)\<lbrace>v\<mapsto>R,q,r'\<rbrace> = QrNil (M\<lbrace>v\<mapsto>R,q,r'\<rbrace>)"
| "(TrCons l M \<theta>)\<lbrace>v\<mapsto>R,q,r'\<rbrace> = QrCons l (M\<lbrace>v\<mapsto>R,q,r'\<rbrace>) (\<theta>\<lbrace>v\<mapsto>R,q,r'\<rbrace>)"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto simp add: abs_fresh fresh_prod code.inject src_fresh usubst_fresh)+
  apply(fresh_guess)+
done

lemma tusubst_fresh_t[simp]: 
  fixes M::trm
  fixes Ms::trm_repl
  fixes x::var
  shows "x \<sharp> [y].M \<Longrightarrow> x \<sharp> (N,q,N') \<Longrightarrow> x \<sharp> (M\<lbrace>y\<mapsto>N,q,N'\<rbrace>)"
  and "x \<sharp> [y].Ms \<Longrightarrow> x \<sharp> (N,q,N') \<Longrightarrow> x \<sharp> (Ms\<lbrace>y\<mapsto>N,q,N'\<rbrace>)"
  apply(nominal_induct M and Ms avoiding: y N q N' rule:trm_trm_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_trail_label usubst_fresh)
done

lemma tusubst_fresh_u[simp]: 
  fixes M::trm
  fixes Ms::trm_repl
  fixes x::uvar
  shows "x \<sharp> [y].M \<Longrightarrow> x \<sharp> (N,q,N') \<Longrightarrow> x \<sharp> (M\<lbrace>y\<mapsto>N,q,N'\<rbrace>)"
  and "x \<sharp> [y].Ms \<Longrightarrow> x \<sharp> (N,q,N') \<Longrightarrow> x \<sharp> (Ms\<lbrace>y\<mapsto>N,q,N'\<rbrace>)"
  apply(nominal_induct M and Ms avoiding: y N q N' rule:trm_trm_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_trail_label usubst_fresh)
done

lemmas tusubst_fresh = tusubst_fresh_t tusubst_fresh_u

lemma tusubst_eqvt[eqvt] :
  fixes M N::trm
  fixes Ms::trm_repl
  fixes u::uvar
  fixes q::trail
  fixes n::code
  fixes \<pi>::"var prm"
  fixes \<pi>'::"uvar prm"
  shows "\<pi> \<bullet> (M\<lbrace>u\<mapsto>N,q,n'\<rbrace>) = (\<pi>\<bullet>M)\<lbrace>\<pi>\<bullet>u\<mapsto>\<pi>\<bullet>N,\<pi>\<bullet>q,\<pi>\<bullet>n'\<rbrace>"
  and "\<pi>' \<bullet> (M\<lbrace>u\<mapsto>N,q,n'\<rbrace>) = (\<pi>'\<bullet>M)\<lbrace>\<pi>'\<bullet>u\<mapsto>\<pi>'\<bullet>N,\<pi>'\<bullet>q,\<pi>'\<bullet>n'\<rbrace>"
  and "\<pi> \<bullet> (Ms\<lbrace>u\<mapsto>N,q,n'\<rbrace>) = (\<pi>\<bullet>Ms)\<lbrace>\<pi>\<bullet>u\<mapsto>\<pi>\<bullet>N,\<pi>\<bullet>q,\<pi>\<bullet>n'\<rbrace>"
  and "\<pi>' \<bullet> (Ms\<lbrace>u\<mapsto>N,q,n'\<rbrace>) = (\<pi>'\<bullet>Ms)\<lbrace>\<pi>'\<bullet>u\<mapsto>\<pi>'\<bullet>N,\<pi>'\<bullet>q,\<pi>'\<bullet>n'\<rbrace>"
  apply(nominal_induct M and Ms avoiding: u N q n' rule:trm_trm_repl.strong_inducts)
  apply(auto simp add:perm_bij fresh_bij eqvt subst_eqvt src_eqvt uvar_qsubst_eqvt)
done

(* corresponds to the \<ltimes>-substitution of Def.4 *)
nominal_primrec (freshness_context : "(v::uvar, r'::code)")
 tlusubst :: "trm \<Rightarrow> uvar \<Rightarrow> trm \<Rightarrow> trail \<Rightarrow> code \<Rightarrow> trm" ("_{_\<mapsto>_,_,_}" [100,60,60,60] 100)
 and tlusubst_repl :: "trm_repl \<Rightarrow> uvar \<Rightarrow> trm \<Rightarrow> trail \<Rightarrow> code \<Rightarrow> trm_repl" ("_{_\<mapsto>_,_,_}" [100,60,60,60] 100)
where 
  "(TVar b){v\<mapsto>R,q,r'} = TVar b"
| "(TUVar u){v\<mapsto>R,q,r'} = (if u = v then R else TUVar u)"
| "y \<sharp> (\<tau>,R,q,r') \<Longrightarrow> (TLam \<tau> y M){v\<mapsto>R,q,r'} = TLam (\<tau>{v\<mapsto>r'}) y (M{v\<mapsto>R,q,r'})"
| "(TApp M1 M2){v\<mapsto>R,q,r'} = TApp (M1{v\<mapsto>R,q,r'}) (M2{v\<mapsto>R,q,r'})"
| "(TBox q0 M){v\<mapsto>R,q,r'} = TBox (Trans (q0{v\<mapsto>r'}) (R\<lbrace>v\<mapsto>R,q,r'\<rbrace>)) (M{v\<mapsto>R,q,r'})"
| "u \<sharp> (\<tau>,M1,v,R,q,r') \<Longrightarrow> (TLetBox \<tau> M1 u M2){v\<mapsto>R,q,r'} = TLetBox (\<tau>{v\<mapsto>r'}) (M1{v\<mapsto>R,q,r'}) u (M2{v\<mapsto>R,q,r'})"
| "(TOInsp \<theta>){v\<mapsto>R,q,r'} = TOInsp (\<theta>{v\<mapsto>R,q,r'})"
| "(TIInsp M \<theta>){v\<mapsto>R,q,r'} = TIInsp (M{v\<mapsto>R,q,r'}) (\<theta>{v\<mapsto>R,q,r'})"
| "(TrNil M){v\<mapsto>R,q,r'} = TrNil (M{v\<mapsto>R,q,r'})"
| "(TrCons l M \<theta>){v\<mapsto>R,q,r'} = TrCons l (M{v\<mapsto>R,q,r'}) (\<theta>{v\<mapsto>R,q,r'})"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto simp: abs_fresh fresh_prod code.inject fresh_ty tusubst_fresh qusubst_fresh)+
  apply(fresh_guess)+
done

(* Identity substitutions *)

fun
  id_subst :: "ctx \<Rightarrow> subst"   
where
  "id_subst []        = []"
| "id_subst ((y,\<tau>)#\<Gamma>) = (y,Var y)#(id_subst \<Gamma>)"

lemma id_lookup: 
  "lookup (id_subst \<Gamma>) x = Var x"
  by(induct \<Gamma>,auto)

lemma id_subst_fresh:
  fixes x::var
  and u::uvar
  shows "x \<sharp> \<Gamma> \<Longrightarrow> x \<sharp> id_subst \<Gamma>"
  and "u \<sharp> \<Gamma> \<Longrightarrow> u \<sharp> id_subst \<Gamma>"
  apply(induct \<Gamma>)
  apply(auto simp add:fresh_list_nil fresh_list_cons fresh_prod fresh_atm)
  done


fun
  id_usubst :: "uctx \<Rightarrow> usubst"   
where
  "id_usubst []        = []"
| "id_usubst ((u,\<tau>)#\<Delta>) = (u,UVar u)#(id_usubst \<Delta>)"

lemma id_usubst_fresh:
  fixes x::var
  and u::uvar
  shows "x \<sharp> \<Delta> \<Longrightarrow> x \<sharp> id_usubst \<Delta>"
  and "u \<sharp> \<Delta> \<Longrightarrow> u \<sharp> id_usubst \<Delta>"
  apply(induct \<Delta>)
  apply(auto simp add:fresh_list_nil fresh_list_cons fresh_prod fresh_atm)
  done

lemma id_ulookup: 
  "ulookup (id_usubst \<Delta>) x = UVar x"
  by(induct \<Delta>,auto)


lemma id_subst_is_id:
  fixes \<sigma>::ty
  fixes M::code
  fixes Ms::code_repl
  shows "\<sigma><id_usubst \<Delta>,id_subst \<Gamma>> = \<sigma>"
  and "M<id_usubst \<Delta>,id_subst \<Gamma>> = M"
  and "Ms<id_usubst \<Delta>,id_subst \<Gamma>> = Ms"
  apply(nominal_induct \<sigma> and M and Ms avoiding: \<Delta> \<Gamma> rule:ty_code_code_repl.strong_inducts)
  apply(auto simp add:id_lookup id_ulookup id_subst_fresh id_usubst_fresh)
done

(* branch types, as defined in Sect. 2.2 *)
primrec bt :: "ty \<Rightarrow> nat \<Rightarrow> ty" where
  "bt \<sigma> 0 = \<sigma>"
| "bt \<sigma> (Suc n)= FunTy \<sigma> (bt \<sigma> n)"

definition bty :: "ty \<Rightarrow> trail_label \<Rightarrow> ty" where
  "bty \<sigma> l = bt \<sigma> (trail_label_arity l)"

lemma bt_eqvt:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  shows "pi\<bullet>(bt \<sigma> n) = bt (pi\<bullet>\<sigma>) n"
  and "pi'\<bullet>(bt \<sigma> n) = bt (pi'\<bullet>\<sigma>) n"
  apply (induct n,auto)
done

lemma trail_label_eqvt[simp]:
  fixes l::trail_label
  fixes pi::"var prm"
  fixes pi'::"uvar prm"
  shows "pi\<bullet>l = l"
  and "pi'\<bullet>l = l"
by (nominal_induct l rule:trail_label.strong_induct,auto)

lemma bty_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  shows "pi\<bullet>(bty \<sigma> l) = bty (pi\<bullet>\<sigma>) (pi\<bullet>l)"
  and "pi'\<bullet>(bty \<sigma> l) = bty (pi'\<bullet>\<sigma>) (pi'\<bullet>l)"
by (auto simp:bty_def bt_eqvt)

(* Typing rules for codes and trails, from Fig.2 (amended) *)
inductive wfc :: "uctx \<Rightarrow> ctx \<Rightarrow> code \<Rightarrow> ty \<Rightarrow> bool" ("_;_ \<turnstile> _ : _" [60,60,60,60] 60)
  and wfc_repl :: "uctx \<Rightarrow> code_repl \<Rightarrow> ty \<Rightarrow> bool" ("_ \<turnstile> _ : _" [60,60,60] 60)
  and wf_ctx :: "uctx \<Rightarrow> ctx \<Rightarrow> bool" ("_ \<turnstile> _" [60,60] 60)
  and wf_uctx :: "uctx \<Rightarrow> bool" ("\<turnstile> _" [60] 60)
  and wf_ty :: "uctx \<Rightarrow> ctx \<Rightarrow> ty \<Rightarrow> bool" ("_;_ \<turnstile> _" [60,60,60] 60)
where
  var_ty : "\<lbrakk>(x,\<tau>) \<in> set \<Gamma>; \<Delta> \<turnstile> \<Gamma> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> Var x : \<tau>"
| uvar_ty : "\<lbrakk>(u,\<tau>) \<in> set \<Delta>; \<Delta> \<turnstile> \<Gamma> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> UVar u : \<tau>"
| lam_ty : "\<lbrakk>\<Delta>;(x,\<tau>)#\<Gamma> \<turnstile> M : \<tau>'; x \<sharp> (\<Gamma>,\<Delta>,\<tau>,\<tau>') \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> Lam \<tau> x M : FunTy \<tau> \<tau>'"
| app_ty : "\<lbrakk>\<Delta>;\<Gamma> \<turnstile> M : FunTy \<tau> \<tau>'  ; \<Delta>;\<Gamma> \<turnstile> N : \<tau> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> App M N : \<tau>'"
| box_ty : "\<lbrakk>\<Delta>;[] \<turnstile> M : \<tau>; \<Delta> \<turnstile> \<Gamma> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> Box M : BoxTy M \<tau>"
| letbox_ty : "\<lbrakk>\<Delta>;\<Gamma> \<turnstile> M : BoxTy N \<tau>  ; (u,\<tau>)#\<Delta>;\<Gamma> \<turnstile> N : \<tau>'; u \<sharp> (M,\<Delta>,\<Gamma>,\<tau>,N) \<rbrakk> \<Longrightarrow> 
                \<Delta>;\<Gamma> \<turnstile> LetBox \<tau> M u N : \<tau>'{u\<mapsto>N}"
| oinsp_ty : "\<lbrakk> \<Delta> \<turnstile> \<theta> : \<tau>; \<Delta> \<turnstile> \<Gamma>\<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> OInsp \<theta> : \<tau>"
| iinsp_ty : "\<lbrakk> \<Delta>;\<Gamma> \<turnstile> M : BoxTy N \<sigma>; \<Delta> \<turnstile> \<theta> : \<tau> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> IInsp M \<theta> : \<tau>"
(* wf_repl *)
| crnil_ty : "\<Delta>;[] \<turnstile> M : \<tau> \<Longrightarrow> \<Delta> \<turnstile> CrNil M : \<tau>" (* M is the default branch *)
| crcons_ty : "\<lbrakk>\<Delta> \<turnstile> \<theta> : \<tau>; \<Delta>;[] \<turnstile> M : bty \<tau> l \<rbrakk> \<Longrightarrow> \<Delta> \<turnstile> CrCons l M \<theta> : \<tau>"
(* wf_ty *)
| pty_ty : "\<Delta> \<turnstile> \<Gamma> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> PTy"
| funty_ty : "\<lbrakk> \<Delta>;\<Gamma> \<turnstile> \<sigma>; \<Delta>;\<Gamma> \<turnstile> \<tau> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> FunTy \<sigma> \<tau>"
| boxty_ty : "\<Delta>;\<Gamma> \<turnstile> s : \<sigma> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> BoxTy s \<sigma>"
(* wf_ctx *)
| nil_wf : "\<turnstile> \<Delta> \<Longrightarrow> \<Delta> \<turnstile> []"
| cons_wf : "\<lbrakk>a \<sharp> (\<Delta>,\<Gamma>); \<Delta>;\<Gamma> \<turnstile> \<tau>\<rbrakk> \<Longrightarrow> \<Delta> \<turnstile> (a,\<tau>)#\<Gamma>"
(* wf_uctx *)
| unil_wf : "\<turnstile> []"
| ucons_wf : "\<lbrakk>u \<sharp> \<Delta>; \<Delta>;[] \<turnstile> \<tau>\<rbrakk> \<Longrightarrow> \<turnstile> (u,\<tau>)#\<Delta>"

inductive wfq :: "uctx \<Rightarrow> ctx \<Rightarrow> trail \<Rightarrow> ty \<Rightarrow> code \<Rightarrow> code \<Rightarrow> bool" ("_;_ \<turnstile> _ : _ \<bar> _ \<triangleright> _" [60,60,60,60,60,60] 60)
  and wfq_repl :: "uctx \<Rightarrow> trail_repl \<Rightarrow> ty \<Rightarrow> code_repl \<Rightarrow> code_repl \<Rightarrow> bool" ("_ \<turnstile> _ : _ \<bar> _ \<triangleright> _" [60,60,60,60,60] 60)
where
  refl_ty : "\<Delta>;\<Gamma> \<turnstile> M : \<tau> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> Refl M : \<tau> \<bar> M \<triangleright> M"
| trans_ty : "\<lbrakk>\<Delta>;\<Gamma> \<turnstile> q1 : \<tau> \<bar> M \<triangleright> R; \<Delta>;\<Gamma> \<turnstile> q2 : \<tau> \<bar> M \<triangleright> R \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> Trans q1 q2 : \<tau> \<bar> M \<triangleright> R"
| beta_ty : "\<lbrakk>\<Delta>;(a,\<sigma>)#\<Gamma> \<turnstile> M : \<tau>; \<Delta>;\<Gamma> \<turnstile> N : \<sigma>; a \<sharp> (\<Gamma>,\<Delta>,N,\<sigma>,\<tau>) \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> Beta \<sigma> a M N : \<tau> \<bar> App (Lam \<sigma> a M) N \<triangleright> M[a\<mapsto>N]"
| betabox_ty : "\<lbrakk>\<Delta>;[] \<turnstile> M : \<sigma>; (u,\<sigma>)#\<Delta>;\<Gamma> \<turnstile> N : \<tau>; u \<sharp> (\<Delta>,\<Gamma>,M,\<sigma>,\<tau>) \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> BetaBox \<sigma> M u N : \<tau> \<bar> LetBox \<sigma> (Box M) u N \<triangleright> N{u\<mapsto>M}"
| iti_ty : "\<lbrakk> \<Delta>;[] \<turnstile> q : \<sigma> \<bar> M \<triangleright> N; \<Delta> \<turnstile> \<theta> : \<tau> ; \<Delta> \<turnstile> \<Gamma> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> ITI q \<theta> : \<tau> \<bar> IInsp (Box N) \<theta>  \<triangleright> scfold (skel_of_trail q) \<theta>"
| oti_ty : "\<lbrakk> \<Delta>;[] \<turnstile> q : \<sigma> \<bar> M \<triangleright> N; \<Delta> \<turnstile> \<theta> : \<tau>; \<Delta> \<turnstile> \<Gamma> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> OTI q \<theta> : \<tau> \<bar> OInsp \<theta>  \<triangleright> scfold (skel_of_trail q) \<theta>"
| qlam_ty : "\<lbrakk> \<Delta>;(a,\<sigma>)#\<Gamma> \<turnstile> q : \<tau> \<bar> M \<triangleright> N; a \<sharp> (\<Gamma>,\<Delta>,\<sigma>,\<tau>) \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> QLam \<sigma> a q : FunTy \<sigma> \<tau> \<bar> Lam \<sigma> a M \<triangleright> Lam \<sigma> a N"
| qapp_ty : "\<lbrakk> \<Delta>;\<Gamma> \<turnstile> q1 : FunTy \<sigma> \<tau> \<bar> M1 \<triangleright> N1; \<Delta>;\<Gamma> \<turnstile> q2 : \<sigma> \<bar> M2 \<triangleright> N2 \<rbrakk> 
             \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> QApp q1 q2 : \<tau> \<bar> App M1 M2 \<triangleright> App N1 N2"
| qlet_ty : "\<lbrakk>\<Delta>;\<Gamma> \<turnstile> q1 : BoxTy M \<sigma> \<bar> M1 \<triangleright> N1; (u,\<sigma>)#\<Delta>;\<Gamma> \<turnstile> q2 : \<tau> \<bar> M2 \<triangleright> N2; u \<sharp> (\<Gamma>,\<Delta>,\<sigma>,q1,M,M1,N1) \<rbrakk> 
             \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> QLet \<sigma> q1 u q2 : \<tau>{u\<mapsto>M} \<bar> LetBox \<sigma> M1 u M2 \<triangleright> LetBox \<sigma> N1 u N2"
| qitb_ty : "\<lbrakk>(\<Delta>;\<Gamma> \<turnstile> q : \<sigma> \<bar> M \<triangleright> N); \<Delta> \<turnstile> \<zeta> : \<tau> \<bar> \<theta> \<triangleright> \<theta>' \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> ITB q \<zeta> : \<tau> \<bar> IInsp M \<theta> \<triangleright> IInsp N \<theta>'"
| qotb_ty : "\<lbrakk>\<Delta> \<turnstile> \<zeta> : \<tau> \<bar> \<theta> \<triangleright> \<theta>'; \<Delta> \<turnstile> \<Gamma> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> OTB \<zeta> : \<tau> \<bar> OInsp \<theta> \<triangleright> OInsp \<theta>'"
| qrnil_ty : "\<Delta>;[] \<turnstile> q : \<tau> \<bar> M \<triangleright> N \<Longrightarrow> \<Delta> \<turnstile> QrNil q : \<tau> \<bar> CrNil M \<triangleright> CrNil N"
| qrcons_ty : "\<lbrakk>\<Delta> \<turnstile> \<zeta> : \<tau> \<bar> \<theta> \<triangleright> \<theta>'; \<Delta>;[] \<turnstile> q : bty \<tau> l \<bar> M \<triangleright> N \<rbrakk> 
               \<Longrightarrow> \<Delta> \<turnstile> QrCons l q \<zeta> : \<tau> \<bar> CrCons l M \<theta> \<triangleright> CrCons l N \<theta>'"

(* Typing rules for terms, from Fig.2 *)
inductive wf :: "uctx \<Rightarrow> ctx \<Rightarrow> trm \<Rightarrow> ty \<Rightarrow> code \<Rightarrow> bool" ("_;_ \<turnstile> _ : _ ; _" [60,60,60,60,60] 60)
  and wf_repl :: "uctx \<Rightarrow> trm_repl \<Rightarrow> ty \<Rightarrow> code_repl \<Rightarrow> bool" ("_ \<turnstile> _ : _ ; _" [60,60,60,60] 60)
where
  tvar_ty : "\<lbrakk>(x,\<tau>) \<in> set \<Gamma>; \<Delta> \<turnstile> \<Gamma> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> TVar x : \<tau> ; Var x"
| tuvar_ty : "\<lbrakk>(u,\<tau>) \<in> set \<Delta>; \<Delta> \<turnstile> \<Gamma> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> TUVar u : \<tau> ; UVar u"
| tlam_ty : "\<lbrakk>\<Delta>;(x,\<tau>)#\<Gamma> \<turnstile> M : \<tau>' ; M'; x \<sharp> (\<tau>,\<tau>',\<Gamma>,\<Delta>) \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> TLam \<tau> x M : FunTy \<tau> \<tau>' ; Lam \<tau> x M'"
| tapp_ty : "\<lbrakk>\<Delta>;\<Gamma> \<turnstile> M : FunTy \<tau> \<tau>' ; M' ; \<Delta>;\<Gamma> \<turnstile> N : \<tau> ; N'\<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> TApp M N : \<tau>' ; App M' N'"
| tbox_ty : "\<lbrakk>\<Delta>;[] \<turnstile> M : \<tau> ; M'; \<Delta>;[] \<turnstile> q : \<tau> \<bar> N \<triangleright> M'; \<Delta> \<turnstile> \<Gamma> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> TBox q M : BoxTy N \<tau> ; N"
| tletbox_ty : "\<lbrakk>\<Delta>;\<Gamma> \<turnstile> M : BoxTy M'' \<tau> ; M' ; (u,\<tau>)#\<Delta>;\<Gamma> \<turnstile> N : \<tau>' ; N'; u \<sharp> (M,M',M'',\<tau>,\<Delta>,\<Gamma>) \<rbrakk> \<Longrightarrow> 
                \<Delta>;\<Gamma> \<turnstile> TLetBox \<tau> M u N : \<tau>'{u\<mapsto>M''} ; LetBox \<tau> M' u N'"
| toinsp_ty : "\<lbrakk> \<Delta> \<turnstile> \<theta> : \<tau> ; \<theta>'; \<Delta> \<turnstile> \<Gamma>\<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> TOInsp \<theta> : \<tau> ; OInsp \<theta>'"
| tiinsp_ty : "\<lbrakk> \<Delta>;\<Gamma> \<turnstile> M : BoxTy N \<sigma> ; M'; \<Delta> \<turnstile> \<theta> : \<tau> ; \<theta>'\<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> TIInsp M \<theta> : \<tau> ; IInsp M' \<theta>'"
(* wf_repl *)
| trnil_ty : "\<Delta>;[] \<turnstile> M : \<tau> ; M' \<Longrightarrow> \<Delta> \<turnstile> TrNil M : \<tau> ; CrNil M'" (* M is the default branch *)
| trcons_ty : "\<lbrakk>\<Delta> \<turnstile> \<theta> : \<tau> ; \<theta>'; \<Delta>;[] \<turnstile> M : bty \<tau> l ; M'\<rbrakk> \<Longrightarrow> \<Delta> \<turnstile> TrCons l M \<theta> : \<tau> ; CrCons l M' \<theta>'"

equivariance wfc

nominal_inductive wfc
  by (auto simp:abs_fresh uvar_usubst_fresh)

equivariance wfq

nominal_inductive wfq
  apply(auto intro: fresh_ty var_fresh_uctx uvar_fresh_ctx)
  apply(simp_all add:abs_fresh var_subst_fresh uvar_usubst_fresh)
done

equivariance wf

nominal_inductive wf
  apply(auto intro: fresh_ty var_fresh_uctx uvar_fresh_ctx)
  apply(simp_all add:abs_fresh usubst_fresh)
done

(* The box-free contexts "\<F>" (terms with holes) from Sect. 2.3 *)
nominal_datatype punch_trm =
  PTHole
| PTLam "ty" "var" "punch_trm"
| PTAppL "punch_trm" "trm" | PTAppR "trm" "punch_trm"
| PTLetL "ty" "punch_trm" "uvar" "trm" | PTLetR "ty" "trm" "uvar" "punch_trm"
| PTOInsp "punch_trm_repl"
| PTIInspL "punch_trm" "trm_repl" | PTIInspR "trm" "punch_trm_repl"
and punch_trm_repl = 
  PTrNil "punch_trm"
| PTrConsL "trail_label" "punch_trm" "trm_repl" | PTrConsR "trail_label" "trm" "punch_trm_repl"

nominal_primrec pt_apply :: "punch_trm \<Rightarrow> trm \<Rightarrow> trm" ("_\<lceil>_\<rceil>" [100,60] 60) 
and ptr_apply :: "punch_trm_repl \<Rightarrow> trm \<Rightarrow> trm_repl" ("_\<lceil>_\<rceil>" [100,60] 60) where
  "PTHole\<lceil>M\<rceil> = M"
| "(PTLam \<sigma> a F)\<lceil>M\<rceil> = TLam \<sigma> a (F\<lceil>M\<rceil>)"
| "(PTAppL F N)\<lceil>M\<rceil> = TApp (F\<lceil>M\<rceil>) N"
| "(PTAppR N F)\<lceil>M\<rceil> = TApp N (F\<lceil>M\<rceil>)"
| "(PTLetL \<sigma> F u N)\<lceil>M\<rceil> = TLetBox \<sigma> (F\<lceil>M\<rceil>) u N"
| "(PTLetR \<sigma> N u F)\<lceil>M\<rceil> = TLetBox \<sigma> N u (F\<lceil>M\<rceil>)"
| "(PTOInsp Fs)\<lceil>M\<rceil> = TOInsp (Fs\<lceil>M\<rceil>)"
| "(PTIInspL F \<theta>)\<lceil>M\<rceil> = TIInsp (F\<lceil>M\<rceil>) \<theta>"
| "(PTIInspR N Fs)\<lceil>M\<rceil> = TIInsp N (Fs\<lceil>M\<rceil>)"
| "(PTrNil F)\<lceil>M\<rceil> = TrNil (F\<lceil>M\<rceil>)"
| "(PTrConsL l F \<theta>)\<lceil>M\<rceil> = TrCons l (F\<lceil>M\<rceil>) \<theta>"
| "(PTrConsR l N Fs)\<lceil>M\<rceil> = TrCons l N (Fs\<lceil>M\<rceil>)"
apply (rule TrueI)+
done

(* computes a reflexivity trail replacement from a given term replacement *)
nominal_primrec refl_repl :: "trm_repl \<Rightarrow> trail_repl"
and refl_repl_aux :: "trm \<Rightarrow> unit" where
  "refl_repl (TrNil M) = QrNil (Refl (t2c M))"
| "refl_repl (TrCons l M \<theta>) = QrCons l (Refl (t2c M)) (refl_repl \<theta>)"
| "refl_repl_aux (TVar _) = ()"
| "refl_repl_aux (TUVar _) = ()"
| "a \<sharp> \<tau> \<Longrightarrow>refl_repl_aux (TLam \<tau> a  _) = ()"
| "refl_repl_aux (TApp _ _) = ()"
| "u \<sharp> (\<tau>,M) \<Longrightarrow> refl_repl_aux (TLetBox \<tau> M u _) = ()"
| "refl_repl_aux (TBox _ _) = ()"
| "refl_repl_aux (TOInsp _) = ()"
| "refl_repl_aux (TIInsp _ _) = ()"
apply (finite_guess)+
apply (rule TrueI)+
apply (fresh_guess)+
done

lemma refl_repl_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  and \<theta>::trm_repl
  shows "pi\<bullet>(refl_repl \<theta>) = refl_repl (pi\<bullet>\<theta>)"
  and "pi'\<bullet>(refl_repl \<theta>) = refl_repl (pi'\<bullet>\<theta>)"
by (nominal_induct \<theta> rule:trm_trm_repl.strong_inducts(2)[of "\<lambda>_ _. True"],auto simp:t2c_eqvt)

lemma refl_repl_fresh:
  fixes \<theta>::trm_repl
  and a::var
  and u::uvar
  shows "a \<sharp> \<theta> \<Longrightarrow> a \<sharp> refl_repl \<theta>"
  and "u \<sharp> \<theta> \<Longrightarrow> u \<sharp> refl_repl \<theta>"
by (nominal_induct \<theta> rule:trm_trm_repl.strong_inducts(2)[of "\<lambda>_ _. True"],auto simp:t2c_fresh)

(* Rather than computing a canonical trail context \<Q>\<^sub>\<F> (Def.6) from \<F> and then applying it to a 
   trail, the following pt_trail_apply performs both operations at once *)
nominal_primrec pt_trail_apply :: "punch_trm \<Rightarrow> trail \<Rightarrow> trail" ("_\<lceil>_\<rceil>" [100,60] 60)
and ptr_trail_apply :: "punch_trm_repl \<Rightarrow> trail \<Rightarrow> trail_repl" ("_\<lceil>_\<rceil>" [100,60] 60) where
  "pt_trail_apply PTHole q = q"
| "(PTLam \<sigma> a F)\<lceil>q\<rceil> = QLam \<sigma> a (F\<lceil>q\<rceil>)"
| "(PTAppL F N)\<lceil>q\<rceil> = QApp (F\<lceil>q\<rceil>) (Refl (t2c N))"
| "(PTAppR N F)\<lceil>q\<rceil> = QApp (Refl (t2c N)) (F\<lceil>q\<rceil>)"
| "(PTLetL \<sigma> F u N)\<lceil>q\<rceil> = QLet \<sigma> (F\<lceil>q\<rceil>) u (Refl (t2c N))"
| "(PTLetR \<sigma> N u F)\<lceil>q\<rceil> = QLet \<sigma> (Refl (t2c N)) u (F\<lceil>q\<rceil>)"
| "(PTOInsp Fs)\<lceil>q\<rceil> = OTB (Fs\<lceil>q\<rceil>)"
| "(PTIInspL F \<theta>)\<lceil>q\<rceil> = ITB (F\<lceil>q\<rceil>) (refl_repl \<theta>)"
| "(PTIInspR N Fs)\<lceil>q\<rceil> = ITB (Refl (t2c N)) (Fs\<lceil>q\<rceil>)"
| "(PTrNil F)\<lceil>q\<rceil> = QrNil (F\<lceil>q\<rceil>)"
| "(PTrConsL l F \<theta>)\<lceil>q\<rceil> = QrCons l (F\<lceil>q\<rceil>) (refl_repl \<theta>)"
| "(PTrConsR l N Fs)\<lceil>q\<rceil> = QrCons l (Refl (t2c N)) (Fs\<lceil>q\<rceil>)"
apply (rule TrueI)+
done

lemma ptrm_apply_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  and F::punch_trm
  and Fs::punch_trm_repl
  and M::trm
  shows "pi\<bullet>(F\<lceil>M\<rceil>) = (pi\<bullet>F)\<lceil>pi\<bullet>M\<rceil>"
  and "pi'\<bullet>(F\<lceil>M\<rceil>) = (pi'\<bullet>F)\<lceil>pi'\<bullet>M\<rceil>"
  and "pi\<bullet>(Fs\<lceil>M\<rceil>) = (pi\<bullet>Fs)\<lceil>pi\<bullet>M\<rceil>"
  and "pi'\<bullet>(Fs\<lceil>M\<rceil>) = (pi'\<bullet>Fs)\<lceil>pi'\<bullet>M\<rceil>"
  by (nominal_induct F and Fs rule:punch_trm_punch_trm_repl.strong_inducts,auto)

lemma pt_trail_apply_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  and F::punch_trm
  and Fs::punch_trm_repl
  and q::trail
  shows "pi\<bullet>(F\<lceil>q\<rceil>) = (pi\<bullet>F)\<lceil>pi\<bullet>q\<rceil>"
  and "pi'\<bullet>(F\<lceil>q\<rceil>) = (pi'\<bullet>F)\<lceil>pi'\<bullet>q\<rceil>"
  and "pi\<bullet>(Fs\<lceil>q\<rceil>) = (pi\<bullet>Fs)\<lceil>pi\<bullet>q\<rceil>"
  and "pi'\<bullet>(Fs\<lceil>q\<rceil>) = (pi'\<bullet>Fs)\<lceil>pi'\<bullet>q\<rceil>"
by (nominal_induct F and Fs rule:punch_trm_punch_trm_repl.strong_inducts,auto simp:t2c_eqvt refl_repl_eqvt)

lemma ptrm_apply_fresh[simp]:
  fixes F::punch_trm
  and Fs::punch_trm_repl
  and M::trm
  and a::var
  and u::uvar
  shows "a \<sharp> (F,M) \<Longrightarrow> a \<sharp> (F\<lceil>M\<rceil>)"
  and "u \<sharp> (F,M) \<Longrightarrow> u \<sharp> (F\<lceil>M\<rceil>)"
  and "a \<sharp> (Fs,M) \<Longrightarrow> a \<sharp> (Fs\<lceil>M\<rceil>)"
  and "u \<sharp> (Fs,M) \<Longrightarrow> u \<sharp> (Fs\<lceil>M\<rceil>)"
  by (nominal_induct F and Fs rule:punch_trm_punch_trm_repl.strong_inducts, 
      simp_all add:abs_fresh fresh_prod)

lemma pt_trail_apply_fresh[simp]:
  fixes F::punch_trm
  and Fs::punch_trm_repl
  and q::trail
  and a::var
  and u::uvar
  shows "a \<sharp> (F,q) \<Longrightarrow> a \<sharp> (F\<lceil>q\<rceil>)"
  and "u \<sharp> (F,q) \<Longrightarrow> u \<sharp> (F\<lceil>q\<rceil>)"
  and "a \<sharp> (Fs,q) \<Longrightarrow> a \<sharp> (Fs\<lceil>q\<rceil>)"
  and "u \<sharp> (Fs,q) \<Longrightarrow> u \<sharp> (Fs\<lceil>q\<rceil>)"
  apply (nominal_induct F and Fs avoiding:q rule:punch_trm_punch_trm_repl.strong_inducts)
  apply (simp_all add:abs_fresh fresh_prod t2c_fresh refl_repl_fresh)
done

lemma tsubst_eqvt[eqvt]:
  fixes \<pi>::"var prm"
  fixes \<pi>'::"uvar prm"
  fixes M::trm
  fixes Ms :: trm_repl
  fixes x::var
  shows "\<pi> \<bullet> (M[x\<mapsto>N]) = (\<pi>\<bullet>M)[\<pi>\<bullet>x\<mapsto>\<pi>\<bullet>N]"
  and "\<pi>' \<bullet> (M[x\<mapsto>N]) = (\<pi>'\<bullet>M)[\<pi>'\<bullet>x\<mapsto>\<pi>'\<bullet>N]"
  and "\<pi> \<bullet> (Ms[x\<mapsto>N]) = (\<pi>\<bullet>Ms)[\<pi>\<bullet>x\<mapsto>\<pi>\<bullet>N]"
  and "\<pi>' \<bullet> (Ms[x\<mapsto>N]) = (\<pi>'\<bullet>Ms)[\<pi>'\<bullet>x\<mapsto>\<pi>'\<bullet>N]"
apply (nominal_induct M and Ms avoiding:x N rule:trm_trm_repl.strong_inducts)
apply (auto simp:perm_bij fresh_bij subst_eqvt qsubst_eqvt t2c_eqvt)
done

lemma tsubst_fresh_t[simp]: 
  fixes M::trm
  fixes Ms::trm_repl
  fixes x::var
  shows "x \<sharp> [y].M \<Longrightarrow> x \<sharp> N \<Longrightarrow> x \<sharp> (M[y\<mapsto>N])"
  and "x \<sharp> [y].Ms \<Longrightarrow> x \<sharp> N \<Longrightarrow> x \<sharp> (Ms[y\<mapsto>N])"
  apply(nominal_induct M and Ms avoiding: y N rule:trm_trm_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_trail_label subst_fresh)
done

lemma tsubst_fresh_u[simp]: 
  fixes M::trm
  fixes Ms::trm_repl
  fixes x::uvar
  shows "x \<sharp> [y].M \<Longrightarrow> x \<sharp> N \<Longrightarrow> x \<sharp> (M[y\<mapsto>N])"
  and "x \<sharp> [y].Ms \<Longrightarrow> x \<sharp> N \<Longrightarrow> x \<sharp> (Ms[y\<mapsto>N])"
  apply(nominal_induct M and Ms avoiding: y N rule:trm_trm_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_trail_label subst_fresh)
done

lemmas tsubst_fresh = tsubst_fresh_t tsubst_fresh_u

lemma tlusubst_eqvt[eqvt]:
  fixes \<pi>::"var prm"
  fixes \<pi>'::"uvar prm"
  fixes M::trm
  fixes Ms :: trm_repl
  fixes x::uvar
  shows "\<pi> \<bullet> (M{x\<mapsto>N,q,N'}) = (\<pi>\<bullet>M){\<pi>\<bullet>x\<mapsto>\<pi>\<bullet>N,\<pi>\<bullet>q,\<pi>\<bullet>N'}"
  and "\<pi>' \<bullet> (M{x\<mapsto>N,q,N'}) = (\<pi>'\<bullet>M){\<pi>'\<bullet>x\<mapsto>\<pi>'\<bullet>N,\<pi>'\<bullet>q,\<pi>'\<bullet>N'}"
  and "\<pi> \<bullet> (Ms{x\<mapsto>N,q,N'}) = (\<pi>\<bullet>Ms){\<pi>\<bullet>x\<mapsto>\<pi>\<bullet>N,\<pi>\<bullet>q,\<pi>\<bullet>N'}"
  and "\<pi>' \<bullet> (Ms{x\<mapsto>N,q,N'}) = (\<pi>'\<bullet>Ms){\<pi>'\<bullet>x\<mapsto>\<pi>'\<bullet>N,\<pi>'\<bullet>q,\<pi>'\<bullet>N'}"
apply (nominal_induct M and Ms avoiding: x N q N' rule:trm_trm_repl.strong_inducts)
apply (auto simp:tusubst_eqvt qusubst_eqvt subst_eqvt fresh_bij perm_bij)
done

lemma tlusubst_fresh_t[simp]: 
  fixes M::trm
  fixes Ms::trm_repl
  fixes x::var
  shows "x \<sharp> [y].M \<Longrightarrow> x \<sharp> (N,q,N') \<Longrightarrow> x \<sharp> (M{y\<mapsto>N,q,N'})"
  and "x \<sharp> [y].Ms \<Longrightarrow> x \<sharp> (N,q,N') \<Longrightarrow> x \<sharp> (Ms{y\<mapsto>N,q,N'})"
  apply(nominal_induct M and Ms avoiding: y N q N' rule:trm_trm_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_trail_label fresh_prod usubst_fresh)
done

lemma tlusubst_fresh_u[simp]: 
  fixes M::trm
  fixes Ms::trm_repl
  fixes x::uvar
  shows "x \<sharp> [y].M \<Longrightarrow> x \<sharp> (N,q,N') \<Longrightarrow> x \<sharp> (M{y\<mapsto>N,q,N'})"
  and "x \<sharp> [y].Ms \<Longrightarrow> x \<sharp> (N,q,N') \<Longrightarrow> x \<sharp> (Ms{y\<mapsto>N,q,N'})"
  apply(nominal_induct M and Ms avoiding: y N q N' rule:trm_trm_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_trail_label fresh_prod usubst_fresh)
done

lemmas tlusubst_fresh = tlusubst_fresh_t tlusubst_fresh_u

(* Term reduction (Fig.3) *) 
inductive red :: "trm \<Rightarrow> trm \<Rightarrow> bool" ("_ \<leadsto> _" [60,60] 60)
where
  red_beta: "x \<sharp> (q,F,N,\<tau>) \<Longrightarrow> TBox q (F\<lceil>TApp (TLam \<tau> x M) N\<rceil>) \<leadsto> 
             TBox (Trans q (F\<lceil>Beta \<tau> x (t2c M) (t2c N)\<rceil>)) (F\<lceil>M[x\<mapsto>N]\<rceil>)"
| red_beta_box: "u \<sharp> (F,M,\<tau>,q1,q2) \<Longrightarrow> TBox q1 (F\<lceil>TLetBox \<tau> (TBox q2 M) u N\<rceil>) \<leadsto> 
    TBox (Trans q1 (F\<lceil>Trans (BetaBox \<tau> (src q2) u (t2c N)) (N\<lbrace>u\<mapsto>M,q2,src q2\<rbrace>)\<rceil>)) (F\<lceil>N{u\<mapsto>M,q2,src q2}\<rceil>)"
| red_oinsp: "TBox q (F\<lceil>TOInsp \<theta>\<rceil>) \<leadsto> TBox (Trans q (F\<lceil>OTI q (t2c_repl \<theta>)\<rceil>)) (F\<lceil>stfold (skel_of_trail q) \<theta>\<rceil>)"
| red_iinsp: "TBox q1 (F\<lceil>TIInsp (TBox q2 M) \<theta>\<rceil>) \<leadsto> TBox (Trans q1 (F\<lceil>ITI q2 (t2c_repl \<theta>)\<rceil>)) (F\<lceil>stfold (skel_of_trail q2) \<theta>\<rceil>)"
| red_box: "M \<leadsto> N \<Longrightarrow> TBox q (F\<lceil>M\<rceil>) \<leadsto> TBox q (F\<lceil>N\<rceil>)"

equivariance red
nominal_inductive red
  apply(auto intro: fresh_ty var_fresh_uctx uvar_fresh_ctx)
  apply(auto simp:fresh_prod abs_fresh t2c_fresh subst_fresh usubst_fresh src_fresh tsubst_fresh)
  done

end