theory Common
imports "~~/src/HOL/Nominal/Nominal"
(*imports "Nominal/Nominal2"*)

begin

(* Common definitions for the two calculi \<lambda>hc and \<lambda>hs *)

(* Two sorts of variables for modal type systems *)

atom_decl var uvar

(* The types of the object language *)
nominal_datatype ty = PTy
                    | FunTy "ty" "ty"
                    | BoxTy "ty"

lemma fresh_ty:
  fixes x::var 
  fixes u::uvar 
  fixes \<tau> :: ty
  shows "u \<sharp> \<tau>"
  and "x \<sharp> \<tau>"
  apply(nominal_induct \<tau> rule: ty.strong_induct)
  by(auto)

lemma ty_eqvt:
  fixes \<tau>::ty
  fixes \<pi>::"var prm"
  fixes \<pi>'::"uvar prm"
  shows "\<pi>\<bullet>\<tau> = \<tau>"
  and  "\<pi>'\<bullet>\<tau> = \<tau>"
  apply(nominal_induct \<tau> rule:ty.strong_induct)
  apply(auto) 
  done

(* trail inspection labels *)
nominal_datatype trail_label = LRefl | LTrans | LBeta | LBetaBox | LITI | LOTI | LQLam | LQApp | LQLet 
                             | LITBn | LITBc | LOTBn | LOTBc

lemma fresh_trail_label :
  fixes x::var
  fixes y::uvar
  fixes l::trail_label
  shows "x \<sharp> l" and "y \<sharp> l"
  by (nominal_induct l rule:trail_label.strong_induct,auto)

lemma swap_trail_label :
  fixes x y::var
  fixes u v::uvar
  fixes l::trail_label
  shows "[(x,y)]\<bullet>l = l" and "[(u,v)]\<bullet>l = l"
  by (nominal_induct l rule:trail_label.strong_induct,auto)

(* The codes of \<lambda>hc *)
nominal_datatype code = Var "var"
                      | UVar "uvar"
                      | Lam "ty" "\<guillemotleft>var\<guillemotright>code"
                      | App "code" "code"
                      | Box "code"
                      | LetBox "ty" "code" "\<guillemotleft>uvar\<guillemotright>code"
                      | OInsp "code_repl"
                      | IInsp "code" "code_repl"
and code_repl = CrNil "code"
              | CrCons "trail_label" "code" "code_repl"

(* \<lambda>hc trails *)
nominal_datatype trail = Refl "code"
                       | Trans "trail" "trail"
                       | Beta "ty" "\<guillemotleft>var\<guillemotright>code" "code"
                       | BetaBox "ty" "code" "\<guillemotleft>uvar\<guillemotright>code"
                       | ITI "trail" "code_repl"
                       | OTI "trail" "code_repl"
                       | QLam "ty" "\<guillemotleft>var\<guillemotright>trail"
                       | QApp "trail" "trail"
                       | QLet "ty" "trail" "\<guillemotleft>uvar\<guillemotright>trail"
                       | ITB "trail" "trail_repl"
                       | OTB "trail_repl"
and trail_repl = QrNil "trail" (* represents the default branch *)
               | QrCons "trail_label" "trail" "trail_repl"

(* branch types, as defined in Sect. 2.2 *)
primrec bt :: "ty \<Rightarrow> nat \<Rightarrow> ty" where
  "bt \<sigma> 0 = \<sigma>"
| "bt \<sigma> (Suc n)= FunTy \<sigma> (bt \<sigma> n)"

nominal_primrec trail_label_arity :: "trail_label \<Rightarrow> nat" where
  "trail_label_arity LRefl = 0"
| "trail_label_arity LTrans = 2"
| "trail_label_arity LBeta = 0"
| "trail_label_arity LBetaBox = 0"
| "trail_label_arity LITI = 0"
| "trail_label_arity LOTI = 0"
| "trail_label_arity LQLam = 1"
| "trail_label_arity LQApp = 2"
| "trail_label_arity LQLet = 2"
| "trail_label_arity LITBn = 0"
| "trail_label_arity LITBc = 2"
| "trail_label_arity LOTBn = 0"
| "trail_label_arity LOTBc = 2"
by auto

definition bty :: "ty \<Rightarrow> trail_label \<Rightarrow> ty" where
  "bty \<sigma> l = bt \<sigma> (trail_label_arity l)"

(* Contexts *)

type_synonym  ctx = "(var * ty) list"

lemma fresh_var_not_in:
  fixes x::var
  and \<Gamma>::ctx
  shows "(x,\<tau>) \<in> set \<Gamma> \<Longrightarrow> x \<sharp> \<Gamma> \<Longrightarrow>  False"
  apply(induct \<Gamma>)
  by(auto simp add:fresh_list_cons fresh_prod fresh_atm)

lemma uvar_fresh_ctx[simp]:
  fixes u::uvar 
  fixes \<Gamma> :: ctx
  shows "u \<sharp> \<Gamma>"
  apply(induct \<Gamma>)
  apply(auto simp add:fresh_list_nil fresh_list_cons fresh_prod fresh_atm fresh_ty)
  done

inductive valid :: "ctx \<Rightarrow> bool" 
where
  v1 : "valid []"
| v2 : "\<lbrakk>valid \<Gamma>; x \<sharp> \<Gamma> \<rbrakk> \<Longrightarrow> valid ((x,\<tau>)#\<Gamma>)"

equivariance valid
nominal_inductive valid
  done

type_synonym  uctx = "(uvar * ty) list"

lemma fresh_uvar_not_in:
  fixes u::uvar
  and \<Delta>::uctx
  shows "(u,\<tau>) \<in> set \<Delta> \<Longrightarrow> u \<sharp> \<Delta> \<Longrightarrow>  False"
  apply(induct \<Delta>)
  by(auto simp add:fresh_list_cons fresh_prod fresh_atm)  

lemma var_fresh_uctx[simp]: 
  fixes x::var
  fixes \<Delta>::uctx
  shows "x \<sharp> \<Delta>"
  apply(induct \<Delta>)
  apply(auto simp add:fresh_list_nil fresh_list_cons fresh_prod fresh_atm fresh_ty)
  done

inductive uvalid :: "uctx \<Rightarrow> bool" 
where
  v1 : "uvalid []"
| v2 : "\<lbrakk>uvalid \<Delta>; u \<sharp> \<Delta> \<rbrakk> \<Longrightarrow> uvalid ((u,\<tau>)#\<Delta>)"

equivariance uvalid
nominal_inductive uvalid
  done

(* a large part of the structure of trails is actually ignored when performing trail inspection:
   to make the formalization simpler, we use a stripped-down version of trails called 
   "trail skeleton", containing only the information needed for trail inspection

   skel_of_trail computes the skeleton of a given trail *)
nominal_datatype trailskel = SRefl
                           | STrans "trailskel" "trailskel"
                           | SBeta
                           | SBetaBox
                           | SITI
                           | SOTI
                           | SLam "trailskel"
                           | SApp "trailskel" "trailskel"
                           | SLet "trailskel" "trailskel"
                           | SITB "trailskel" "trailskel_repl"
                           | SOTB "trailskel_repl"
and trailskel_repl = SrNil "trailskel" (* represents the default branch *)
               | SrCons "trail_label" "trailskel" "trailskel_repl"

lemma fresh_trailskel[simp] :
  fixes x::var
  fixes y::uvar
  fixes s::trailskel
  fixes ss::trailskel_repl
  shows "x \<sharp> s" and "y \<sharp> s" and "x \<sharp> ss" and "y \<sharp> ss"
  by (nominal_induct s and ss rule:trailskel_trailskel_repl.strong_inducts,auto simp:fresh_trail_label)

lemma swap_trailskel[simp] :
  fixes x y::var
  fixes u v::uvar
  fixes s::trailskel
  fixes ss::trailskel_repl
  shows "[(x,y)]\<bullet>s = s" and "[(u,v)]\<bullet>s = s"
  and "[(x,y)]\<bullet>ss = ss" and "[(u,v)]\<bullet>ss = ss"
  by (nominal_induct s and ss rule:trailskel_trailskel_repl.strong_inducts,auto simp:swap_trail_label)

nominal_primrec skel_of_trail :: "trail \<Rightarrow> trailskel"
  and skel_of_trail_repl :: "trail_repl \<Rightarrow> trailskel_repl"
where
  "skel_of_trail (Refl _) = SRefl"
| "skel_of_trail (Trans q1 q2) = STrans (skel_of_trail q1) (skel_of_trail q2)"
| "a \<sharp> N \<Longrightarrow> skel_of_trail (Beta _ a M N) = SBeta"
| "u \<sharp> M \<Longrightarrow> skel_of_trail (BetaBox _ M u N) = SBetaBox"
| "skel_of_trail (ITI _ _) = SITI"
| "skel_of_trail (OTI _ _) = SOTI"
| "skel_of_trail (QLam _ _ q) = SLam (skel_of_trail q)"
| "skel_of_trail (QApp q1 q2) = SApp (skel_of_trail q1) (skel_of_trail q2)"
| "u \<sharp> q1 \<Longrightarrow> skel_of_trail (QLet _ q1 u q2) = SLet (skel_of_trail q1) (skel_of_trail q2)"
| "skel_of_trail (ITB q qs) = SITB (skel_of_trail q) (skel_of_trail_repl qs)"
| "skel_of_trail (OTB qs) = SOTB (skel_of_trail_repl qs)"
| "skel_of_trail_repl (QrNil q) = SrNil (skel_of_trail q)"
| "skel_of_trail_repl (QrCons l q qs) = SrCons l (skel_of_trail q) (skel_of_trail_repl qs)"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto simp:fresh_trailskel)
  apply(fresh_guess add:fresh_trail_label fresh_ty)+
  done

lemma skel_of_trail_eqvt[eqvt] :
  fixes \<pi> :: "var prm"
  fixes \<pi>':: "uvar prm"
  shows "\<pi> \<bullet> (skel_of_trail q) = skel_of_trail (\<pi> \<bullet> q)"
  and   "\<pi>' \<bullet> (skel_of_trail q) = skel_of_trail (\<pi>' \<bullet> q)"
  and "\<pi> \<bullet> (skel_of_trail_repl qs) = skel_of_trail_repl (\<pi> \<bullet> qs)"
  and   "\<pi>' \<bullet> (skel_of_trail_repl qs) = skel_of_trail_repl (\<pi>' \<bullet> qs)"
apply (nominal_induct q and qs rule:trail_trail_repl.strong_inducts)
apply (auto simp add:perm_bij fresh_bij)
done

end
