theory Modal
imports "~~/src/HOL/Nominal/Nominal" "Common"

begin

(* Definitions and proofs on the \<lambda>hs calculus, Sections 3 and 4 *)

nominal_datatype exp = Var "var"
                     | UVar "uvar"
                     | Lam "ty" "\<guillemotleft>var\<guillemotright>exp"
                     | App "exp" "exp"
                     | Box "exp"
                     | LetBox "ty" "exp" "\<guillemotleft>uvar\<guillemotright>exp"
                     | Insp "exp_repl"
                     | IInsp "exp" "exp_repl"
and exp_repl = ErNil "exp"
             | ErCons "trail_label" "exp" "exp_repl"

fun mker :: "(trail_label \<times> exp) list \<Rightarrow> exp \<Rightarrow> exp_repl" where
 "mker [] bdef = ErNil bdef"
|"mker ((x,M)#l) bdef = ErCons x M (mker l bdef)"

fun mktr :: "(trail_label \<times> trail) list \<Rightarrow> trail \<Rightarrow> trail_repl" where
 "mktr [] bdef = QrNil bdef"
|"mktr ((x,q)#l) bdef = QrCons x q (mktr l bdef)"

lemma bty_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  shows "pi\<bullet>(bty \<sigma> l) = bty (pi\<bullet>\<sigma>) (pi\<bullet>l)"
  and "pi'\<bullet>(bty \<sigma> l) = bty (pi'\<bullet>\<sigma>) (pi'\<bullet>l)"
apply(auto simp:ty_eqvt)
apply(nominal_induct l rule:trail_label.strong_induct,auto)
apply(nominal_induct l rule:trail_label.strong_induct,auto)
done

nominal_primrec
  bedom :: "exp_repl \<Rightarrow> trail_label list"
  and bedaux :: "exp \<Rightarrow> unit"
where
  "bedom (ErNil Mdef) = []"
| "bedom (ErCons x M tr) = x#bedom tr"
| "bedaux (Var _) = ()"
| "bedaux (UVar _) = ()"
| "bedaux (Lam \<tau> a M) = ()"
| "u \<sharp> M \<Longrightarrow> bedaux (LetBox \<tau> M u N) = ()"
| "bedaux (App _ _) = ()"
| "bedaux (Box _) = ()"
| "bedaux (Insp _) = ()"
| "bedaux (IInsp _ _) = ()"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto)
  apply(fresh_guess add:fresh_ty)+
  done

primrec
  lambda_lift :: "ty \<Rightarrow> nat \<Rightarrow> exp \<Rightarrow> exp"
where 
  "lambda_lift _ 0 M = M"
| "lambda_lift \<tau> (Suc n) M = fresh_fun (\<lambda>a. Lam \<tau> a (lambda_lift \<tau> n M))"

nominal_primrec
  bedefault :: "exp_repl \<Rightarrow> exp"
  and bedefaux :: "exp \<Rightarrow> unit"
where
  "bedefault (ErNil M) = M"
| "bedefault (ErCons _ _ Ms) = bedefault Ms"
| "bedefaux (Var _) = ()"
| "bedefaux (UVar _) = ()"
| "bedefaux (Lam \<tau> a M)= ()"
| "u \<sharp> M \<Longrightarrow> bedefaux (LetBox \<tau> M u N) = ()"
| "bedefaux (App _ _) = ()"
| "bedefaux (Box _) = ()"
| "bedefaux (Insp _) = ()"
| "bedefaux (IInsp _ _) = ()"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto)
  apply(fresh_guess add:fresh_ty)+
  done

lemma bedefault_eqvt[eqvt] :
  fixes \<pi>::"var prm"
  fixes \<pi>'::"uvar prm"
  shows "\<pi> \<bullet> bedefault \<theta> = bedefault (\<pi> \<bullet> \<theta>)"
  and "\<pi>' \<bullet> bedefault \<theta> = bedefault (\<pi>' \<bullet> \<theta>)"
  apply (nominal_induct \<theta> rule:exp_exp_repl.strong_inducts(2))
  apply (simp) 
  apply (auto)
done

lemma bedefault_fresh :
  fixes a::var
  fixes u::uvar
  shows "a \<sharp> \<theta> \<Longrightarrow> a \<sharp> bedefault \<theta>"
  and "u \<sharp> \<theta> \<Longrightarrow> u \<sharp> bedefault \<theta>"
by (nominal_induct \<theta> rule:exp_exp_repl.strong_inducts(2),auto)

nominal_primrec
  belookup :: "exp_repl \<Rightarrow> trail_label \<Rightarrow> exp option"
  and belaux :: "exp \<Rightarrow> trail_label \<Rightarrow> unit"
where
  "belookup (ErNil _) y = None"
| "belookup (ErCons x M tr) y = (if x=y then Some M else belookup tr y)"
| "belaux (Var _) y = ()"
| "belaux (UVar _) y = ()"
| "belaux (Lam \<tau> a M) y = ()"
| "u \<sharp> M \<Longrightarrow> belaux (LetBox \<tau> M u N) y = ()"
| "belaux (App _ _) y = ()"
| "belaux (Box _) y = ()"
| "belaux (Insp _) y = ()"
| "belaux (IInsp _ _) y = ()"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto)
  apply(fresh_guess add:fresh_trail_label fresh_ty)+
  done

lemma belookup_eqvt[eqvt] :
  fixes \<pi>::"var prm"
  fixes \<pi>'::"uvar prm"
  shows "\<pi> \<bullet> belookup \<theta> l = belookup (\<pi> \<bullet> \<theta>) (\<pi> \<bullet> l)"
  and "\<pi>' \<bullet> belookup \<theta> l = belookup (\<pi>' \<bullet> \<theta>) (\<pi>' \<bullet> l)"
  apply (nominal_induct \<theta> rule:exp_exp_repl.strong_inducts(2))
  apply (simp) 
  apply (auto)
  apply (subgoal_tac "trail_label = l")
  apply (simp)
  apply (simp only:perm_bij)
  apply (subgoal_tac "trail_label = l")
  apply (simp)
  apply (simp only:perm_bij)
done

lemma belookup_fresh[simp] :
  fixes a::var
  fixes u::uvar
  shows "a \<sharp> \<theta> \<Longrightarrow> a \<sharp> belookup \<theta> l"
  and "u \<sharp> \<theta> \<Longrightarrow> u \<sharp> belookup \<theta> l"
by (nominal_induct \<theta> rule:exp_exp_repl.strong_inducts(2)[of "\<lambda>_ _. True"],auto simp:fresh_some fresh_none)

nominal_primrec
  bqdefault :: "trail_repl \<Rightarrow> trail"
  and bqdefaux :: "trail \<Rightarrow> unit"
where
  "bqdefault (QrNil q) = q"
| "bqdefault (QrCons _ _ qs) = bqdefault qs"
| "bqdefaux (Refl _) = ()"
| "bqdefaux (Trans _ _) = ()"
| "a \<sharp> N \<Longrightarrow> bqdefaux (Beta \<sigma> a M N) = ()"
| "u \<sharp> M \<Longrightarrow> bqdefaux (BetaBox \<sigma> M u N) = ()"
| "bqdefaux (ITI _ _) = ()"
| "bqdefaux (OTI _ _) = ()"
| "bqdefaux (QLam _ _ _) = ()"
| "bqdefaux (QApp _ _) = ()"
| "u \<sharp> q1 \<Longrightarrow> bqdefaux (QLet \<sigma> q1 u q2) = ()"
| "bqdefaux (ITB _ _) = ()"
| "bqdefaux (OTB _) = ()"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto)
  apply(fresh_guess add:fresh_trail_label fresh_ty)+
  done

nominal_primrec
  bqlookup :: "trail_repl \<Rightarrow> trail_label \<Rightarrow> trail option"
  and bqlaux :: "trail \<Rightarrow> trail_label \<Rightarrow> unit"
where
  "bqlookup (QrNil _) y = None"
| "bqlookup (QrCons x q tr) y = (if x=y then Some q else bqlookup tr y)"
| "bqlaux (Refl _) y = ()"
| "bqlaux (Trans _ _) y = ()"
| "a \<sharp> N \<Longrightarrow> bqlaux (Beta _ a M N) y = ()"
| "u \<sharp> M \<Longrightarrow> bqlaux (BetaBox _ M u N) y = ()"
| "bqlaux (ITI _ _) y = ()"
| "bqlaux (OTI _ _) y = ()"
| "bqlaux (QLam _ _ _) y = ()"
| "bqlaux (QApp _ _) y = ()"
| "u \<sharp> q1 \<Longrightarrow> bqlaux (QLet _ q1 u q2) y = ()"
| "bqlaux (ITB _ _) y = ()"
| "bqlaux (OTB _) y = ()"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto)
  apply(fresh_guess add:fresh_trail_label fresh_ty)+
  done

nominal_primrec
  bqdom :: "trail_repl \<Rightarrow> trail_label list"
  and bqdaux :: "trail \<Rightarrow> unit"
where
  "bqdom (QrNil qdef) = []"
| "bqdom (QrCons x q tr) = x#bqdom tr"
| "bqdaux (Refl _) = ()"
| "bqdaux (Trans _ _) = ()"
| "a \<sharp> N \<Longrightarrow> bqdaux (Beta _ a M N) = ()"
| "u \<sharp> M \<Longrightarrow> bqdaux (BetaBox _ M u N) = ()"
| "bqdaux (ITI _ _) = ()"
| "bqdaux (OTI _ _) = ()"
| "bqdaux (QLam _ _ _) = ()"
| "bqdaux (QApp _ _) = ()"
| "u \<sharp> q1 \<Longrightarrow> bqdaux (QLet _ q1 u q2) = ()"
| "bqdaux (ITB _ _) = ()"
| "bqdaux (OTB _) = ()"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(auto)
  apply(fresh_guess add:fresh_ty)+
  done

primrec mybind :: "exp option \<Rightarrow> (exp \<Rightarrow> exp) \<Rightarrow> exp \<Rightarrow> exp" where
  "mybind None f b = b"
| "mybind (Some a) f b = f a"

lemma mybind_eqvt[eqvt] :
  fixes \<pi> :: "var prm"
  fixes \<pi>':: "uvar prm"
  shows "\<pi> \<bullet> (mybind M f b) = mybind (\<pi> \<bullet> M) (\<pi> \<bullet> f) (\<pi> \<bullet> b)"
  and   "\<pi>' \<bullet> (mybind M f b) = mybind (\<pi>' \<bullet> M) (\<pi>' \<bullet> f) (\<pi>' \<bullet> b)"
  by (cases M,auto simp add:perm_fun_def perm_pi_simp)+

lemma mybind_fresh_None :
  fixes x::var
  fixes u::uvar
  shows "x \<sharp> b \<Longrightarrow> x \<sharp> mybind None f b"
  and "u \<sharp> b \<Longrightarrow> u \<sharp> mybind None f b"
  by auto

lemma mybind_fresh_Some :
  fixes x::var
  fixes u::uvar
  shows "x \<sharp> f a \<Longrightarrow> x \<sharp> mybind (Some a) f b"
  and "u \<sharp> f a \<Longrightarrow> u \<sharp> mybind (Some a) f b"
  by auto

lemma mybind_fresh2 :
  fixes x::var
  fixes u::uvar
  shows "(M = None \<Longrightarrow> x \<sharp> b) \<Longrightarrow> (\<And>a. M = Some a \<Longrightarrow> x \<sharp> f a) \<Longrightarrow> x \<sharp> mybind M f b"
  and "(M = None \<Longrightarrow> u \<sharp> b) \<Longrightarrow> (\<And>a. M = Some a \<Longrightarrow> u \<sharp> f a) \<Longrightarrow> u \<sharp> mybind M f b"
  by (cases M, auto simp: mybind_fresh_None mybind_fresh_Some)+

lemma mybind_fresh[simp] :
  fixes x::var
  fixes u::uvar
  shows "x \<sharp> mybind M f b \<Longrightarrow> (M = None \<and> x \<sharp> b) \<or> (\<exists>a. M = Some a \<and> x \<sharp> f a)"
  and "u \<sharp> mybind M f b \<Longrightarrow> (M = None \<and> u \<sharp> b) \<or> (\<exists>a. M = Some a \<and> u \<sharp> f a)"
  by (cases M,simp_all)+
  
(*
  Common.fin_supp(2): finite (supp ?x)
  Common.fs_uvar_class.axioms: finite (supp ?x)
  Common.fs_uvar_class.fs_uvar1: finite (supp ?x)

lemma mybind_finsupp[simp] :
  shows "finite (supp (mybind None f b)::var set)"
  and "finite (supp f::var set) \<Longrightarrow> finite (supp (\<lambda>a. mybind (Some a) f b) ::var set)"
  and "finite (supp (mybind None f b)::uvar set)"
  and "finite (supp f::uvar set) \<Longrightarrow> finite (supp (\<lambda>a. mybind (Some a) f b) ::uvar set)"
  by (auto simp add:exp.supp Common.fin_supp)
*)

definition qfbranch :: "exp_repl \<Rightarrow> trail_label \<Rightarrow> (exp \<Rightarrow> exp) \<Rightarrow> exp" where
  "qfbranch \<theta> l f = mybind (belookup \<theta> l) f (bedefault \<theta>)"

lemma qfbranch_eqvt[eqvt] :
  fixes \<pi> :: "var prm"
  fixes \<pi>':: "uvar prm"
  shows "\<pi> \<bullet> (qfbranch \<theta> l f) = qfbranch (\<pi> \<bullet> \<theta>) (\<pi> \<bullet> l) (\<pi> \<bullet> f)"
  and   "\<pi>' \<bullet> (qfbranch \<theta> l f) = qfbranch (\<pi>' \<bullet> \<theta>) (\<pi>' \<bullet> l) (\<pi>' \<bullet> f)"
by (auto simp add:qfbranch_def eqvts)

lemma qfbranch_fresh:
  fixes x::var
  fixes u::uvar
  shows "x \<sharp> bedefault \<theta> \<Longrightarrow> (\<And>l M. belookup \<theta> l = Some M \<Longrightarrow> x \<sharp> f M) \<Longrightarrow> x \<sharp> qfbranch \<theta> l f" 
  and "u \<sharp> bedefault \<theta> \<Longrightarrow> (\<And>l M. belookup \<theta> l = Some M \<Longrightarrow> u \<sharp> f M) \<Longrightarrow> u \<sharp> qfbranch \<theta> l f"
apply (simp_all add:qfbranch_def)
apply (cases "belookup \<theta> l",simp_all)+
done

lemma qfbranch_fresh2 :
  fixes x::var
  fixes u::uvar
  shows "x \<sharp> qfbranch \<theta> l f \<Longrightarrow> (belookup \<theta> l = None \<and> x \<sharp> bedefault \<theta>) \<or> (\<exists>M. belookup \<theta> l = Some M \<and> x \<sharp> f M)"
  and "u \<sharp> qfbranch \<theta> l f \<Longrightarrow> (belookup \<theta> l = None \<and> u \<sharp> bedefault \<theta>) \<or> (\<exists>M. belookup \<theta> l = Some M \<and> u \<sharp> f M)"
  by (simp_all add:qfbranch_def)

(*
lemma qfbranch_Beta[simp] :
  fixes a::var
  shows "a \<sharp> (N,\<theta>) \<Longrightarrow> a \<sharp> qfbranch \<theta> LBeta (\<lambda>M. M)"
  apply (rule qfbranch_fresh)
  sorry

lemma const_supp_aux :
  fixes f :: "'a \<Rightarrow> 'b"
  fixes x :: 'x
  fixes M :: 'b
  assumes hconst:"\<And>a. f a = M"
  assumes at:"at TYPE('x)"
  shows "{y. [(x,y)]\<bullet>M \<noteq> M} = {y. [(x,y)]\<bullet>f \<noteq> f}"
apply(rule,rule,simp)
apply(subgoal_tac "[(x, xa)] \<bullet> f = f \<Longrightarrow> [(x, xa)] \<bullet> M = M", blast)
apply(simp add:perm_fun_def assms fun_eq_iff)
apply(rule,rule,simp)
apply(subgoal_tac "[(x, xa)] \<bullet> M = M \<Longrightarrow> [(x, xa)] \<bullet> f = f", blast)
apply(simp add:perm_fun_def assms fun_eq_iff)
done

lemma const_supp :
  fixes f :: "'a \<Rightarrow> 'b"
  fixes M :: 'b
  assumes hconst:"\<And>a. f a = M"
  assumes at:"at TYPE('x)"
  shows "(supp f::'x set) = supp M"
apply(simp add: supp_def)
apply(rule,rule,simp)
proof-
  fix x::'x
  assume "infinite {b. [(x, b)] \<bullet> f \<noteq> f}"
  from this show "infinite {y. [(x, y)] \<bullet> M \<noteq> M}"
  by (auto simp:at const_supp_aux[of f _ x, OF hconst])
next
  show "{a::'x. infinite {b. [(a, b)] \<bullet> M \<noteq> M}} \<subseteq> {a. infinite {b. [(a, b)] \<bullet> f \<noteq> f}}"
  proof (rule,rule,simp)
    fix x::'x
    assume "infinite {b. [(x, b)] \<bullet> M \<noteq> M}"
    from this show "infinite {b. [(x, b)] \<bullet> f \<noteq> f}" 
    by(simp add:at const_supp_aux[of f _ x, OF hconst])
  qed
qed

lemma fresh_funE:
  fixes x::'x
  assumes fresh:"x \<sharp> (\<lambda>_. M)"
  assumes at:"at TYPE('x)"
  shows "x \<sharp> M"
  using assms
apply (subgoal_tac "(supp (\<lambda>_. M)::'x set) = supp M")
apply (simp add:fresh_def, force)
apply (rule const_supp,auto)
done
*)

nominal_primrec
  sfold :: "trailskel \<Rightarrow> exp_repl \<Rightarrow> exp"
  and sfaux :: "trailskel_repl \<Rightarrow> exp_repl \<Rightarrow> exp \<Rightarrow> trail_label \<Rightarrow> exp"
where
  "sfold SRefl \<theta> = qfbranch \<theta> LRefl (\<lambda>M. M)"
| "sfold (STrans s1 s2) \<theta> = qfbranch \<theta> LTrans (\<lambda>M. App (App M (sfold s1 \<theta>)) (sfold s2 \<theta>))"
| "sfold SBeta \<theta> = qfbranch \<theta> LBeta (\<lambda>M. M)"
| "sfold SBetaBox \<theta> = qfbranch \<theta> LBetaBox (\<lambda>M. M)"
| "sfold SITI \<theta> = qfbranch \<theta> LITI (\<lambda>M. M)"
| "sfold SOTI \<theta> = qfbranch \<theta> LOTI (\<lambda>M. M)"
| "sfold (SLam s) \<theta> = qfbranch \<theta> LQLam (\<lambda>M. App M (sfold s \<theta>))"
| "sfold (SApp s1 s2) \<theta> = qfbranch \<theta> LQApp (\<lambda>M. App (App M (sfold s1 \<theta>)) (sfold s2 \<theta>))"
| "sfold (SLet s1 s2) \<theta> = qfbranch \<theta> LQLet (\<lambda>M. App (App M (sfold s1 \<theta>)) (sfold s2 \<theta>))"
| "sfold (SITB s ss) \<theta> = qfbranch \<theta> LITBc (\<lambda>M. qfbranch \<theta> LITBn (\<lambda>N. sfaux ss \<theta> (App (App M N) (sfold s \<theta>)) LITBc))"
| "sfold (SOTB ss) \<theta> = qfbranch \<theta> LOTBn (\<lambda>M. sfaux ss \<theta> M LOTBc)"
| "sfaux (SrNil sdef) \<theta> = (\<lambda>Mnil Lcons. qfbranch \<theta> Lcons (\<lambda>M. App (App M (sfold sdef \<theta>)) Mnil))"
| "sfaux (SrCons x s ss) \<theta> = (\<lambda>Mnil Lcons. qfbranch \<theta> Lcons (\<lambda>M. App (App M (sfold s \<theta>)) (sfaux ss \<theta> Mnil Lcons)))"
by (auto)

lemma sfold_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  shows "pi\<bullet>(sfold q \<theta>) = sfold (pi\<bullet>q) (pi\<bullet>\<theta>)"
  and "pi'\<bullet>(sfold q \<theta>) = sfold (pi'\<bullet>q) (pi'\<bullet>\<theta>)"
  and "\<And>M l. pi\<bullet>(sfaux qs \<theta> M l) = sfaux (pi\<bullet>qs) (pi\<bullet>\<theta>) (pi\<bullet>M) (pi\<bullet>l)"
  and "\<And>M l. pi'\<bullet>(sfaux qs \<theta> M l) = sfaux (pi'\<bullet>qs) (pi'\<bullet>\<theta>) (pi'\<bullet>M) (pi'\<bullet>l)"
apply(nominal_induct q and qs rule:trailskel_trailskel_repl.strong_inducts,auto)
apply(simp add:qfbranch_eqvt, rule arg_cong[of _ _ "qfbranch (_ \<bullet> \<theta>) _"],
      simp add: perm_fun_def perm_pi_simp)+
apply(auto simp add:fun_eq_iff, simp add:qfbranch_eqvt, rule arg_cong[of _ _ "qfbranch (_ \<bullet> \<theta>) _"],
      simp add:perm_fun_def perm_pi_simp)
apply(simp add:qfbranch_eqvt, rule arg_cong[of _ _ "qfbranch (_ \<bullet> \<theta>) _"],
      simp add: perm_fun_def perm_pi_simp qfbranch_eqvt)+
done  

lemma sfold_fresh :
  fixes a :: var
  fixes u :: uvar
  shows "a \<sharp> \<theta> \<Longrightarrow> a \<sharp> sfold s \<theta>"
  and "u \<sharp> \<theta> \<Longrightarrow> u \<sharp> sfold s \<theta>"
  and "a \<sharp> M \<Longrightarrow> a \<sharp> \<theta> \<Longrightarrow> a \<sharp> sfaux ss \<theta> M l"
  and "u \<sharp> M \<Longrightarrow> u \<sharp> \<theta> \<Longrightarrow> u \<sharp> sfaux ss \<theta> M l"
apply (nominal_induct s and ss avoiding:\<theta> M l rule:trailskel_trailskel_repl.strong_inducts, simp_all)
apply (rule qfbranch_fresh, rule bedefault_fresh, simp)
apply (subgoal_tac "a \<sharp> belookup \<theta> l", simp add:fresh_some)
apply (rule belookup_fresh,simp)
apply (rule qfbranch_fresh, rule bedefault_fresh, simp)
apply (subgoal_tac "u \<sharp> belookup \<theta> l", simp add:fresh_some)
apply (rule belookup_fresh,simp)
apply (rule qfbranch_fresh, rule bedefault_fresh, simp)
apply (subgoal_tac "a \<sharp> belookup \<theta> l", simp add:fresh_some)
apply (rule belookup_fresh,simp)
apply (rule qfbranch_fresh, rule bedefault_fresh, simp_all)
apply (subgoal_tac "u \<sharp> belookup \<theta> l", simp add:fresh_some)
apply (rule belookup_fresh,simp)
apply (rule qfbranch_fresh, rule bedefault_fresh, simp,
       subgoal_tac "a \<sharp> belookup \<theta> l", simp add:fresh_some,
       rule belookup_fresh,simp,
       rule qfbranch_fresh, rule bedefault_fresh, simp,
       subgoal_tac "u \<sharp> belookup \<theta> l", simp add:fresh_some,
       rule belookup_fresh,simp)+
apply (rule qfbranch_fresh, rule bedefault_fresh, simp_all)
apply (rule qfbranch_fresh, rule bedefault_fresh, simp_all)
apply (subgoal_tac "a \<sharp> exp.App (exp.App M Ma) (sfold trailskel \<theta>)", auto)
apply (subgoal_tac "a \<sharp> belookup \<theta> l", simp add:fresh_some, rule belookup_fresh, simp)
apply (subgoal_tac "a \<sharp> belookup \<theta> la", simp add:fresh_some, rule belookup_fresh, simp)
apply (rule qfbranch_fresh, rule bedefault_fresh, simp_all)
apply (rule qfbranch_fresh, rule bedefault_fresh, simp_all)
apply (subgoal_tac "u \<sharp> exp.App (exp.App M Ma) (sfold trailskel \<theta>)", auto)
apply (subgoal_tac "u \<sharp> belookup \<theta> l", simp add:fresh_some, rule belookup_fresh, simp)
apply (subgoal_tac "u \<sharp> belookup \<theta> la", simp add:fresh_some, rule belookup_fresh, simp)
apply (rule qfbranch_fresh, rule bedefault_fresh, simp)
apply (subgoal_tac "a \<sharp> belookup \<theta> l", simp add:fresh_some)
apply (rule belookup_fresh,simp)
apply (rule qfbranch_fresh, rule bedefault_fresh, simp_all)
apply (subgoal_tac "u \<sharp> belookup \<theta> l", simp add:fresh_some)
apply (rule belookup_fresh,simp)
apply (rule qfbranch_fresh, rule bedefault_fresh, simp)
apply (subgoal_tac "a \<sharp> belookup \<theta> la", simp add:fresh_some)
apply (rule belookup_fresh,simp)
apply (rule qfbranch_fresh, rule bedefault_fresh, simp_all)
apply (subgoal_tac "u \<sharp> belookup \<theta> la", simp add:fresh_some)
apply (rule belookup_fresh,simp)
apply (rule qfbranch_fresh, rule bedefault_fresh, simp)
apply (subgoal_tac "a \<sharp> belookup \<theta> la", simp add:fresh_some)
apply (rule belookup_fresh,simp)
apply (rule qfbranch_fresh, rule bedefault_fresh, simp_all)
apply (subgoal_tac "u \<sharp> belookup \<theta> la", simp add:fresh_some)
apply (rule belookup_fresh,simp)
done

(* defines recursion over trail q using branches \<theta>, i.e. Def.1 adapted to \<lambda>hs *)
definition qfold :: "trail \<Rightarrow> exp_repl \<Rightarrow> exp"
where "qfold q \<theta> = sfold (skel_of_trail q) \<theta>"

definition qfaux :: "trail_repl \<Rightarrow> exp_repl \<Rightarrow> exp \<Rightarrow> trail_label \<Rightarrow> exp"
where "qfaux qs \<theta> = sfaux (skel_of_trail_repl qs) \<theta>"

lemma qfold_simps[simp]:
  "qfold (Refl M) \<theta> = qfbranch \<theta> LRefl (\<lambda>M. M)"
  "qfold (Trans q1 q2) \<theta> = qfbranch \<theta> LTrans (\<lambda>M. App (App M (qfold q1 \<theta>)) (qfold q2 \<theta>))"
  "x \<sharp> M2 \<Longrightarrow> qfold (Beta \<tau> x M1 M2) \<theta> = qfbranch \<theta> LBeta (\<lambda>M. M)"
  "u \<sharp> M1 \<Longrightarrow> qfold (BetaBox \<tau> M1 u M2) \<theta> = qfbranch \<theta> LBetaBox (\<lambda>M. M)"
  "qfold (ITI q \<theta>') \<theta> = qfbranch \<theta> LITI (\<lambda>M. M)"
  "qfold (OTI q \<theta>') \<theta> = qfbranch \<theta> LOTI (\<lambda>M. M)"
  "qfold (QLam \<tau> x q) \<theta> = qfbranch \<theta> LQLam (\<lambda>M. App M (qfold q \<theta>))"
  "qfold (QApp q1 q2) \<theta> = qfbranch \<theta> LQApp (\<lambda>M. App (App M (qfold q1 \<theta>)) (qfold q2 \<theta>))"
  "u \<sharp> q1 \<Longrightarrow> qfold (QLet \<tau> q1 u q2) \<theta> = qfbranch \<theta> LQLet (\<lambda>M. App (App M (qfold q1 \<theta>)) (qfold q2 \<theta>))"
  "qfold (ITB q qs) \<theta> = qfbranch \<theta> LITBc (\<lambda>M. qfbranch \<theta> LITBn (\<lambda>N. qfaux qs \<theta> (App (App M N) (qfold q \<theta>)) LITBc))"
  "qfold (OTB qs) \<theta> = qfbranch \<theta> LOTBn (\<lambda>M. qfaux qs \<theta> M LOTBc)"
  "qfaux (QrNil qdef) \<theta> = (\<lambda>Mnil Lcons. qfbranch \<theta> Lcons (\<lambda>M. App (App M (qfold qdef \<theta>)) Mnil))"
  "qfaux (QrCons l q qs) \<theta> = (\<lambda>Mnil Lcons. qfbranch \<theta> Lcons (\<lambda>M. App (App M (qfold q \<theta>)) (qfaux qs \<theta> Mnil Lcons)))"
by (simp_all add:qfold_def qfaux_def)

lemma qfold_eqvt[eqvt] :
  fixes \<pi> :: "var prm"
  shows "\<pi> \<bullet> (qfold q \<theta>) = qfold (\<pi> \<bullet> q) (\<pi> \<bullet> \<theta>)"
  and "\<pi> \<bullet> (qfaux qs \<theta> Mnil Lcons) = qfaux (\<pi> \<bullet> qs) (\<pi> \<bullet> \<theta>) (\<pi>  \<bullet> Mnil) (\<pi> \<bullet> Lcons)"
  by (simp_all only: qfold_def qfaux_def eqvts)

lemma qfold_fresh :
  fixes a :: var
  fixes u :: uvar
  shows "a \<sharp> \<theta> \<Longrightarrow> a \<sharp> qfold q \<theta>"
  and "u \<sharp> \<theta> \<Longrightarrow> u \<sharp> qfold q \<theta>"
  and "a \<sharp> M \<Longrightarrow> a \<sharp> \<theta> \<Longrightarrow> a \<sharp> qfaux qs \<theta> M l"
  and "u \<sharp> M \<Longrightarrow> u \<sharp> \<theta> \<Longrightarrow> u \<sharp> qfaux qs \<theta> M l"
by (simp_all add:qfold_def qfaux_def sfold_fresh)

(* substitutions *)

type_synonym subst = "(var * exp) list"

fun
  lookup :: "subst \<Rightarrow> var \<Rightarrow> exp"   
where
  "lookup [] x        = Var x"
| "lookup ((y,M)#\<sigma>) x = (if x=y then M else lookup \<sigma> x)"

lemma lookup_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  shows "pi\<bullet>(lookup \<theta> x) = lookup (pi\<bullet>\<theta>) (pi\<bullet>x)"
  and "pi'\<bullet>(lookup \<theta> x) = lookup (pi'\<bullet>\<theta>) (pi'\<bullet>x)"
by (induct \<theta>) (auto simp add: perm_bij)

lemma lookup_fresh:
  fixes x ::"var" 
  fixes u ::"uvar"
  shows "x \<sharp> \<theta> \<Longrightarrow> x \<sharp> y \<Longrightarrow> x \<sharp> lookup \<theta> y"
  and "u \<sharp> \<theta> \<Longrightarrow>  u \<sharp> lookup \<theta> y"
apply (induct \<theta>)
by (auto simp add: fresh_atm fresh_prod fresh_list_cons)

lemma lookup_default:
  "x \<sharp> \<gamma> \<Longrightarrow>  lookup \<gamma> x = Var x"
  apply (induct \<gamma>)
  by (auto simp add: fresh_atm fresh_prod fresh_list_cons)

type_synonym usubst = "(uvar * exp) list"

fun
  ulookup :: "usubst \<Rightarrow> uvar \<Rightarrow> exp"   
where
  "ulookup [] x        = UVar x"
| "ulookup ((y,M)#\<sigma>) x = (if x=y then M else ulookup \<sigma> x)"

lemma ulookup_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  shows "pi\<bullet>(ulookup \<theta> x) = ulookup (pi\<bullet>\<theta>) (pi\<bullet>x)"
  and "pi'\<bullet>(ulookup \<theta> x) = ulookup (pi'\<bullet>\<theta>) (pi'\<bullet>x)"
by (induct \<theta>) (auto simp add: perm_bij)

lemma ulookup_fresh:
  fixes x ::"var" 
  fixes u::"uvar"
  shows "x \<sharp> \<theta> \<Longrightarrow> x \<sharp> ulookup \<theta> v"
  and "u \<sharp> \<theta> \<Longrightarrow> u \<sharp> v \<Longrightarrow>  u \<sharp> ulookup \<theta> v"
apply (induct \<theta>)
by (auto simp add: fresh_atm fresh_prod fresh_list_cons)

lemma ulookup_default:
  "u \<sharp> \<theta> \<Longrightarrow>  ulookup \<theta> u = UVar u"
  apply (induct \<theta>)
  by (auto simp add: fresh_atm fresh_prod fresh_list_cons)

(* this substitution corresponds to Def. 2, adapted to \<lambda>hs *)
nominal_primrec subst :: "exp \<Rightarrow> usubst \<Rightarrow> subst \<Rightarrow> exp" ("_<_,_>" [100,60] 100)
 and subst_repl :: "exp_repl \<Rightarrow> usubst \<Rightarrow> subst \<Rightarrow> exp_repl" ("_<_,_>" [100,60] 100)
where 
  "(Var y)<S,T> = (lookup T y)"
| "(UVar u)<S,T> = (ulookup S u)"
| "y \<sharp> (S,T) \<Longrightarrow> (Lam \<tau> y M)<S,T> = Lam \<tau> y (M<S,T>)"
| "(App M1 M2)<S,T> = App (M1<S,T>) (M2<S,T>)"
| "(Box M)<S,T> = Box (M<S,T>)"
| "v \<sharp> (M1,S,T) \<Longrightarrow> (LetBox \<tau> M1 v M2)<S,T> = LetBox \<tau> (M1<S,T>) v (M2<S,T>)"
| "(Insp er)<S,T> = Insp (er<S,T>)"
| "(IInsp M er)<S,T> = IInsp (M<S,T>) (er<S,T>)"
| "(ErNil M)<S,T> = ErNil (M<S,T>)"
| "(ErCons l M er)<S,T> = ErCons l (M<S,T>) (er<S,T>)"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(simp_all add: abs_fresh fresh_prod exp.inject fresh_ty lookup_eqvt ulookup_eqvt)+
  apply(fresh_guess)+
  done

lemma subst_eqvt[eqvt]:
  fixes \<pi>::"var prm"
  fixes \<pi>'::"uvar prm"
  fixes M::exp
  fixes Ms :: exp_repl
  fixes x::var
  shows "\<pi> \<bullet> (M<S,T>) = (\<pi>\<bullet>M)<\<pi>\<bullet>S,\<pi>\<bullet>T>"
  and "\<pi>' \<bullet> (M<S,T>) = (\<pi>'\<bullet>M)<\<pi>'\<bullet>S,\<pi>'\<bullet>T>"
  and "\<pi> \<bullet> (Ms<S,T>) = (\<pi>\<bullet>Ms)<\<pi>\<bullet>S,\<pi>\<bullet>T>"
  and "\<pi>' \<bullet> (Ms<S,T>) = (\<pi>'\<bullet>Ms)<\<pi>'\<bullet>S,\<pi>'\<bullet>T>"
  apply(nominal_induct M and Ms avoiding: S T rule:exp_exp_repl.strong_inducts)
  apply(auto simp add:perm_bij fresh_bij lookup_eqvt ulookup_eqvt)
  done

lemma subst2_fresh: 
  fixes M ::exp
  fixes Ms ::exp_repl
  fixes x::var
  fixes u::uvar
  shows "x \<sharp> M \<Longrightarrow> x \<sharp> S \<Longrightarrow> x \<sharp> T \<Longrightarrow> x \<sharp> M<S,T> "  
  and "u \<sharp> M \<Longrightarrow> u \<sharp> S \<Longrightarrow> u \<sharp> T \<Longrightarrow> u \<sharp> M<S,T> "
  and "x \<sharp> Ms \<Longrightarrow> x \<sharp> S \<Longrightarrow> x \<sharp> T \<Longrightarrow> x \<sharp> Ms<S,T> "  
  and "u \<sharp> Ms \<Longrightarrow> u \<sharp> S \<Longrightarrow> u \<sharp> T \<Longrightarrow> u \<sharp> Ms<S,T> "  
  apply(nominal_induct M and Ms avoiding: S T rule:exp_exp_repl.strong_inducts)
  apply(auto simp add: abs_fresh fresh_ty fresh_list_cons fresh_list_nil lookup_fresh ulookup_fresh)
  done

lemma var_id:
  fixes M :: exp
  fixes Ms :: exp_repl
  shows "x \<sharp> \<gamma> \<Longrightarrow> M<\<delta>,(x,Var x)#\<gamma>> = M<\<delta>,\<gamma>>"
  and "x \<sharp> \<gamma> \<Longrightarrow> Ms<\<delta>,(x,Var x)#\<gamma>> = Ms<\<delta>,\<gamma>>"
  apply(nominal_induct M and Ms avoiding: \<gamma> \<delta> x rule:exp_exp_repl.strong_inducts)
  apply(auto simp add: fresh_list_cons fresh_prod lookup_default)
  done

lemma uvar_id:
  fixes M :: exp
  fixes Ms :: exp_repl
  shows "u \<sharp> \<delta> \<Longrightarrow>M<(u,UVar u)#\<delta>,\<gamma>> = M<\<delta>,\<gamma>>"
  and"u \<sharp> \<delta> \<Longrightarrow>Ms<(u,UVar u)#\<delta>,\<gamma>> = Ms<\<delta>,\<gamma>>"
  apply(nominal_induct M and Ms avoiding: \<gamma> \<delta> u rule:exp_exp_repl.strong_inducts)
  apply(auto simp add: fresh_list_cons fresh_prod ulookup_default)
  done

abbreviation
  subst1 :: "exp \<Rightarrow> var \<Rightarrow> exp \<Rightarrow> exp" ("_[_\<mapsto>_]" [60,60,60] 100) 
where
  "M[x\<mapsto>N] \<equiv>  M<[],[(x,N)]>"

abbreviation
  subst1_repl :: "exp_repl \<Rightarrow> var \<Rightarrow> exp \<Rightarrow> exp_repl" ("_[_\<mapsto>_]" [60,60,60] 100) 
where
  "subst1_repl Ms x N \<equiv>  Ms<[],[(x,N)]>"

lemma subst_lam_simp[simp]:
  "x\<sharp> (y,N) \<Longrightarrow> (Lam ty x M)[y\<mapsto>N] = Lam ty x (M[y\<mapsto>N])"  
  by(simp add: fresh_list_cons fresh_list_nil)
  
lemma subst_letbox_simp[simp]:
  "u \<sharp> (M,Q) \<Longrightarrow> (LetBox \<tau> M u N)[y\<mapsto>Q] = LetBox \<tau> (M[y\<mapsto> Q]) u (N[y\<mapsto>Q])"  
  by(simp add: fresh_list_cons fresh_list_nil fresh_atm)

lemma subst_fresh: 
  fixes M::exp
  fixes Ms::exp_repl
  fixes x::var
  shows "x \<sharp> [y].M \<Longrightarrow> x \<sharp> N \<Longrightarrow> x \<sharp> (M[y\<mapsto>N])"
  and "x \<sharp> [y].Ms \<Longrightarrow> x \<sharp> N \<Longrightarrow> x \<sharp> (Ms[y\<mapsto>N])"
  apply(nominal_induct M and Ms avoiding: y N rule:exp_exp_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_trail_label fresh_ty)
  done

lemma var_subst_fresh: 
  fixes M::exp
  fixes Ms::exp_repl
  fixes x::var
  shows "x \<sharp> N \<Longrightarrow> x \<sharp> (M[x\<mapsto>N])"  
  and "x \<sharp> N \<Longrightarrow> x \<sharp> (Ms[x\<mapsto>N])"  
  apply(nominal_induct M and Ms avoiding: x N rule:exp_exp_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_trail_label fresh_ty)
  done

lemma var_subst_id:
  fixes M::exp
  fixes Ms::exp_repl
  shows "x \<sharp> M \<Longrightarrow> M[x\<mapsto>N] = M"
  and "x \<sharp> Ms \<Longrightarrow> Ms[x\<mapsto>N] = Ms"
  apply(nominal_induct M and Ms avoiding: x N rule:exp_exp_repl.strong_inducts)
  apply(auto simp add: fresh_atm abs_fresh )
  done

lemma subst1_merge:
  fixes M::exp
  fixes Ms::exp_repl
  shows  "x \<sharp> \<delta> \<Longrightarrow> x \<sharp> \<gamma> \<Longrightarrow> M<\<delta>,(x, N) # \<gamma>> = M<\<delta>, \<gamma>>[x\<mapsto>N]"
  and "x \<sharp> \<delta> \<Longrightarrow> x \<sharp> \<gamma> \<Longrightarrow> Ms<\<delta>,(x, N) # \<gamma>> = Ms<\<delta>, \<gamma>>[x\<mapsto>N]"
  apply(nominal_induct M and Ms avoiding: \<gamma> x \<delta> N rule:exp_exp_repl.strong_inducts)
  apply(auto simp add:lookup_default fresh_atm var_subst_id lookup_fresh)
  apply(auto simp add:ulookup_default var_subst_id ulookup_fresh)
  apply(auto simp add: fresh_list_cons subst2_fresh fresh_prod fresh_atm fresh_list_nil)
  done

abbreviation
  usubst1 :: "exp \<Rightarrow> uvar \<Rightarrow> exp \<Rightarrow> exp" ("_{_\<mapsto>_}" [100,60,60] 100) 
where
  "M{u\<mapsto>N} \<equiv>  M<[(u,N)],[]>"

abbreviation
  usubst1_repl :: "exp_repl \<Rightarrow> uvar \<Rightarrow> exp \<Rightarrow> exp_repl" ("_{_\<mapsto>_}" [100,60,60] 100) 
where
  "usubst1_repl Ms u N \<equiv>  Ms<[(u,N)],[]>"

lemma usubst_lam_simp[simp]:
  "x\<sharp> (N) \<Longrightarrow> (Lam ty x M){u\<mapsto>N} = Lam ty x (M{u\<mapsto>N})"  
  by(simp add: fresh_list_cons fresh_list_nil fresh_atm)
  
lemma usubst_letbox_simp[simp]:
  "v \<sharp> (M,u,Q) \<Longrightarrow> (LetBox \<tau> M v N){u\<mapsto>Q} = LetBox \<tau> (M{u\<mapsto>Q}) v (N{u\<mapsto>Q})"  
  by(simp add: fresh_list_cons fresh_list_nil fresh_atm fresh_prod)
  
lemma uvar_usubst_fresh: 
  fixes M::exp
  fixes Ms::exp_repl
  fixes u::uvar
  shows "u \<sharp> N \<Longrightarrow> u \<sharp> (M{u\<mapsto>N})"
  and "u \<sharp> N \<Longrightarrow> u \<sharp> (Ms{u\<mapsto>N})"  
  apply(nominal_induct M and Ms avoiding: u N rule:exp_exp_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_trail_label fresh_ty)
  done

lemma usubst_fresh: 
  fixes M::exp
  fixes Ms::exp_repl
  fixes u::uvar
  shows "u \<sharp> [v].M \<Longrightarrow> u \<sharp> N \<Longrightarrow> u \<sharp> (M{v\<mapsto>N})"  
  and "u \<sharp> [v].Ms \<Longrightarrow> u \<sharp> N \<Longrightarrow> u \<sharp> (Ms{v\<mapsto>N})"  
  apply(nominal_induct M and Ms avoiding: v N rule:exp_exp_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_ty fresh_trail_label)
  done

lemma subst_fresh': 
  fixes M::exp
  fixes Ms::exp_repl
  fixes u::uvar
  fixes a::var
  shows "u \<sharp> [a].M \<Longrightarrow> u \<sharp> N \<Longrightarrow> u \<sharp> (subst1 M a N)" 
  and "u \<sharp> [a].Ms \<Longrightarrow> u \<sharp> N \<Longrightarrow> u \<sharp> (subst1_repl Ms a N)" 
  apply(nominal_induct M and Ms avoiding: a N rule:exp_exp_repl.strong_inducts)
  apply(auto simp add:exp.fresh abs_fresh fresh_ty)
  done

lemma usubst_fresh': 
  fixes M::exp
  fixes Ms::exp_repl
  fixes a::var
  shows "a \<sharp> [v].M \<Longrightarrow> a \<sharp> N \<Longrightarrow> a \<sharp> (M{v\<mapsto>N})"  
  and "a \<sharp> [v].Ms \<Longrightarrow> a \<sharp> N \<Longrightarrow> a \<sharp> (Ms{v\<mapsto>N})"  
  apply(nominal_induct M and Ms avoiding: v N rule:exp_exp_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_ty )
  done

lemma uvar_subst_id:
  fixes M::exp
  fixes Ms::exp_repl
  shows "u \<sharp> M \<Longrightarrow> M{u\<mapsto>N} = M"
  and "u \<sharp> Ms \<Longrightarrow> Ms{u\<mapsto>N} = Ms"
  apply(nominal_induct M and Ms avoiding: u N rule:exp_exp_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh) 
  done

lemma usubst1_merge:
  fixes M::exp
  fixes Ms::exp_repl
  shows "u \<sharp> \<delta> \<Longrightarrow> u \<sharp> \<gamma> \<Longrightarrow> M<(u,N)#\<delta>, \<gamma>> = M<\<delta>, \<gamma>>{u\<mapsto>N}"
  and "u \<sharp> \<delta> \<Longrightarrow> u \<sharp> \<gamma> \<Longrightarrow> Ms<(u,N)#\<delta>, \<gamma>> = Ms<\<delta>, \<gamma>>{u\<mapsto>N}"
  apply(nominal_induct M and Ms avoiding: \<gamma> u \<delta> N rule:exp_exp_repl.strong_inducts)
  apply(auto simp add:lookup_default fresh_atm uvar_subst_id lookup_fresh)
  apply(auto simp add:ulookup_default uvar_subst_id ulookup_fresh fresh_atm)
  apply(auto simp add: fresh_list_cons subst2_fresh fresh_prod fresh_atm fresh_list_nil)
  done

(* Identity substitutions *)

fun
  id_subst :: "ctx \<Rightarrow> subst"   
where
  "id_subst []        = []"
| "id_subst ((y,\<tau>)#\<Gamma>) = (y,Var y)#(id_subst \<Gamma>)"

lemma id_lookup: 
  "lookup (id_subst \<Gamma>) x = Var x"
  by(induct \<Gamma>,auto)

lemma id_subst_fresh:
  fixes x::var
  and u::uvar
  shows "x \<sharp> \<Gamma> \<Longrightarrow> x \<sharp> id_subst \<Gamma>"
  and "u \<sharp> \<Gamma> \<Longrightarrow> u \<sharp> id_subst \<Gamma>"
  apply(induct \<Gamma>)
  apply(auto simp add:fresh_list_nil fresh_list_cons fresh_prod fresh_atm)
  done


fun
  id_usubst :: "uctx \<Rightarrow> usubst"   
where
  "id_usubst []        = []"
| "id_usubst ((u,\<tau>)#\<Delta>) = (u,UVar u)#(id_usubst \<Delta>)"

lemma id_usubst_fresh:
  fixes x::var
  and u::uvar
  shows "x \<sharp> \<Delta> \<Longrightarrow> x \<sharp> id_usubst \<Delta>"
  and "u \<sharp> \<Delta> \<Longrightarrow> u \<sharp> id_usubst \<Delta>"
  apply(induct \<Delta>)
  apply(auto simp add:fresh_list_nil fresh_list_cons fresh_prod fresh_atm)
  done

lemma id_ulookup: 
  "ulookup (id_usubst \<Delta>) x = UVar x"
  by(induct \<Delta>,auto)


lemma id_subst_is_id:
  fixes M::exp
  fixes Ms::exp_repl
  shows "M<id_usubst \<Delta>,id_subst \<Gamma>> = M"
  and "Ms<id_usubst \<Delta>,id_subst \<Gamma>> = Ms"
  apply(nominal_induct M and Ms avoiding: \<Delta> \<Gamma> rule:exp_exp_repl.strong_inducts)
  apply(auto simp add:id_lookup id_ulookup id_subst_fresh id_usubst_fresh)
  done


lemma id_usubst_is_id:
  fixes M::exp
  fixes Ms::exp_repl
  shows "M<id_usubst \<Delta>,id_subst \<Gamma>> = M"
  and "Ms<id_usubst \<Delta>,id_subst \<Gamma>> = Ms"
  apply(nominal_induct M and Ms avoiding: \<Gamma> \<Delta> rule:exp_exp_repl.strong_inducts)
  apply(auto simp add: id_subst_fresh id_lookup id_usubst_fresh id_ulookup)
  done

(* Typing (Fig. 4 in the paper) *)
inductive wf :: "uctx \<Rightarrow> ctx \<Rightarrow> exp \<Rightarrow> ty \<Rightarrow> bool" ("_;_ \<turnstile> _ : _" [60,60,60,60] 60)
  and wf_repl :: "uctx \<Rightarrow> exp_repl \<Rightarrow> ty \<Rightarrow> bool" ("_ \<turnstile> _ : _" [60,60,60] 60)
where
  var_ty : "\<lbrakk>(x,\<tau>) \<in> set \<Gamma>; valid \<Gamma>; uvalid \<Delta> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> Var x : \<tau>"
| uvar_ty : "\<lbrakk>(u,\<tau>) \<in> set \<Delta>; valid \<Gamma>; uvalid \<Delta> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> UVar u : \<tau>"
| lam_ty : "\<lbrakk>\<Delta>;(x,\<tau>)#\<Gamma> \<turnstile> M : \<tau>'; x \<sharp> \<Gamma> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> Lam \<tau> x M : FunTy \<tau> \<tau>'"
| app_ty : "\<lbrakk>\<Delta>;\<Gamma> \<turnstile> M : FunTy \<tau> \<tau>'  ; \<Delta>;\<Gamma> \<turnstile> N : \<tau> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> App M N : \<tau>'"
| box_ty : "\<lbrakk>\<Delta>;[] \<turnstile> M : \<tau>; valid \<Gamma> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> Box M : BoxTy \<tau>"
| letbox_ty : "\<lbrakk>\<Delta>;\<Gamma> \<turnstile> M : BoxTy \<tau>  ; (u,\<tau>)#\<Delta>;\<Gamma> \<turnstile> N : \<tau>'; u \<sharp> (M,\<Delta>) \<rbrakk> \<Longrightarrow> 
                \<Delta>;\<Gamma> \<turnstile> LetBox \<tau> M u N : \<tau>'"
| tinsp_ty : "\<lbrakk> \<Delta> \<turnstile> Ms : \<tau>; valid \<Gamma>\<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> Insp Ms : \<tau>"
| tiinsp_ty : "\<lbrakk> \<Delta>;\<Gamma> \<turnstile> M : BoxTy \<sigma>; \<Delta> \<turnstile> Ns : \<tau> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> IInsp M Ns : \<tau>"
(* wf_repl *)
| ernil_ty : "\<Delta>;[] \<turnstile> M : \<tau> \<Longrightarrow> \<Delta> \<turnstile> ErNil M : \<tau>" (* M is the default branch *)
| ercons_ty : "\<lbrakk>\<Delta> \<turnstile> Ms : \<tau>; \<Delta>;[] \<turnstile> M : bty \<tau> l \<rbrakk> \<Longrightarrow> \<Delta> \<turnstile> ErCons l M Ms : \<tau>"

inductive_cases wf_cases[elim]:
"\<Delta>;\<Gamma> \<turnstile> Var x : \<tau>"
"\<Delta>;\<Gamma> \<turnstile> UVar u : \<tau>"
"\<Delta>;\<Gamma> \<turnstile> Lam \<tau> x M : \<tau>'"
"\<Delta>;\<Gamma> \<turnstile> App M N : \<tau>'"
"\<Delta>;\<Gamma> \<turnstile> Box M : \<tau>"
"\<Delta>;\<Gamma> \<turnstile> LetBox \<tau> M u N : \<tau>'"

inductive_cases wf_repl_cases[elim]:
"\<Delta> \<turnstile> ErNil M : \<tau>"
"\<Delta> \<turnstile> ErCons l M Ms : \<tau>"

equivariance wf

nominal_inductive wf
  apply(auto intro: fresh_ty var_fresh_uctx uvar_fresh_ctx)
  apply(simp_all add:abs_fresh)
  done

lemma wf_valid:
  shows "\<Delta>;\<Gamma> \<turnstile> M : \<tau> \<Longrightarrow> uvalid \<Delta> \<and> valid \<Gamma>"
  and "\<Delta> \<turnstile> Ms : \<tau> \<Longrightarrow> uvalid \<Delta>"
  apply(nominal_induct \<Delta> \<Gamma> M \<tau> and \<Delta> Ms \<tau> rule:wf_wf_repl.strong_inducts)
  by (auto elim:valid.cases)

(* Strong inversion for lambda and box *)

lemma strong_lam_wf_elim:
  "\<Delta>;\<Gamma> \<turnstile> Lam \<tau>1 x M : \<tau> \<Longrightarrow> x \<sharp> \<Gamma> \<Longrightarrow> \<exists> \<tau>2. \<Delta>;(x,\<tau>1)#\<Gamma> \<turnstile>M : \<tau>2 \<and> FunTy \<tau>1 \<tau>2 = \<tau>"
  apply(subgoal_tac "\<exists> u::uvar. u \<sharp> (\<Gamma>,\<Delta>,\<tau>,M)")
  apply(clarify)
  apply(erule_tac x="x" and u="u" in wf.strong_cases)
  apply(simp_all add: exp.inject ty.inject alpha abs_fresh var_fresh_uctx fresh_ty)
  apply(rule exists_fresh, auto,finite_guess)
  done

lemma strong_letbox_wf_elim:
  "\<Delta>;\<Gamma> \<turnstile> LetBox \<tau>' M u N : \<tau> \<Longrightarrow> u \<sharp> (M,\<Delta>) \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> M : BoxTy \<tau>' \<and> (u,\<tau>')#\<Delta>;\<Gamma> \<turnstile> N : \<tau>"
  apply(subgoal_tac "\<exists> x::var. x \<sharp> (\<Gamma>,\<Delta>,M,N)")
  apply(clarify)
  apply(erule_tac x="x" and u="u" in wf.strong_cases)
  apply(simp_all add: exp.inject ty.inject alpha abs_fresh uvar_fresh_ctx fresh_ty)
  apply(rule exists_fresh,auto,finite_guess)
  done

(* Weakening *)

definition
  "sub_context" :: "'a list \<Rightarrow> 'a list \<Rightarrow> bool" ("_ \<subseteq> _" [60,60] 60) 
where
  "xs \<subseteq> ys \<equiv> \<forall>x. x\<in>set xs \<longrightarrow> x\<in>set ys"


lemma var_in_fresh: 
  fixes x::var
  and \<Gamma> ::ctx
  shows "x \<sharp> \<Gamma> \<Longrightarrow> (y, \<tau>) \<in> set \<Gamma> \<Longrightarrow> x \<sharp> y"
  apply(induct \<Gamma>)
  apply(auto simp add:fresh_list_cons)
  done

lemma uvar_in_fresh: 
  fixes u::uvar
  and \<Delta> ::uctx
  shows "u \<sharp> \<Delta> \<Longrightarrow> (v, \<tau>) \<in> set \<Delta> \<Longrightarrow> u \<sharp> v"
  apply(induct \<Delta>) 
  apply(auto simp add:fresh_list_cons)
  done

lemma var_sub_fresh:
  fixes x::var
  and \<Gamma> \<Gamma>'::ctx
  shows "\<Gamma> \<subseteq> \<Gamma>' \<Longrightarrow> x \<sharp> \<Gamma>' \<Longrightarrow> x \<sharp> \<Gamma>"
  apply(induct \<Gamma>)
  apply(simp add:fresh_list_nil)
  apply(clarsimp, auto simp add:fresh_list_cons fresh_prod)
  apply(simp add: sub_context_def)
  apply(subgoal_tac "(a, b) \<in> set \<Gamma>'",auto intro: var_in_fresh)
  apply(auto simp add:sub_context_def fresh_ty)
  done

lemma uvar_sub_fresh:
  fixes u::uvar
  and \<Delta> \<Delta>'::uctx
  shows "\<Delta> \<subseteq> \<Delta>' \<Longrightarrow> u \<sharp> \<Delta>' \<Longrightarrow> u \<sharp> \<Delta>"
  apply(induct \<Delta>)
  apply(simp add:fresh_list_nil)
  apply(clarsimp, auto simp add:fresh_list_cons fresh_prod)
  apply(simp add: sub_context_def)
  apply(subgoal_tac "(a, b) \<in> set \<Delta>'",auto intro: uvar_in_fresh)
  apply(auto simp add:sub_context_def fresh_ty)
  done

lemma sub_context_refl :
  fixes \<Gamma>::ctx
  fixes \<Delta>::uctx
  shows "\<Gamma> \<subseteq> \<Gamma>"
  and "\<Delta> \<subseteq> \<Delta>"
  by (simp_all add:sub_context_def)

lemma weakening:
  "\<Delta>;\<Gamma> \<turnstile> M : \<tau> \<Longrightarrow> \<Gamma> \<subseteq> \<Gamma>' \<Longrightarrow> \<Delta> \<subseteq> \<Delta>' \<Longrightarrow> valid \<Gamma>' \<Longrightarrow> uvalid \<Delta>' \<Longrightarrow> \<Delta>';\<Gamma>' \<turnstile> M : \<tau>"
  and "\<Delta> \<turnstile> Ms : \<tau> \<Longrightarrow> \<Delta> \<subseteq> \<Delta>' \<Longrightarrow> uvalid \<Delta>' \<Longrightarrow> \<Delta>' \<turnstile> Ms : \<tau>"
  apply(nominal_induct \<Delta> \<Gamma> M \<tau> and \<Delta> Ms \<tau>
        avoiding: \<Delta>' \<Gamma>' rule:wf_wf_repl.strong_inducts)
  apply(rule wf_wf_repl.intros,simp add:sub_context_def,simp_all)
  apply(rule wf_wf_repl.intros,simp add:sub_context_def,simp_all)
  apply(rule wf_wf_repl.intros)
  apply(subgoal_tac "(x, \<tau>) # \<Gamma> \<subseteq> (x, \<tau>) # \<Gamma>'", auto intro:valid.intros)
  apply(simp add:sub_context_def)
  apply(rule wf_wf_repl.intros,auto)
  apply(rule wf_wf_repl.intros,auto,simp add:sub_context_def valid.intros)
  apply(rule wf_wf_repl.intros,auto)
  apply(subgoal_tac "(u, \<tau>) # \<Delta> \<subseteq> (u, \<tau>) # \<Delta>' \<and> uvalid ((u,\<tau>)#\<Delta>')")
  apply(simp)
  apply(simp add:sub_context_def uvalid.intros)
  apply(rule wf_wf_repl.intros,auto)
  apply(rule wf_wf_repl.intros,auto)
  apply(rule wf_wf_repl.intros)
  apply(subgoal_tac "[] \<subseteq> [] \<and> valid []")
  apply(auto simp add:sub_context_refl,rule valid.intros)
  apply(rule wf_wf_repl.intros,auto)
  apply(subgoal_tac "[] \<subseteq> [] \<and> valid []")
  apply(auto simp add:sub_context_refl,rule valid.intros)
  done

inductive wf_subst:: "uctx \<Rightarrow> ctx \<Rightarrow> usubst \<Rightarrow> subst \<Rightarrow> uctx \<Rightarrow> ctx \<Rightarrow> bool" 
   ("_;_ \<turnstile> <_,_> : _;_" [60,60,60,60,60,60] 60)
where
  wf_subst_nil: "\<lbrakk>uvalid \<Delta>; valid \<Gamma> \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> <[],[]> : [];[]"
| wf_subst_cons1: "\<lbrakk>\<Delta>;\<Gamma> \<turnstile> <\<delta>,\<gamma>> : \<Delta>';\<Gamma>'; \<Delta>;\<Gamma> \<turnstile> M : \<tau>; x \<sharp>(\<gamma>,\<Gamma>') \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> <\<delta>,(x,M)#\<gamma>> : \<Delta>';(x,\<tau>)#\<Gamma>'"
| wf_subst_cons2: "\<lbrakk>\<Delta>;\<Gamma> \<turnstile> <\<delta>,\<gamma>> : \<Delta>';\<Gamma>'; \<Delta>;[] \<turnstile> M : \<tau>; u \<sharp>(\<delta>,\<Delta>') \<rbrakk> \<Longrightarrow> \<Delta>;\<Gamma> \<turnstile> <(u,M)#\<delta>,\<gamma>> : (u,\<tau>)#\<Delta>';\<Gamma>'"

equivariance wf_subst
nominal_inductive wf_subst
  done

lemma wf_subst_valid:
  "\<Delta>';\<Gamma>' \<turnstile> <\<delta>,\<gamma>> : \<Delta>;\<Gamma> \<Longrightarrow> valid \<Gamma>'"
  apply(nominal_induct \<Delta>' \<Gamma>' \<delta> \<gamma> \<Delta> \<Gamma> rule: wf_subst.strong_induct)
  by(auto)
lemma wf_subst_uvalid:
  "\<Delta>';\<Gamma>' \<turnstile> <\<delta>,\<gamma>> : \<Delta>;\<Gamma> \<Longrightarrow> uvalid \<Delta>'"
  apply(nominal_induct \<Delta>' \<Gamma>' \<delta> \<gamma> \<Delta> \<Gamma> rule: wf_subst.strong_induct)
  by(auto)


lemma lookup_wf_subst:
  "\<Delta>';\<Gamma>' \<turnstile> <\<delta>,\<gamma>> : \<Delta>;\<Gamma> \<Longrightarrow> (x,\<tau>) \<in> set \<Gamma> \<Longrightarrow> \<Delta>';\<Gamma>' \<turnstile> lookup \<gamma> x : \<tau>"
  apply(nominal_induct \<Delta>' \<Gamma>' \<delta> \<gamma> \<Delta> \<Gamma> avoiding: x \<tau> rule: wf_subst.strong_induct)
  apply(auto)
  apply(frule fresh_var_not_in,simp_all)
  done

lemma ulookup_wf_subst:
  "\<Delta>';\<Gamma>' \<turnstile> <\<delta>,\<gamma>> : \<Delta>;\<Gamma> \<Longrightarrow> (u,\<tau>) \<in> set \<Delta> \<Longrightarrow> \<Delta>';\<Gamma>' \<turnstile> ulookup \<delta> u : \<tau>"
  apply(nominal_induct \<Delta>' \<Gamma>' \<delta> \<gamma> \<Delta> \<Gamma> avoiding: u \<tau> rule: wf_subst.strong_induct)
  apply(auto)
  apply(erule weakening, simp_all add:sub_context_def wf_subst_valid wf_subst_uvalid)
  apply(frule fresh_uvar_not_in,simp_all)
  done


lemma empty_fresh: 
  fixes x::"var"
  shows "\<Delta>;\<Gamma> \<turnstile> M : \<tau> \<Longrightarrow> x \<sharp> \<Gamma> \<Longrightarrow> x \<sharp> M"
  and "\<Delta> \<turnstile> Ms : \<tau> \<Longrightarrow> x \<sharp> Ms"
  apply(nominal_induct \<Delta> \<Gamma> M \<tau> and \<Delta> Ms \<tau> avoiding: x rule: wf_wf_repl.strong_inducts)
  apply(auto simp add: fresh_atm fresh_list_nil abs_fresh fresh_list_cons fresh_ty)
  apply(erule fresh_var_not_in,simp)
  apply(rule fresh_trail_label)
  done

lemma fresh_sub_id:
  fixes x::"var"
  fixes M::"exp"
  fixes Ms::"exp_repl"
  shows "x \<sharp> M \<Longrightarrow> M<\<delta>,(x,N)#\<gamma>> = M<\<delta>,\<gamma>>"
  and "x \<sharp> Ms \<Longrightarrow> Ms<\<delta>,(x,N)#\<gamma>> = Ms<\<delta>,\<gamma>>"
  apply(nominal_induct M and Ms avoiding: x N \<delta> \<gamma> rule:exp_exp_repl.strong_inducts)
  apply(auto simp add:fresh_atm abs_fresh fresh_prod fresh_list_cons)
  done

lemma box_empty: "\<Delta>;[] \<turnstile> M : \<tau> \<Longrightarrow> M<\<delta>,\<gamma>> = M<\<delta>,[]>"
  apply(induct \<gamma>)
  apply(simp)
  apply(clarify)
  apply(subgoal_tac "M<\<delta>,(a,b)#\<gamma>> = M<\<delta>,\<gamma>>",simp)
  apply(rule fresh_sub_id, erule empty_fresh, simp add: fresh_list_nil)
  done

lemma box_empty_repl: "\<Delta> \<turnstile> Ms : \<tau> \<Longrightarrow> Ms<\<delta>,\<gamma>> = Ms<\<delta>,[]>"
  apply(induct \<gamma>)
  apply(simp)
  apply(clarify)
  apply(subgoal_tac "Ms<\<delta>,(a,b)#\<gamma>> = Ms<\<delta>,\<gamma>>",simp)
  apply(rule fresh_sub_id, erule empty_fresh)
  done

lemma wf_subst_weakening:
  "\<Delta>';\<Gamma>' \<turnstile> <\<delta>,\<gamma>> : \<Delta>;\<Gamma> \<Longrightarrow> \<Gamma>' \<subseteq> \<Gamma>'' \<Longrightarrow> \<Delta>' \<subseteq> \<Delta>'' \<Longrightarrow> valid \<Gamma>'' \<Longrightarrow> uvalid \<Delta>'' \<Longrightarrow> \<Delta>'';\<Gamma>'' \<turnstile> <\<delta>,\<gamma>> : \<Delta>;\<Gamma> "
  apply(nominal_induct \<Delta>' \<Gamma>' \<delta> \<gamma> \<Delta> \<Gamma> avoiding: \<Gamma>'' \<Delta>'' rule: wf_subst.strong_induct)
  apply(erule wf_subst.intros,simp)
  apply(rule wf_subst.intros,auto, erule weakening,simp_all add:sub_context_def valid.intros)+
  done

lemma wf_subst_usubst: "\<Delta>';\<Gamma>' \<turnstile> <\<delta>,\<gamma>> : \<Delta>;\<Gamma> \<Longrightarrow> \<Delta>';[] \<turnstile> <\<delta>,[]> : \<Delta>;[]"
  apply(nominal_induct \<Delta>' \<Gamma>' \<delta> \<gamma> \<Delta> \<Gamma> rule: wf_subst.strong_induct)
  apply(auto intro:wf_subst.intros uvalid.intros valid.intros)
  done

lemma id_subst_simple:
  fixes M::exp
  fixes Ms::exp_repl
  shows "M<id_usubst \<Delta>,(x, N) # id_subst \<Gamma>> = M[x\<mapsto>N]"
  and "Ms<id_usubst \<Delta>,(x, N) # id_subst \<Gamma>> = Ms[x\<mapsto>N]"
  apply(nominal_induct M and Ms avoiding: \<Delta> x N \<Gamma> rule:exp_exp_repl.strong_inducts)
  apply(auto simp add:id_lookup id_ulookup )
  apply(auto simp add: fresh_prod fresh_list_cons id_subst_fresh id_usubst_fresh fresh_list_nil)
  done

lemma id_usubst_simple:
  fixes M::exp
  fixes Ms::exp_repl
  shows "M<(u,N)#id_usubst \<Delta>,id_subst \<Gamma>> = M{u\<mapsto>N}"
  and "Ms<(u,N)#id_usubst \<Delta>,id_subst \<Gamma>> = Ms{u\<mapsto>N}"
  apply(nominal_induct M and Ms avoiding: \<Delta> u N \<Gamma> rule:exp_exp_repl.strong_inducts)
  apply(auto simp add:id_lookup id_ulookup )
  apply(auto simp add: fresh_prod fresh_list_cons id_subst_fresh id_usubst_fresh fresh_list_nil)
  done

lemma id_usubst_wf:
"\<Delta> \<subseteq> \<Delta>' \<Longrightarrow>  valid \<Gamma>' \<Longrightarrow> uvalid \<Delta> \<Longrightarrow> uvalid \<Delta>' \<Longrightarrow> \<Delta>';\<Gamma>' \<turnstile> <id_usubst \<Delta>,[]> : \<Delta>;[]"
  apply(induct \<Delta>)
  apply(simp_all)
  apply(rule wf_subst.intros,simp_all add:valid.intros)
  apply(clarsimp)
  apply(erule uvalid.cases,simp,clarsimp)
  apply(rule wf_subst.intros)
  apply(simp add:sub_context_def)
  apply(rule wf_wf_repl.intros, simp add:sub_context_def, simp add:valid.intros, simp)
  apply(simp add:fresh_prod id_usubst_fresh)
  done

lemma id_subst2_wf:
"\<Gamma> \<subseteq> \<Gamma>' \<Longrightarrow> uvalid \<Delta> \<Longrightarrow> valid \<Gamma> \<Longrightarrow> valid \<Gamma>' \<Longrightarrow> \<Delta>;\<Gamma>' \<turnstile> <id_usubst \<Delta>,id_subst \<Gamma>> : \<Delta>;\<Gamma>"
  apply(induct \<Gamma>)
  apply(simp_all) 
  apply(rule id_usubst_wf,simp add:sub_context_def, simp_all add:valid.intros)
  apply(clarsimp)
  apply(erule valid.cases,simp,clarsimp)
  apply(rule wf_subst.intros)
  apply(simp add:sub_context_def)
  apply(rule wf_wf_repl.intros, simp add:sub_context_def, simp add:valid.intros, simp)
  apply(simp add:fresh_prod id_subst_fresh)
  done

(* Reduction rules (Fig.4 in the paper)
   In the paper, there is a single congruence rule for contexts.
   Here, we prefer to use a congruence rule for every subcase. 
 *)  
inductive red :: "exp \<Rightarrow> exp \<Rightarrow> bool" ("_ \<leadsto> _" [60,60] 60)
and red_repl :: "exp_repl \<Rightarrow> exp_repl \<Rightarrow> bool" ("_ \<leadsto> _" [60,60] 60)
where
  red_beta: "x \<sharp> (N,\<tau>) \<Longrightarrow> App (Lam \<tau> x M) N \<leadsto> M[x\<mapsto>N]"
| red_beta_box: "u \<sharp> M \<Longrightarrow> LetBox \<tau> (Box M) u N \<leadsto> N{u\<mapsto>M}"
| red_insp: "Insp Ms \<leadsto> sfold (skel_of_trail q) Ms" (* non-deterministic inspection *)
| red_iinsp: "IInsp M \<theta> \<leadsto> sfold (skel_of_trail q) \<theta>" 
| red_repl: "\<theta> \<leadsto> \<theta>' \<Longrightarrow> Insp \<theta> \<leadsto> Insp \<theta>'"
| red_irepl1: "M \<leadsto> M' \<Longrightarrow> IInsp M \<theta> \<leadsto> IInsp M' \<theta>"
| red_irepl2: "\<theta> \<leadsto> \<theta>' \<Longrightarrow> IInsp M \<theta> \<leadsto> IInsp M \<theta>'"
| red_app1: "M \<leadsto> M' \<Longrightarrow> App M N \<leadsto> App M' N"
| red_app2: "N \<leadsto> N' \<Longrightarrow> App M N \<leadsto> App M N'"
| red_lam:  "M \<leadsto> M' \<Longrightarrow> Lam \<tau> x M \<leadsto> Lam \<tau> x M'"
| red_box:  "M \<leadsto> M' \<Longrightarrow> Box M \<leadsto> Box M'"
| red_letbox1: "\<lbrakk>M \<leadsto> M'; u \<sharp> (M,M')\<rbrakk> \<Longrightarrow> LetBox \<tau> M u N \<leadsto> LetBox \<tau> M' u N"
| red_letbox2: "\<lbrakk>N \<leadsto> N'; u \<sharp> M\<rbrakk> \<Longrightarrow> LetBox \<tau> M u N \<leadsto> LetBox \<tau> M u N'"
| red_nil: "M \<leadsto> M' \<Longrightarrow> ErNil M \<leadsto> ErNil M'"
| red_hd : "M \<leadsto> M' \<Longrightarrow> ErCons l M \<theta> \<leadsto> ErCons l M' \<theta>"
| red_tl : "\<theta> \<leadsto> \<theta>' \<Longrightarrow> ErCons l M \<theta> \<leadsto> ErCons l M \<theta>'"

inductive_cases red_cases[elim]:
"Insp \<theta> \<leadsto> M"
"IInsp N \<theta> \<leadsto> M"

inductive_cases red_repl_cases[elim]:
"ErNil M \<leadsto> \<theta>"
"ErCons l M \<theta> \<leadsto> \<theta>'"

lemma swap_inv:
  fixes x y :: var
  shows "[(x,y)] \<bullet> (M::exp) = [(y,x)] \<bullet> M"
  apply (subgoal_tac "[(x,y)]\<bullet>[(x,y)] \<bullet> (M::exp) = [(x,y)]\<bullet>[(y,x)] \<bullet> M \<Longrightarrow> [(x,y)] \<bullet> (M::exp) = [(y,x)] \<bullet> M")
  defer
  apply (simp add:perm_bij)
  apply (subgoal_tac "[(x,y)]\<bullet>[(x,y)] \<bullet> (M::exp) = [(x,y)]\<bullet>[(y,x)] \<bullet> M")
  apply (auto simp add:perm_swap)
  done

lemma uswap_inv:
  fixes x y :: uvar
  shows "[(x,y)] \<bullet> (M::exp) = [(y,x)] \<bullet> M"
  apply (subgoal_tac "[(x,y)]\<bullet>[(x,y)] \<bullet> (M::exp) = [(x,y)]\<bullet>[(y,x)] \<bullet> M \<Longrightarrow> [(x,y)] \<bullet> (M::exp) = [(y,x)] \<bullet> M")
  defer
  apply (simp add:perm_bij)
  apply (subgoal_tac "[(x,y)]\<bullet>[(x,y)] \<bullet> (M::exp) = [(x,y)]\<bullet>[(y,x)] \<bullet> M")
  apply (auto simp add:perm_swap)
  done

equivariance red
nominal_inductive red
  apply(auto intro: fresh_ty var_fresh_uctx uvar_fresh_ctx)
  apply(simp_all add:abs_fresh var_subst_fresh uvar_usubst_fresh)
  done

lemma red_fresh:
  fixes a::var
  fixes M N::exp
  fixes \<theta> \<theta>'::exp_repl
  shows "M \<leadsto> N \<Longrightarrow> a \<sharp> M \<Longrightarrow> a \<sharp> N"
  and "\<theta> \<leadsto> \<theta>' \<Longrightarrow> a \<sharp> \<theta> \<Longrightarrow> a \<sharp> \<theta>'"
  using assms
  apply(nominal_induct M N and \<theta> \<theta>' rule: red_red_repl.strong_inducts, auto simp:exp.fresh)
  apply(auto simp:subst_fresh usubst_fresh' abs_fresh sfold_fresh)
  done

lemma red_ufresh:
  fixes u::uvar
  fixes M N::exp
  fixes \<theta> \<theta>'::exp_repl
  shows "M \<leadsto> N \<Longrightarrow> u \<sharp> M \<Longrightarrow> u \<sharp> N"
  and "\<theta> \<leadsto> \<theta>' \<Longrightarrow> u \<sharp> \<theta> \<Longrightarrow> u \<sharp> \<theta>'"
  using assms
  apply(nominal_induct M N and \<theta> \<theta>' rule: red_red_repl.strong_inducts, auto simp:exp.fresh)
  apply(auto simp:subst_fresh' usubst_fresh abs_fresh sfold_fresh)
  done

lemma red_lam_inv: 
  "Lam \<tau> x M \<leadsto> M' \<Longrightarrow> \<exists>M''. M'=Lam \<tau> x M'' \<and> M \<leadsto> M''"
  proof (erule red.cases,simp_all add:exp.inject alpha,auto)
    fix R R'::exp
    fix \<sigma> y
    assume ha:"R \<leadsto> R'" and hb:"x \<sharp> R"
    show "\<exists>M''. R' = [(y, x)] \<bullet> M'' \<and> y \<sharp> M'' \<and> [(x, y)] \<bullet> R \<leadsto> M''"
    proof (rule exI[of _ "[(y,x)]\<bullet>R'"],auto simp:perm_swap)
      from ha have "[(y,x)]\<bullet>R \<leadsto> [(y,x)]\<bullet>R'" using eqvts by auto
      moreover from hb have "y \<sharp> [(y,x)]\<bullet>R" by (simp add:fresh_left swap_simps)
      ultimately show "y \<sharp> [(y,x)]\<bullet>R'" by (auto simp: red_fresh)
    next
      have "[(y, x)] \<bullet> R' = [(x, y)] \<bullet> R'" by (simp add:swap_inv) 
      from this show "[(x, y)] \<bullet> R \<leadsto> [(y, x)] \<bullet> R'" using ha eqvts by auto
    qed
  qed

lemma subst_exchange:
  fixes x::var
  fixes M::exp
  fixes Ms::exp_repl
  shows "x \<sharp> \<delta> \<Longrightarrow> x \<sharp> \<gamma> \<Longrightarrow> (M[x\<mapsto>N])<\<delta>,\<gamma>> = (M<\<delta>,\<gamma>>)[x\<mapsto>N<\<delta>,\<gamma>>]"
  and "x \<sharp> \<delta> \<Longrightarrow> x \<sharp> \<gamma> \<Longrightarrow> (Ms[x\<mapsto>N])<\<delta>,\<gamma>> = (Ms<\<delta>,\<gamma>>)[x\<mapsto>N<\<delta>,\<gamma>>]"
  apply(nominal_induct M and Ms avoiding: \<gamma> \<delta> x N rule:exp_exp_repl.strong_inducts)
  apply(auto simp add: lookup_default var_subst_id lookup_fresh ulookup_fresh fresh_atm
                       fresh_list_cons fresh_list_nil subst2_fresh)
  done

lemma usubst_exchange:
  fixes u::uvar
  fixes M::exp
  fixes Ms::exp_repl
  shows "u \<sharp> \<delta> \<Longrightarrow> u \<sharp> \<gamma> \<Longrightarrow> M{u\<mapsto>N}<\<delta>,\<gamma>> = M<\<delta>,\<gamma>>{u\<mapsto>N<\<delta>,\<gamma>>}"
  and "u \<sharp> \<delta> \<Longrightarrow> u \<sharp> \<gamma> \<Longrightarrow> Ms{u\<mapsto>N}<\<delta>,\<gamma>> = Ms<\<delta>,\<gamma>>{u\<mapsto>N<\<delta>,\<gamma>>}"
  apply(nominal_induct M and Ms avoiding: \<gamma> \<delta> u N rule:exp_exp_repl.strong_inducts)
  apply(auto simp add: ulookup_default uvar_subst_id lookup_fresh ulookup_fresh fresh_atm
                       fresh_list_cons fresh_list_nil subst2_fresh)
  done

lemma bedefault_subst[simp]:
  shows "(bedefault \<theta>)<S,T> = bedefault (\<theta><S,T>)"
  by (nominal_induct \<theta> rule:exp_exp_repl.strong_inducts(2),auto)

lemma belookup_subst_cases_aux:
  shows "(belookup \<theta> l = None \<and> belookup (\<theta><S,T>) l = None) \<or>
         (\<exists>M. belookup \<theta> l = Some M \<and> belookup (\<theta><S,T>) l = Some (M<S,T>))"
  by (nominal_induct \<theta> rule:exp_exp_repl.strong_inducts(2)[of "\<lambda>_ _. True"],simp_all)

lemma belookup_subst_cases:
  shows "(belookup \<theta> l = None \<Longrightarrow> belookup (\<theta><S,T>) l = None \<Longrightarrow> P)
         \<Longrightarrow> (\<And>M. belookup \<theta> l = Some M \<Longrightarrow> belookup (\<theta><S,T>) l = Some (M<S,T>) \<Longrightarrow> P)
         \<Longrightarrow> P"
using belookup_subst_cases_aux[of \<theta> l S T]
by (auto)

(* prove it for sfold instead *)
lemma qfold_subst[simp]:
  shows "(qfold q \<theta>)<S,T> = qfold q (\<theta><S,T>)"
  and "\<And>Mnil Lcons. (qfaux qs \<theta> Mnil Lcons)<S,T> = qfaux qs (\<theta><S,T>) (Mnil<S,T>) Lcons"
  apply (nominal_induct q and qs rule:trail_trail_repl.strong_inducts,auto simp:qfbranch_def)
  apply (cases \<theta> S T LRefl rule: belookup_subst_cases,auto)
  apply (cases \<theta> S T LTrans rule: belookup_subst_cases,auto)
  apply (cases \<theta> S T LBeta rule: belookup_subst_cases,auto)
  apply (cases \<theta> S T LBetaBox rule: belookup_subst_cases,auto)
  apply (cases \<theta> S T LITI rule: belookup_subst_cases,auto)
  apply (cases \<theta> S T LOTI rule: belookup_subst_cases,auto)
  apply (cases \<theta> S T LQLam rule: belookup_subst_cases,auto)
  apply (cases \<theta> S T LQApp rule: belookup_subst_cases,auto)
  apply (cases \<theta> S T LQLet rule: belookup_subst_cases,auto)
  apply (cases \<theta> S T LITBc rule: belookup_subst_cases,auto)
  apply (cases \<theta> S T LITBn rule: belookup_subst_cases,auto)
  apply (cases \<theta> S T LOTBn rule: belookup_subst_cases,auto)
proof-
  fix q N l
  assume "qfold q \<theta><S,T> = qfold q (\<theta><S,T>)"
  from this show "mybind (belookup \<theta> l) (\<lambda>M. exp.App (exp.App M (qfold q \<theta>)) N) (bedefault \<theta>)<S,T> =
       mybind (belookup (\<theta><S,T>) l) (\<lambda>M. exp.App (exp.App M (qfold q (\<theta><S,T>))) (N<S,T>)) (bedefault (\<theta><S,T>))"
  by (cases \<theta> S T l rule: belookup_subst_cases,auto)
next
  fix q qs N l
  assume "qfold q \<theta><S,T> = qfold q (\<theta><S,T>)"
         "(\<And>R l0. qfaux qs \<theta> R l0<S,T> = qfaux qs (\<theta><S,T>) (R<S,T>) l0)"
  from this 
  show "mybind (belookup \<theta> l) (\<lambda>M. exp.App (exp.App M (qfold q \<theta>)) (qfaux qs \<theta> N l))
        (bedefault \<theta>)<S,T> =
       mybind (belookup (\<theta><S,T>) l)
        (\<lambda>M. exp.App (exp.App M (qfold q (\<theta><S,T>))) (qfaux qs (\<theta><S,T>) (N<S,T>) l)) (bedefault (\<theta><S,T>))"
  by (cases \<theta> S T l rule: belookup_subst_cases,auto)
qed

(* Lemma 6 *)
lemma red_subst: 
  fixes M N::exp
  fixes \<theta> \<theta>'::exp_repl
  shows "M \<leadsto> N \<Longrightarrow> M<\<delta>,\<gamma>> \<leadsto> N<\<delta>,\<gamma>>"
  and "\<theta> \<leadsto> \<theta>' \<Longrightarrow> \<theta><\<delta>,\<gamma>> \<leadsto> \<theta>'<\<delta>,\<gamma>>"
  apply(nominal_induct M N and \<theta> \<theta>' avoiding: \<gamma> \<delta> rule: red_red_repl.strong_inducts)
  apply(auto simp add:subst_exchange usubst_exchange intro:red_red_repl.intros simp add:fresh_prod subst2_fresh)
  apply(subgoal_tac "(qfold q Ms)<\<delta>,\<gamma>> = qfold q (Ms<\<delta>,\<gamma>>)",simp add:qfold_def,auto intro: red_red_repl.intros)
  apply(subgoal_tac "(qfold q \<theta>)<\<delta>,\<gamma>> = qfold q (\<theta><\<delta>,\<gamma>>)",simp add:qfold_def,auto intro: red_red_repl.intros)
done

(* Neutral term (Def.9) *)
inductive nt :: "exp \<Rightarrow> bool" 
where
  var_nt: "nt (Var x)"
| uvar_nt: "nt (UVar u)"
| app_nt: "nt (App M N)"
| letbox_nt: "u \<sharp> M \<Longrightarrow> nt (LetBox \<tau> M u N)"
| insp_nt: "nt (Insp \<theta>)"
| iinsp_nt: "nt (IInsp M \<theta>)"

equivariance nt
nominal_inductive nt
  apply(simp add:exp.inject abs_fresh fresh_ty)
  done

definition NT :: "exp set"
where 
  "NT \<equiv> {M . nt M}"

lemma nt_eqvt :
  fixes \<pi>::"uvar prm"
  shows "\<pi>\<bullet>NT = NT"
apply (simp add: NT_def,auto simp add:nt.eqvt(2) pt_set_bij1a pt_uvar_inst at_uvar_inst)
apply (subgoal_tac "nt (\<pi> \<bullet> rev \<pi> \<bullet> x)", simp add:perm_pi_simp)
apply (auto simp add:nt.eqvt(2) pt_set_bij1a pt_uvar_inst at_uvar_inst)
done

lemma var_nt:
  shows "Var x \<in> NT"
  and "UVar u \<in> NT"
  apply(simp_all add: NT_def)
  apply(auto intro: nt.intros)
  done

definition NF :: "exp set"
where 
  "NF \<equiv> {M . (\<forall>N. M \<leadsto> N \<longrightarrow> False)}"

lemma var_nf:
  "Var x \<in> NF"
  apply(auto simp add:NF_def)
  apply(erule red.cases, simp_all)
  done

lemma uvar_nf:
  "UVar u \<in> NF"
  apply(auto simp add:NF_def)
  apply(erule red.cases, simp_all)
  done

(* Strongly normalizing terms *)
inductive sn :: "exp \<Rightarrow> bool" 
where 
  step_sn : "(\<And> N. M \<leadsto> N \<Longrightarrow> sn N) \<Longrightarrow> sn M"
equivariance sn
nominal_inductive sn
done

inductive sn_repl :: "exp_repl \<Rightarrow> bool" 
where 
  step_sn_repl : "(\<And> \<theta>'. \<theta> \<leadsto> \<theta>' \<Longrightarrow> sn_repl \<theta>') \<Longrightarrow> sn_repl \<theta>"
equivariance sn_repl
nominal_inductive sn_repl
done

lemma SN_preserved: 
  fixes t1 t2::exp
  assumes a: "sn t1" "t1 \<leadsto> t2"
  shows "sn t2"
using a 
by (cases) (auto)

lemma double_SN_aux:
  assumes a: "sn a"
  and b: "sn b"
  and hyp: "\<And>x z.
    \<lbrakk>\<And>y. x \<leadsto> y \<Longrightarrow> sn y; \<And>y. x \<leadsto> y \<Longrightarrow> P y z;
     \<And>u. z \<leadsto> u \<Longrightarrow> sn u; \<And>u. z \<leadsto> u \<Longrightarrow> P x u\<rbrakk> \<Longrightarrow> P x z"
  shows "P a b"
proof -
  from a
  have r: "\<And>b. sn b \<Longrightarrow> P a b"
  proof (induct a rule: sn.induct)
    case (step_sn x)
    note SNI' = step_sn
    have "sn b" by fact
    thus ?case
    proof (induct b rule: sn.induct)
      case (step_sn y)
      show ?case
        apply (rule hyp)
        apply (erule SNI')
        apply (erule SNI')
        apply (rule sn.step_sn)
        apply (erule step_sn)+
        done
    qed
  qed
  from b show ?thesis by (rule r)
qed

(* well-founded induction to reason on pairs of strongly normalizing terms *)
lemma double_SN[consumes 2]:
  fixes a b::exp and P::"exp \<Rightarrow> exp \<Rightarrow> bool"
  assumes ha: "sn a"
  and     hb: "sn b" 
  and     hc: "\<And>x z. \<lbrakk>\<And>y. x \<leadsto> y \<Longrightarrow> P y z; \<And>u. z \<leadsto> u \<Longrightarrow> P x u\<rbrakk> \<Longrightarrow> P x z"
  shows "P a b"
using ha hb hc
apply(rule_tac double_SN_aux)
apply(assumption)+
apply(blast)
done


definition SN :: "exp set"
where "SN \<equiv> {M . sn M}"

lemma nf_subseteq_sn: "NF \<subseteq> SN"
  by(auto simp add:NF_def SN_def intro:step_sn)

(* we require CRs to be equivariant in box variables
definition cr_eqvt :: "exp set \<Rightarrow> bool"
where
  "cr_eqvt C \<equiv> (\<forall> \<pi>::uvar prm. \<pi>\<bullet>C = C)"
*)

(* Definitions on candidates of reducibility (Sect. 4.1) *)
definition cr1 :: "exp set \<Rightarrow> bool"
where 
  "cr1 C \<equiv> C \<subseteq> SN"

definition cr2 :: "exp set \<Rightarrow> bool"
where 
  "cr2 C \<equiv> \<forall> M N. M \<in> C \<and> M \<leadsto> N \<longrightarrow> N \<in> C"

definition cr3 :: "exp set \<Rightarrow> bool"
where 
  "cr3 C \<equiv> \<forall> M. (M \<in> NT \<and> (\<forall> N. M \<leadsto> N \<longrightarrow> N \<in> C)) \<longrightarrow> M \<in> C"

(* Pre-candidates (Def.8) *)
definition PCR :: "exp set set"
where
  "PCR \<equiv> {C .  cr1 C \<and> cr2 C}"

lemma using_pcr1:
  "M \<in> C \<Longrightarrow> C \<in> PCR \<Longrightarrow> M \<in> SN"
  by(auto simp add: PCR_def cr1_def)

lemma using_pcr2:
  "M \<leadsto> N \<Longrightarrow> C \<in> PCR \<Longrightarrow> M \<in> C \<Longrightarrow> N \<in> C"
  by(auto simp add: PCR_def cr2_def)

lemma pcr_intro:
  "cr1 C \<Longrightarrow> cr2 C \<Longrightarrow> C \<in> PCR"
  by(auto simp add: PCR_def)

(* Candidates (Def.10) *)
definition CR :: "exp set set"
where
  "CR \<equiv> {C .  cr1 C \<and> cr2 C \<and> cr3 C}"

lemma using_cr1:
  "M \<in> C \<Longrightarrow> C \<in> CR \<Longrightarrow> M \<in> SN"
  by(auto simp add: CR_def cr1_def)

lemma using_cr2:
  "M \<leadsto> N \<Longrightarrow> C \<in> CR \<Longrightarrow> M \<in> C \<Longrightarrow> N \<in> C"
  by(auto simp add: CR_def cr2_def)

lemma using_cr3:
  "M \<in> NT \<Longrightarrow> C \<in> CR \<Longrightarrow> (\<forall> N. M \<leadsto> N \<longrightarrow> N \<in> C) \<Longrightarrow> M \<in> C"
  by(auto simp add: CR_def cr3_def)

lemma cr_intro:
  " cr1 C \<Longrightarrow> cr2 C \<Longrightarrow> cr3 C \<Longrightarrow> C \<in> CR"
  by(auto simp add: CR_def)

(* Lemma 8(1) *)
lemma sn_in_cr:
  "SN \<in> CR"
  apply(auto simp add:CR_def SN_def cr1_def cr2_def cr3_def)
  apply(erule sn.cases,simp)
  apply(rule sn.intros,simp)
  done

(* Lemma 8(3) *)
lemma var_in_c:
  shows "C \<in> CR \<Longrightarrow> Var x \<in> C"
  and "C \<in> CR \<Longrightarrow> UVar u \<in> C"
  apply(rule using_cr3, simp add:NT_def, rule nt.intros)
  apply(assumption)
  apply(auto)
  apply(erule red.cases,simp_all)
  apply(rule using_cr3, simp add:NT_def, rule nt.intros)
  apply(assumption)
  apply(auto)
  apply(erule red.cases,simp_all)
  done

(* Lemma 8(2) *)
lemma neut_nf_in_c:
  "C \<in> CR \<Longrightarrow> NT \<inter> NF \<subseteq> C"
  apply(auto)
  apply(erule using_cr3)
  apply(simp_all add:NF_def)  
  done

(* Substitutive subset of a candidate: Def.11 *)
definition repl :: "exp set \<Rightarrow> uvar \<Rightarrow> exp set \<Rightarrow> exp set"
where
"repl C u D \<equiv> {M . M \<in> C \<and> (\<forall> N. N \<in> D \<longrightarrow> (M{u\<mapsto>N}) \<in> C)}"

lemma repl_eqvt:
  fixes \<pi>::"uvar prm"
  shows "\<pi> \<bullet> repl C u D = repl (\<pi>\<bullet>C) (\<pi>\<bullet>u) (\<pi>\<bullet>D)"
apply(simp add:repl_def,rule,rule)
apply(simp add:pt_set_bij1a pt_uvar_inst at_uvar_inst)
apply(auto simp:subst_eqvt perm_pi_simp)
apply(simp add:pt_set_bij1a pt_uvar_inst at_uvar_inst)
apply(auto simp:subst_eqvt perm_pi_simp)
proof-
  fix M N
  assume "N \<in> D"
  from this have "rev \<pi> \<bullet> \<pi> \<bullet> N \<in> D" by (simp add:perm_pi_simp)
  moreover assume "\<forall>N. rev \<pi> \<bullet> N \<in> D \<longrightarrow> (rev \<pi> \<bullet> M){u\<mapsto>rev \<pi> \<bullet> N} \<in> C"
  ultimately have "(rev \<pi> \<bullet> M){u\<mapsto>rev \<pi> \<bullet> \<pi> \<bullet> N} \<in> C" by blast
  from this show "(rev \<pi> \<bullet> M){u\<mapsto>N} \<in> C" by (simp add:perm_pi_simp)
qed

lemma repl_in_pcr:
  "C \<in> CR \<Longrightarrow> repl C u D \<in> PCR"
  apply(rule pcr_intro)
  apply(simp add: cr1_def repl_def,clarify, erule using_cr1,simp)
  apply(simp add: cr2_def repl_def)
  apply(auto)
  apply(erule using_cr2,simp,simp)
  apply(erule_tac x="Na" in allE,auto)
  apply(rule using_cr2, erule red_subst,auto)
  done

definition RedFun :: "exp set \<Rightarrow> exp set \<Rightarrow> exp set"
where "RedFun C D \<equiv> {M. \<forall> N. N \<in> C \<longrightarrow> (App M N) \<in> D}"

lemma redfunI:
  "\<forall> N. N \<in> C \<longrightarrow> (App M N) \<in> D \<Longrightarrow> M \<in> RedFun C D"
  by (auto simp add:RedFun_def)

lemma redfunE:
  "M \<in> RedFun C D \<Longrightarrow> N \<in> C \<Longrightarrow> App M N \<in> D"
  by (auto simp add:RedFun_def)

definition RedBox :: "exp set \<Rightarrow> exp set"
where "RedBox D \<equiv> {M. \<forall> C u N \<tau>.  u \<sharp> M \<and> C \<in> CR \<and> N \<in> repl C u D  \<longrightarrow> LetBox \<tau> M u N \<in> C}"

lemma redboxE:
  " M \<in> RedBox D \<Longrightarrow>  u \<sharp> M \<Longrightarrow> C \<in> CR \<Longrightarrow> N \<in> repl C u D  \<Longrightarrow> LetBox \<tau> M u N \<in> C"
  by(simp add:RedBox_def)

lemma redboxI:
  "(\<forall> C u N \<tau>.   u \<sharp> M \<and> C \<in> CR \<and> N \<in> repl C u D  \<longrightarrow> LetBox \<tau> M u N \<in> C) \<Longrightarrow> M \<in> RedBox D"
  by(simp add:RedBox_def)

(* Reducible sets (Def. 12) *)
nominal_primrec Red :: "ty \<Rightarrow> exp set"
where
  "Red PTy = SN"
| "Red (FunTy \<tau>1 \<tau>2) = RedFun (Red \<tau>1) (Red \<tau>2)"
| "Red (BoxTy \<tau>) = RedBox (Red \<tau>)"
by auto

inductive_set Red_repl :: "ty \<Rightarrow> exp_repl set" for \<tau>::ty
where
  rr_ernil [intro]: "M \<in> Red \<tau> \<Longrightarrow> ErNil M \<in> Red_repl \<tau>"
| rr_ercons [intro]: "M \<in> Red (bty \<tau> l) \<Longrightarrow> Ns \<in> Red_repl \<tau> \<Longrightarrow> ErCons l M Ns \<in> Red_repl \<tau>"

lemma app_subterm_sn_IH:
  "sn (App M N) \<Longrightarrow> sn M"
  apply(nominal_induct M\<equiv>"App M N" arbitrary: M rule:sn.strong_induct)
  apply(rule sn.intros)
  apply(blast intro:red_red_repl.intros)
  done

lemma app_subterm_sn:
  "App M N \<in> SN \<Longrightarrow> M \<in> SN"
  apply(simp add:SN_def)
  by(erule app_subterm_sn_IH)
  
lemma red_letbox1_strong:
  fixes M N Q :: exp
  shows "M \<leadsto> Q \<Longrightarrow> LetBox \<tau> M u N \<leadsto> LetBox \<tau> Q u N"
  apply(subgoal_tac "\<exists> v::uvar. v \<sharp> (M,u,N,Q)",clarify)
  apply(subgoal_tac "LetBox \<tau> M u N = LetBox \<tau> M v ([(v,u)]\<bullet>N) \<and> LetBox \<tau> Q u N = LetBox \<tau> Q v ([(v,u)]\<bullet>N)")
  apply(clarsimp)
  apply(erule red_red_repl.intros)
  apply(simp)
  apply(auto)
  apply(simp add:exp.inject, rule sym, simp add:alpha fresh_atm fresh_prod)+
  apply(rule exists_fresh,auto,finite_guess)
  done

lemma red_letbox2_strong:
  fixes M N Q :: exp
  shows "N \<leadsto> Q \<Longrightarrow> LetBox \<tau> M u N \<leadsto> LetBox \<tau> M u Q"
  apply(subgoal_tac "\<exists> v::uvar. v \<sharp> (M,u,N,Q)",clarify)
  apply(subgoal_tac "LetBox \<tau> M u N = LetBox \<tau> M v ([(v,u)]\<bullet>N) \<and> LetBox \<tau> M u Q = LetBox \<tau> M v ([(v,u)]\<bullet>Q)")
  apply(clarsimp)
  apply(rule red_red_repl.intros,auto simp:eqvts)
  apply(simp add:exp.inject, rule sym, simp add:alpha fresh_atm fresh_prod)+
  apply(rule exists_fresh,auto,finite_guess)
  done

lemma letbox_subterm_sn_IH:
  " sn (LetBox \<tau> M u N) \<Longrightarrow> sn M "
  apply(nominal_induct M\<equiv>"LetBox \<tau> M u N" arbitrary: M rule:sn.strong_induct)
  apply(rule sn.intros)
  apply(drule red_letbox1_strong,auto simp add: fresh_prod)
  done

lemma letbox_subterm_sn:
  "LetBox \<tau> M u N \<in> SN \<Longrightarrow> M \<in> SN"
  apply(simp add:SN_def) 
  by(erule letbox_subterm_sn_IH)
  



lemma red_p_cr:
  shows "Red PTy \<in> CR"
  apply(simp_all)
  by(rule sn_in_cr)
  
lemma red_fun_cr1:
  shows "C \<in> CR \<Longrightarrow> D \<in> CR \<Longrightarrow> cr1 (RedFun C D)"
  apply(auto simp add:cr1_def RedFun_def)
  apply(subgoal_tac "\<exists> y::var . True",clarify)
  apply(erule_tac x="Var y" in allE,simp_all)
  apply(rule using_cr1)
  apply(rule app_subterm_sn)
  apply(rule using_cr1, erule mp, simp_all add: sn_in_cr)
  apply(erule var_in_c)
  done

lemma  red_fun_cr2:
  "C \<in> CR \<Longrightarrow> D \<in> CR \<Longrightarrow> cr2 (RedFun C D)"
  apply(auto simp add:cr2_def)
  apply(simp add:RedFun_def,auto)
  apply(rule using_cr2, simp_all)
  apply(erule red_app1, erule_tac x="Na" in allE,simp)
done

lemma nt_not_lam: "Lam ty x M \<in> NT = False"
  by(auto elim:nt.cases simp add: NT_def)

(*
lemma cr3_app :
  "C \<in> CR \<Longrightarrow> M \<in> NT \<Longrightarrow> (\<forall>R. App M N \<leadsto> R \<longrightarrow> R \<in> C) \<Longrightarrow> App M N \<in> C"
  apply(rule using_cr3, auto simp:NT_def nt.intros)
  done

lemma cr3_app1 :
  "C \<in> CR \<Longrightarrow> M \<in> NT \<Longrightarrow> (\<forall>R. M \<leadsto> R \<longrightarrow> App R N \<in> C) \<Longrightarrow> (\<forall>S. N \<leadsto> S \<longrightarrow> App M S \<in> C) \<Longrightarrow> App M N \<in> C"
  apply(rule cr3_app,auto)
  apply(rule red.cases,auto simp:exp.inject nt_not_lam)
  done
*)

lemma  red_fun_cr3_aux:
  "R \<in> SN \<Longrightarrow> C \<in> CR \<Longrightarrow> D \<in> CR \<Longrightarrow> M \<in> NT \<Longrightarrow>
   (\<forall>N. M \<leadsto> N \<longrightarrow> N \<in> RedFun C D) \<Longrightarrow> R \<in> C \<longrightarrow> App M R \<in> D"
  apply(simp add:SN_def, induct rule: sn.induct,clarify)
  apply(rule using_cr3, auto simp: NT_def nt.intros)
  apply(erule red.cases, auto)
  apply(auto simp:exp.inject, erule nt.cases, auto)
  apply(auto simp:redfunE)
  apply(subgoal_tac "N' \<in> C", auto)
  apply(auto simp:using_cr2)
  done

lemma  red_fun_cr3:
  "C \<in> CR \<Longrightarrow> D \<in> CR \<Longrightarrow> cr3 (RedFun C D)"
  apply(auto simp add: cr3_def)
  apply(rule redfunI,clarify)
  apply(subgoal_tac "N \<in> SN")
  apply(auto simp: red_fun_cr3_aux)
  apply(auto simp:using_cr1)
  done

lemma red_fun_cr:
  "Red A \<in> CR \<Longrightarrow> Red B \<in> CR \<Longrightarrow> Red (FunTy A B) \<in> CR"
  apply(auto intro: cr_intro red_fun_cr1 red_fun_cr2 red_fun_cr3)
  done

lemma sn_eqvt_IH:
  fixes \<pi>::"uvar prm"
  shows "sn (\<pi> \<bullet> M) \<Longrightarrow> sn M"
  apply(induct M\<equiv>"\<pi>\<bullet>M" arbitrary: M rule:sn.induct)
  apply(rule sn.intros)
  apply(simp add:eqvts)
done

lemma eqvt_sn_base:
  fixes u v::"uvar"
  shows "sn M \<Longrightarrow> sn ([(u,v)] \<bullet> M)"
  apply(induct M rule:sn.induct)
  apply(rule sn.intros)
  apply(subgoal_tac "sn ([(u,v)]\<bullet>[(u,v)]\<bullet>N)", simp add:perm_swap)
  apply(subgoal_tac "M \<leadsto> [(u,v)]\<bullet>N",simp)
  apply(subgoal_tac "[(u,v)]\<bullet>[(u,v)]\<bullet>M \<leadsto> [(u,v)]\<bullet>N",simp add:perm_swap)
  apply(simp add:eqvts)
done

lemma SN_eqvt_base:
  fixes a b::"uvar"
  shows "[(a,b)]\<bullet> SN = SN"
  apply(auto simp add:SN_def perm_set_def)
  apply(erule eqvt_sn_base)
  apply(rule_tac x="[(a,b)]\<bullet>x" in exI, simp add:perm_swap)
  apply(erule eqvt_sn_base)
done

(*
lemma sn_lam : "sn M \<Longrightarrow> sn (Lam \<tau> x M)"
  proof (induct rule:sn.induct, rule sn.intros)
  fix M N
  assume hred:"Lam \<tau> x M \<leadsto> N"
  assume ih:"\<And>R. M \<leadsto> R \<Longrightarrow> sn (Lam \<tau> x R)"
  from this have "\<exists>N'. N = Lam \<tau> x N'" sorry
  from this obtain N' where "N = Lam \<tau> x N'" by blast
  from hred this ih show "sn N"
  proof (simp, cases,auto simp:exp.inject alpha)
    fix xa M M'
    assume hred:"M \<leadsto> M'"
    assume "(\<And>R. [(x, xa)] \<bullet> M \<leadsto> R \<Longrightarrow> sn (Lam \<tau> x R))"
    moreover from hred have "[(x, xa)] \<bullet> M \<leadsto> [(x, xa)] \<bullet> M'" by (simp add:eqvts)
    ultimately show "sn (Lam \<tau> x ([(x, xa)] \<bullet> M'))" by auto
  qed
qed

lemma sn_app : "sn M \<Longrightarrow> sn N \<Longrightarrow> sn (App M N)"
  proof (rule double_SN[of M N],auto,rule sn.intros)
  fix M' N' R
  assume hred:"App M' N' \<leadsto> R"
  assume ih:"\<And>M''. M' \<leadsto> M'' \<Longrightarrow> sn (App M'' N')" "\<And>N''. N' \<leadsto> N'' \<Longrightarrow> sn (App M' N'')"
  from this have "\<exists>N'. N = Lam \<tau> x N'" sorry
  from this obtain N' where "N = Lam \<tau> x N'" by blast
  from hred this ih show "sn N"
  proof (simp, cases,auto simp:exp.inject alpha)
    fix xa M M'
    assume hred:"M \<leadsto> M'"
    assume "(\<And>R. [(x, xa)] \<bullet> M \<leadsto> R \<Longrightarrow> sn (Lam \<tau> x R))"
    moreover from hred have "[(x, xa)] \<bullet> M \<leadsto> [(x, xa)] \<bullet> M'" by (simp add:eqvts)
    ultimately show "sn (Lam \<tau> x ([(x, xa)] \<bullet> M'))" by auto
  qed
qed
*)

lemma exp_swap_self[simp]:
  fixes x::var and M::exp and Ms::exp_repl
  shows "[(x,x)]\<bullet>M = M"
  and "[(x,x)]\<bullet>Ms = Ms"
  apply(nominal_induct M and Ms rule:exp_exp_repl.strong_inducts)
  apply(auto simp:exp.inject calc_atm swap_trail_label ty_eqvt)
  done

lemma lam_rename:
  fixes x y::var
  assumes "y \<sharp> M"
  shows "Lam \<tau> x M = Lam \<tau> y ([(x,y)]\<bullet>M)"
  using assms
  apply (auto simp add: exp.inject alpha)
  defer
  apply (subgoal_tac "([(x,y)]\<bullet>y) \<sharp> [(x,y)]\<bullet>M",simp add:calc_atm)
  apply (simp add:fresh_bij)
  apply (auto simp:perm_swap)
  apply (subgoal_tac "([(x,y)]\<bullet>y) \<sharp> [(x,y)]\<bullet>M",subgoal_tac "[(x,y)]\<bullet>y = x",auto simp:fresh_bij)
  apply (simp add:calc_atm)
  done

lemma SN_eqvt:
  fixes \<pi>::"uvar prm"
  shows "\<pi>\<bullet> SN = SN"
proof(induct \<pi>, simp,clarify)
  fix u v :: uvar and \<pi> :: "uvar prm"
  assume ih:"\<pi>\<bullet>SN = SN"
  have "((u,v)#\<pi>)\<bullet>SN = ([(u,v)] @ \<pi>)\<bullet>SN" by simp
  also have "... = [(u,v)]\<bullet>\<pi>\<bullet>SN" by (rule pt2[of "[(u,v)]" \<pi> SN], auto simp:pt_uvar_inst)
  finally show "((u, v) # \<pi>) \<bullet> SN = SN" by (auto simp:ih SN_eqvt_base)
qed

lemma cr1_eqvt:
  fixes \<pi>::"uvar prm"
  assumes "cr1 C"
  shows "cr1 (\<pi>\<bullet>C)"
proof (simp add:cr1_def,auto simp add: pt_set_bij1a pt_uvar_inst at_uvar_inst)
  fix M
  assume "rev \<pi> \<bullet> M \<in> C"
  from assms this have "rev \<pi> \<bullet> M \<in> SN" by (auto simp add:cr1_def)
  from this show "M \<in> SN" by (simp add:pt_set_bij1 pt_uvar_inst at_uvar_inst SN_eqvt)
qed

lemma cr2_eqvt:
  fixes \<pi>::"uvar prm"
  assumes "cr2 C"
  shows "cr2 (\<pi>\<bullet>C)"
proof (simp add:cr2_def,auto simp add: pt_set_bij1a pt_uvar_inst at_uvar_inst)
  fix M N
  assume hM:"rev \<pi> \<bullet> M \<in> C"
  assume "M \<leadsto> N"
  from this have hred:"rev \<pi> \<bullet> M \<leadsto> rev \<pi> \<bullet> N" by (rule red.eqvt)
  from assms hM this show "rev \<pi> \<bullet> N \<in> C" by (auto simp add:cr2_def)
qed

lemma cr3_eqvt:
  fixes \<pi>::"uvar prm"
  assumes "cr3 C"
  shows "cr3 (\<pi>\<bullet>C)"
using assms
proof (simp add:cr3_def,auto simp add: pt_set_bij1a pt_uvar_inst at_uvar_inst)
  fix M N
  assume "M \<in> NT"
  from this have hM:"rev \<pi> \<bullet> M \<in> NT" by (auto simp:nt_eqvt pt_set_bij1 pt_uvar_inst at_uvar_inst)
  assume hcr:"\<forall>M. M \<in> NT \<and> (\<forall>N. M \<leadsto> N \<longrightarrow> N \<in> C) \<longrightarrow> M \<in> C"
  assume hred:"\<forall>N. M \<leadsto> N \<longrightarrow> rev \<pi> \<bullet> N \<in> C"
  have hred':"\<forall>N. rev \<pi>\<bullet>M \<leadsto> N \<longrightarrow> N \<in> C"
  proof(auto)
    fix N
    assume "rev \<pi> \<bullet> M \<leadsto> N"
    from this have "\<pi>\<bullet>rev \<pi>\<bullet>M \<leadsto> \<pi>\<bullet>N" by (auto simp:red.eqvt(2)[of "rev \<pi> \<bullet> M" N \<pi>])
    from this have "M \<leadsto> \<pi>\<bullet>N" by (simp add:perm_pi_simp)
    from this hred show "N \<in> C" by (auto simp add:perm_pi_simp)
  qed
  from hcr hM this show "rev \<pi> \<bullet> M \<in> C" by blast
qed

lemma cr_eqvt :
  fixes \<pi>::"uvar prm"
  shows "\<pi>\<bullet>CR = CR"
apply (simp add:CR_def,rule)
defer
apply (rule,simp add: pt_set_bij1a pt_uvar_inst at_uvar_inst cr1_eqvt cr2_eqvt cr3_eqvt)
apply (rule,subgoal_tac "cr1 (\<pi>\<bullet>rev \<pi>\<bullet>x) \<and> cr2 (\<pi>\<bullet>rev \<pi>\<bullet>x) \<and> cr3 (\<pi>\<bullet>rev \<pi>\<bullet>x)",simp add:perm_pi_simp)
apply (auto simp add: pt_set_bij1a pt_uvar_inst at_uvar_inst cr1_eqvt cr2_eqvt cr3_eqvt)
done

lemma redfun_eqvt:
  fixes \<pi>::"uvar prm"
  shows "\<pi> \<bullet> RedFun C D = RedFun (\<pi>\<bullet>C) (\<pi>\<bullet>D)"
proof (auto, rule redfunI,auto)
  fix M N
  assume "M \<in> \<pi> \<bullet> RedFun C D"
  from this have hM:"rev \<pi> \<bullet> M \<in> RedFun C D" by (simp add: pt_set_bij1a pt_uvar_inst at_uvar_inst)
  assume hN:"N \<in> \<pi> \<bullet> C"
  from this have hN:"rev \<pi> \<bullet> N \<in> C" by (simp add: pt_set_bij1a pt_uvar_inst at_uvar_inst)
  show "App M N \<in> \<pi> \<bullet> D"
    by (simp add:pt_set_bij1a pt_uvar_inst at_uvar_inst,rule redfunE[OF hM hN])
next show "\<And>x. x \<in> RedFun (\<pi> \<bullet> C) (\<pi> \<bullet> D) \<Longrightarrow> x \<in> \<pi> \<bullet> RedFun C D"
proof (simp add: pt_set_bij1a pt_uvar_inst at_uvar_inst, rule redfunI,auto)
  fix M N
  assume hM:"M \<in> RedFun (\<pi> \<bullet> C) (\<pi> \<bullet> D)"
  assume "N \<in> C"
  from this have hN: "\<pi>\<bullet>N \<in> \<pi>\<bullet>C" by (simp add:pt_set_bij pt_uvar_inst at_uvar_inst)
  have "App M (\<pi>\<bullet>N) \<in> \<pi>\<bullet>D" by (rule redfunE[OF hM hN])
  from this show "App (rev \<pi> \<bullet> M) N \<in> D" 
    by (auto simp add:pt_set_bij1a pt_uvar_inst at_uvar_inst perm_pi_simp)
  qed
qed

lemma redbox_eqvt:
  fixes \<pi>::"uvar prm"
  shows "\<pi> \<bullet> RedBox C = RedBox (\<pi>\<bullet>C)"
proof (auto, rule redboxI,auto)
  fix M N D u \<tau>
  assume "M \<in> \<pi> \<bullet> RedBox C"
  from this have hM:"rev \<pi> \<bullet> M \<in> RedBox C" by (simp add: pt_set_bij1a pt_uvar_inst at_uvar_inst)
  assume "(u::uvar) \<sharp> M"
  from this have hu:"(rev \<pi> \<bullet> u) \<sharp> rev \<pi> \<bullet> M" by (auto simp: pt_fresh_bij1 pt_uvar_inst at_uvar_inst)
  assume "D \<in> CR"
  from this have hD:"rev \<pi>\<bullet>D \<in> CR" by (simp add: pt_set_bij1 pt_uvar_inst at_uvar_inst cr_eqvt)
  assume "N \<in> repl D u (\<pi> \<bullet> C)"
  from this have hN:"rev \<pi>\<bullet>N \<in> repl (rev \<pi>\<bullet>D) (rev \<pi>\<bullet>u) C" 
    by (simp add:pt_set_bij1 pt_uvar_inst at_uvar_inst repl_eqvt perm_pi_simp)
  from hM hu hD hN redboxE have "LetBox \<tau> (rev \<pi>\<bullet>M) (rev \<pi>\<bullet>u) (rev \<pi>\<bullet>N) \<in> rev \<pi> \<bullet> D" by blast
  from this have "rev \<pi> \<bullet> (LetBox \<tau> M u N) \<in> rev \<pi> \<bullet> D" by (simp add: ty_eqvt)
  moreover have "(rev \<pi> \<bullet> (LetBox \<tau> M u N) \<in> rev \<pi> \<bullet> D) = (LetBox \<tau> M u N \<in> D)" 
    by (rule pt_set_bij[of "rev \<pi>" "LetBox \<tau> M u N" D], auto simp add: pt_uvar_inst at_uvar_inst)
  ultimately show "LetBox \<tau> M u N \<in> D" by simp
next
show "\<And>x. x \<in> RedBox (\<pi> \<bullet> C) \<Longrightarrow> x \<in> \<pi> \<bullet> RedBox C"
proof (simp add: pt_set_bij1a pt_uvar_inst at_uvar_inst, rule redboxI, auto)
  fix M N D u \<tau>
  assume hM:"M \<in> RedBox (\<pi>\<bullet>C)"
  assume "(u::uvar) \<sharp> rev \<pi> \<bullet> M"
  from this have "(\<pi>\<bullet> u)\<sharp>(\<pi>\<bullet>rev \<pi>\<bullet>M)" by (auto simp: pt_fresh_bij1 pt_uvar_inst at_uvar_inst)
  from this have hu:"(\<pi> \<bullet> u) \<sharp> M" by (simp add:perm_pi_simp)
  assume "D \<in> CR"
  from this have hD:"\<pi> \<bullet> D \<in> CR" by (simp add: pt_set_bij1 pt_uvar_inst at_uvar_inst cr_eqvt)
  assume "N \<in> repl D u C"
  from this have hN:"(\<pi>\<bullet>N) \<in> repl (\<pi>\<bullet>D) (\<pi>\<bullet>u) (\<pi>\<bullet>C)"
    by (simp add:pt_set_bij1 pt_uvar_inst at_uvar_inst repl_eqvt perm_pi_simp)
  from hM hu hD hN redboxE have "LetBox \<tau> M (\<pi>\<bullet>u) (\<pi>\<bullet>N) \<in> \<pi>\<bullet>D" by blast
  from this have "\<pi> \<bullet> (LetBox \<tau> (rev \<pi>\<bullet>M) u N) \<in> \<pi> \<bullet> D" by (simp add:perm_pi_simp ty_eqvt)
  moreover have "(\<pi> \<bullet> (LetBox \<tau> (rev \<pi>\<bullet>M) u N) \<in> \<pi> \<bullet> D) = (LetBox \<tau> (rev \<pi>\<bullet>M) u N \<in> D)" 
    by (rule pt_set_bij[of \<pi> "LetBox \<tau> (rev \<pi>\<bullet>M) u N" D], auto simp add: pt_uvar_inst at_uvar_inst)
  ultimately show "LetBox \<tau> (rev \<pi>\<bullet>M) u N \<in> D" by simp
  qed
qed

lemma red_eqvt:
  shows "\<And>\<pi>::uvar prm. \<pi> \<bullet> Red A = Red A"
  apply(nominal_induct A rule:ty.strong_induct)
  apply(simp_all add:SN_eqvt redfun_eqvt redbox_eqvt)
done

(*
lemma alpha_helper: "v \<sharp>  (N,Na) \<Longrightarrow> u \<sharp> N \<Longrightarrow> LetBox \<tau> N v ([(u, v)] \<bullet> Na) = LetBox \<tau> N u Na"
sorry
*)

lemma nt_not_box: "Box M \<in> NT = False"
  by(auto elim:nt.cases simp add: NT_def)

lemma sn_sub_induction:
       "sn N \<Longrightarrow> Red A \<in> CR \<Longrightarrow>
       M \<in> NT \<Longrightarrow>
       \<forall>N. M \<leadsto> N \<longrightarrow> N \<in> RedBox (Red A) \<Longrightarrow>
       u \<sharp> M \<Longrightarrow> C \<in> CR \<Longrightarrow> N \<in> repl C u (Red A) \<Longrightarrow>  LetBox \<tau> M u N \<in> C"
  apply(induct N rule:sn.induct)
  apply(rule using_cr3, simp add:NT_def nt.intros)
  apply(simp)
  apply(auto)
  apply(erule red.cases)
  apply(simp_all add: exp.inject nt_not_box)
  apply(subgoal_tac "LetBox \<tau>' M' ua Na = LetBox \<tau>' M' u Ma",simp)
  apply(subgoal_tac "M' \<in> RedBox (Red A)")
  apply(subgoal_tac "u \<sharp> M'")
  apply(rule redboxE,auto)
  apply(auto simp:red_ufresh)
  apply(auto simp add:alpha exp.inject pt_swap_bij' pt_uvar_inst at_uvar_inst pt_fresh_left at_calc)
  apply(subgoal_tac "N' \<in> repl C u (Red A)",auto)
  apply(rule using_pcr2, auto simp:repl_in_pcr)
  apply(subgoal_tac "LetBox \<tau> M ua N' = LetBox \<tau> M u ([(ua,u)]\<bullet>N')",simp)
  apply(subgoal_tac "[(u, ua)] \<bullet> Na \<leadsto> [(ua, u)] \<bullet> N'")
  apply(subgoal_tac "[(ua, u)] \<bullet> N' \<in> repl C u (Red A)",auto)
  apply(rule using_pcr2, auto simp:repl_in_pcr)
  apply(subgoal_tac "[(ua, u)] \<bullet> N' = [(ua, u)] \<bullet> [(ua, u)] \<bullet> [(u,ua)] \<bullet> N'")
  apply(simp add:pt_swap_bij pt_uvar_inst at_uvar_inst red.eqvt)
  apply(simp add:pt_swap_bij' pt_uvar_inst at_uvar_inst)
  apply(simp add:exp.inject alpha pt_swap_bij pt_fresh_left pt_uvar_inst at_uvar_inst at_calc red_ufresh)
done

lemma red_box_cr1:
  "Red A \<in> CR \<Longrightarrow>  cr1 (RedBox (Red A))"
  apply(auto simp add:cr1_def )
  apply(subgoal_tac "(\<exists> u::uvar. u \<sharp> x)",clarify)
  apply(rule_tac u="u" and N="UVar u" in letbox_subterm_sn)
  apply(erule redboxE)  
  apply(simp)
  apply(simp add:sn_in_cr)
  apply(simp add: repl_def, rule conjI)
  apply(rule var_in_c, simp add:sn_in_cr) 
  apply(clarsimp)
  apply(erule using_cr1,simp)  
  apply(rule exists_fresh,auto,finite_guess)
  done

lemma  red_box_cr2:
  "Red A \<in> CR \<Longrightarrow> cr2 (RedBox (Red A))"
  apply(auto simp add:cr2_def )
  apply(rule redboxI,auto)
  apply(subgoal_tac "(\<exists> v::uvar. v \<sharp> (M,N,Na))",clarify)
  apply(subgoal_tac "LetBox \<tau> N v ([(u,v)]\<bullet>Na) \<in> [(u,v)]\<bullet>C")
  defer
  apply(rule using_cr2,simp_all)
  apply(rule red_letbox1_strong,simp)
  apply(simp add:pt_set_bij1 pt_uvar_inst at_uvar_inst cr_eqvt)
  apply(rule redboxE,simp_all)
  apply(simp add:pt_set_bij1 pt_uvar_inst at_uvar_inst cr_eqvt)
  apply(simp add:pt_set_bij1 pt_uvar_inst at_uvar_inst pt_swap_bij repl_eqvt red_eqvt at_swap_simps)
  apply(rule exists_fresh,auto,finite_guess)
  apply(subgoal_tac "[(u,v)]\<bullet>LetBox \<tau> N v ([(u, v)] \<bullet> Na) \<in> [(u,v)]\<bullet> [(u,v)] \<bullet> C")
  apply(simp add:pt_uvar_inst at_uvar_inst pt_fresh_fresh pt_swap_bij at_swap_simps ty_eqvt)
  apply(rule pt_set_bij2)
  apply(auto simp: pt_uvar_inst at_uvar_inst)
done

lemma  red_box_cr3:
  "Red A \<in> CR \<Longrightarrow> cr3 (RedBox (Red A))"
  apply(auto simp add: cr3_def )
  apply(rule redboxI,clarify)
  apply(subgoal_tac "N \<in> SN",simp add:SN_def,erule sn_sub_induction,auto)
  apply(erule using_pcr1)
  apply(erule repl_in_pcr)
  done

lemma red_box_cr: 
  "Red A \<in> CR \<Longrightarrow> Red (BoxTy A) \<in> CR"
  apply(auto intro: cr_intro red_box_cr1 red_box_cr2 red_box_cr3)
done

(* Lemma 10 *)
lemma red_cr:
  shows "Red \<tau> \<in> CR"
  apply(nominal_induct rule:ty.strong_induct)
  apply(rule red_p_cr)
  apply(erule red_fun_cr, simp)
  apply(erule red_box_cr)
  done

lemma var_is_red:
  "Var x \<in> Red \<tau>"
  apply(auto simp:var_in_c red_cr)
  done

lemma uvar_is_red:
  "UVar x \<in> Red \<tau>"
  apply(auto simp:var_in_c red_cr)
  done

(* adapted interpretations (Def. 13) are split into simple interpretations (red_ctx) and audited
   interpretations (red_uctx), each defined by means of inductive rules, but are otherwise the same 
   as in the paper *)
inductive
  red_ctx :: "ctx \<Rightarrow> subst \<Rightarrow>  bool"   
where
  red_ctx_nil:  "red_ctx [] []"
| red_ctx_cons: "\<lbrakk>M \<in> Red \<tau>; red_ctx \<Gamma> \<gamma>;y \<sharp> \<Gamma> \<rbrakk> \<Longrightarrow> red_ctx ((y,\<tau>)#\<Gamma>) ((y,M)#\<gamma>)"

lemma lookup_is_red:
  "red_ctx \<Gamma> \<gamma> \<Longrightarrow> (x,\<tau>) \<in> set \<Gamma> \<Longrightarrow> lookup \<gamma> x \<in> Red \<tau>"
  apply(induct \<Gamma> \<gamma> rule:red_ctx.induct)
  apply(auto, frule fresh_var_not_in,auto)
  done

lemma red_ctx_valid:
  "red_ctx \<Gamma> \<gamma> \<Longrightarrow> valid \<Gamma>"
  apply(induct \<Gamma> \<gamma> rule:red_ctx.induct)
  apply(auto intro:valid.intros)
  done

fun
  Red_ctx :: "ctx \<Rightarrow> subst set"   
where
  "Red_ctx \<Gamma>        = { \<gamma> . red_ctx \<Gamma> \<gamma>}"


inductive
  red_uctx :: "uctx \<Rightarrow> usubst \<Rightarrow>  bool"   
where
  red_uctx_nil:  "red_uctx [] []"
| red_uctx_cons: "\<lbrakk>M \<in> Red \<tau>; red_uctx \<Delta> \<delta>;u \<sharp> \<Delta> \<rbrakk> \<Longrightarrow> red_uctx ((u,\<tau>)#\<Delta>) ((u,M)#\<delta>)"

lemma ulookup_is_red:
  "red_uctx \<Delta> \<delta> \<Longrightarrow> (u,\<tau>) \<in> set \<Delta> \<Longrightarrow> ulookup \<delta> u \<in> Red \<tau>"
  apply(induct \<Delta> \<delta> rule:red_uctx.induct)
  apply(auto, frule fresh_uvar_not_in,auto)
  done

lemma red_uctx_valid:
  "red_uctx \<Delta> \<delta> \<Longrightarrow> uvalid \<Delta>"
  apply(induct \<Delta> \<delta> rule:red_uctx.induct)
  apply(auto intro:uvalid.intros)
  done


fun
  Red_uctx :: "ctx \<Rightarrow> subst set"   
where
  "Red_uctx \<Delta>        = { \<delta> . red_ctx \<Delta> \<delta>}"

lemma app_is_nt: "App M N \<in> NT"
  by (simp add:NT_def nt.intros)

lemma subst_same : 
  fixes M::exp and Ms::exp_repl
  shows "M[x\<mapsto>Var x] = M"
  and "Ms[x\<mapsto>Var x] = Ms"
  by (nominal_induct M and Ms avoiding:x rule:exp_exp_repl.strong_inducts,auto)

lemma fresh_uvar_swap:
  fixes u::uvar
    and x y::var
    and M::exp
    and Ms::exp_repl
  shows "(u\<sharp>[(x,y)]\<bullet>M) = u \<sharp> M"
  and "(u\<sharp>[(x,y)]\<bullet>Ms) = u \<sharp> Ms"
  apply(nominal_induct M and Ms rule:exp_exp_repl.strong_inducts)
  apply(auto simp add:calc_atm fresh_atm abs_fresh fresh_trail_label swap_trail_label ty_eqvt)
  done

(*
lemma subst_letbox_simp[simp]:
  "u \<sharp> (M,Q) \<Longrightarrow> (LetBox \<tau> M u N)[y\<mapsto>Q] = LetBox \<tau> (M[y\<mapsto> Q]) u (N[y\<mapsto>Q])"  
*)
lemma subst_rename:
  fixes M::exp and Ms::exp_repl
  shows "y\<sharp>M \<Longrightarrow> subst1 M x N = ([(y,x)]\<bullet>M)[y\<mapsto>N]"
  and "y\<sharp>Ms \<Longrightarrow> subst1_repl Ms x N = ([(y,x)]\<bullet>Ms)[y\<mapsto>N]"
  apply (nominal_induct M and Ms avoiding: x y N rule: exp_exp_repl.strong_inducts)
  apply (auto simp add: calc_atm fresh_atm abs_fresh fresh_uvar_swap swap_trail_label ty_eqvt)
  done

lemma red_lam: 
  assumes hlam: "\<forall> N. N \<in> Red A  \<longrightarrow> subst1 M x N \<in> Red B"
  shows "Lam A x M \<in> RedFun (Red A) (Red B)"
proof -
  have b1: "sn M" 
  proof -
    have "Var x\<in>Red A" by (auto simp:var_in_c red_cr)
    then have "M[x\<mapsto>Var x]\<in>Red B" using hlam by simp
    then have "M \<in> Red B" by (simp add: subst_same)
    then have "M \<in> SN" by (auto simp:using_cr1 red_cr)
    then show "sn M" by (simp add: SN_def)
  qed
  show "Lam A x M \<in> RedFun (Red A) (Red B)"
  proof (rule redfunI,auto)
    fix N
    assume b2: "N \<in> Red A"
    then have "N \<in> SN" by (auto simp: using_cr1 red_cr)
    then have b3: "sn N" by (simp add:SN_def)
    show "App (Lam A x M) N \<in> Red B" using b1 b3 b2 hlam
    proof(induct M N rule: double_SN)
      fix M N (*xa z*)
      assume ih1: "\<And>M'.  \<lbrakk>M \<leadsto> M'; N\<in>Red A; \<forall>R. R \<in> Red A \<longrightarrow> M'[x\<mapsto>R]\<in>Red B\<rbrakk> \<Longrightarrow> App (Lam A x M') N \<in> Red B" 
      assume ih2: "\<And>N'.  \<lbrakk>N \<leadsto> N'; N'\<in>Red A; \<forall>R. R \<in> Red A \<longrightarrow> M[x\<mapsto>R]\<in>Red B\<rbrakk> \<Longrightarrow> App (Lam A x M) N' \<in> Red B"
      assume as1: "N \<in> Red A"
      assume as2: "\<forall>R. R\<in>Red A \<longrightarrow> M[x\<mapsto>R]\<in>Red B"
      have "\<forall>S. App (Lam A x M) N \<leadsto> S \<longrightarrow> S \<in> Red B"
      proof(intro strip)
        fix S
        assume red: "App (Lam A x M) N \<leadsto> S"
        moreover
        { assume "\<exists>M'. M \<leadsto> M' \<and> S = App (Lam A x M') N"
          then obtain M' where a1: "M \<leadsto> M'" and a2: "S = App (Lam A x M') N" by blast
          have "App (Lam A x M') N\<in>Red B" using ih1 a1 as1 as2
            apply(auto)
            apply(drule_tac x="M'" in meta_spec)
            apply(simp)
            apply(drule meta_mp)
            prefer 2
            apply(auto)[1]
            apply(rule allI, rule impI)
            apply(drule_tac x="R" in spec)
            apply(simp)
            apply(subgoal_tac "Red B \<in> CR")(*A*)
            apply(subgoal_tac "cr2 (Red B)")(*A*)
            apply(simp add:cr2_def)[1]
            apply(drule_tac x="M[x\<mapsto>R]" in spec)
            apply(drule_tac x="M'[x\<mapsto>R]" in spec)
            apply(simp add: red_subst)
            apply(simp add: CR_def,auto simp: red_cr)
            done
          then have "S\<in>Red B" using a2 by simp
        }
        moreover
        { assume "\<exists>N'. N \<leadsto> N' \<and> S = App (Lam A x M) N'"
          then obtain N' where b1: "N \<leadsto> N'" and b2: "S = App (Lam A x M) N'" by blast
          have "App (Lam A x M) N'\<in>Red B" using ih2 b1 as1 as2
            apply(auto)
            apply(drule_tac x="N'" in meta_spec)
            apply(simp)
            apply(drule meta_mp)
            apply(auto simp:using_cr2 red_cr)
            done
          then have "S\<in>Red B" using b2 by simp
        }
        moreover
        { assume "S = M[x\<mapsto>N]"
          then have "S\<in>Red B" using as1 as2 by auto
        }
        ultimately show "S \<in> Red B" 
          (* one wants to use the strong elimination principle; for this one 
             has to know that x\<sharp>u *)
        apply(cases, auto simp:exp.inject)
        defer
        apply(drule red_lam_inv)
        apply(auto)[1]
        apply(auto simp add: alpha subst_rename)
        done
      qed
      moreover have "App (Lam A x M) N \<in> NT" by (rule app_is_nt)
      moreover have "Red B \<in> CR" by (rule red_cr)
      ultimately show "App (Lam A x M) N \<in> Red B" by (auto simp: using_cr3)
    qed
  qed
qed

lemma using_repl: "M \<in> repl C u D \<Longrightarrow> u \<sharp> N \<Longrightarrow> N \<in> D \<Longrightarrow> M{u\<mapsto>N} \<in> C"
  by (simp add: repl_def)

lemma uvar_fresh_sym :
  fixes u v::uvar
  shows "u \<sharp> v \<Longrightarrow> v \<sharp> u"
by (simp add: at_fresh at_uvar_inst pt_uvar_inst)

lemma subst_swap_merge:
  fixes u v::uvar
  fixes M N::exp
  fixes \<theta>::exp_repl
  shows "v \<sharp> M \<Longrightarrow> ([(u,v)]\<bullet>M){v\<mapsto>N} = M{u\<mapsto>N}"
  and "v \<sharp> \<theta> \<Longrightarrow> ([(u,v)]\<bullet>\<theta>){v\<mapsto>N} = \<theta>{u\<mapsto>N}"
apply (nominal_induct M and \<theta> avoiding:u v N rule:exp_exp_repl.strong_inducts)
apply (auto simp:at_calc at_uvar_inst pt_uvar_inst)
apply (simp add: dj_var_uvar at_fresh_ineq pt_fresh_fresh at_uvar_inst pt_uvar_inst)
apply (simp add: at_fresh at_uvar_inst)
apply (simp add: abs_fresh)
apply (subgoal_tac "([(u,v)]\<bullet>var) \<sharp> (v,N)",simp)
apply (subgoal_tac "[(u, v)] \<bullet> var = var",auto simp:ty_eqvt)
apply (rule pt_fresh_fresh,auto simp:at_fresh at_uvar_inst pt_uvar_inst)
apply (simp add: dj_var_uvar at_fresh_ineq pt_fresh_fresh at_uvar_inst pt_uvar_inst)+
apply (subgoal_tac "uvar \<sharp> ([(u,v)]\<bullet>exp1,v,N)",auto simp:abs_fresh fresh_prod)
apply (simp add: pt_fresh_left at_calc at_uvar_inst pt_uvar_inst)
apply (simp add: at_fresh at_uvar_inst)
apply (auto simp add: fresh_trail_label pt_fresh_fresh at_uvar_inst pt_uvar_inst)
done

lemma subst_eq_aux:
  fixes u v::uvar
  fixes M N R S::exp
  assumes "v \<sharp> M"
  assumes "R = [(u,v)]\<bullet>M"
  shows "M{u\<mapsto>N} = R{v\<mapsto>N}"
by (auto simp:subst_swap_merge assms)

lemma red_box_aux : 
  "sn M \<Longrightarrow> sn N \<Longrightarrow> C \<in> CR \<Longrightarrow> 
   M \<in> Red A \<longrightarrow> u \<sharp> M \<longrightarrow> N \<in> repl C u (Red A) \<longrightarrow> LetBox \<tau> (Box M) u N \<in> C"
  apply(rule double_SN[of M N],auto)
  apply(rule using_cr3,auto simp:NT_def nt.intros)
  apply(erule red.cases)
  apply(simp_all add: exp.inject nt_not_box alpha)
  proof-
    fix M1 N1 M2 N2 R :: exp
    fix v \<tau>'
    assume hred:"M2 \<in> Red A" "N1 \<in> repl C u (Red A)"
    assume hf:"\<tau> = \<tau>' \<and> M1 = M2 \<and> (u = v \<and> N1 = N2 \<or> u \<noteq> v \<and> N1 = [(u, v)] \<bullet> N2 \<and> u \<sharp> N2)" "u \<sharp> M2" "v \<sharp> M2"
    from this have "N2{v\<mapsto>M2} = N1{u\<mapsto>M1}"
      by (cases "u = v",auto,rule_tac subst_eq_aux,auto simp:at_ds5 pt3 pt_fresh_fresh at_uvar_inst pt_uvar_inst)
    from this hred show "N2{v\<mapsto>M2} \<in> C" by (auto simp: repl_def hf)
  next
    fix M1 N1 M2 N2 R S1 :: exp
    fix v \<tau>'
    assume ih:"\<And>y. S1 \<leadsto> y \<Longrightarrow> y \<in> Red A \<longrightarrow> u \<sharp> y \<longrightarrow> exp.LetBox \<tau>' (exp.Box y) u N1 \<in> C"
    assume hf:"\<tau> = \<tau>' \<and> exp.Box S1 = M1 \<and> (u = v \<and> N1 = N2 \<or> u \<noteq> v \<and> N1 = [(u, v)] \<bullet> N2 \<and> u \<sharp> N2)" "u \<sharp> S1"
    assume hM:"M1 \<leadsto> M2"
    assume hS1:"S1 \<in> Red A"
    assume hN1:"N1 \<in> repl C u (Red A)"
    from hM hf have "\<exists>S2. M2 = Box S2 \<and> S1 \<leadsto> S2" by (erule_tac red.cases,auto simp:exp.inject)
    from this obtain S2 where hS2:"M2 = Box S2" "S1 \<leadsto> S2" by blast
    from this hS1 have hredS2:"S2 \<in> Red A" by (auto simp: using_cr2 red_cr)
    from hf hS2 have hf':"u \<sharp> S2" by (auto simp: red_ufresh)
    from hf have "LetBox \<tau>' M2 v N2 = LetBox \<tau>' M2 u N1"
      by (simp add:exp.inject,cases "u=v",auto simp:alpha pt_swap_bij' pt_fresh_left at_calc at_uvar_inst pt_uvar_inst)
    from this hS2 hredS2 hf' ih show "LetBox \<tau>' M2 v N2 \<in> C" by (auto)
  next
    fix M1 N1 N1' M2 N2 R S1 :: exp
    fix v \<tau>'
    assume hC:"C \<in> CR"
    assume ih:"\<And>w. N1 \<leadsto> w \<Longrightarrow> w \<in> repl C u (Red A) \<longrightarrow> exp.LetBox \<tau>' M1 u w \<in> C"
    assume hf:"\<tau> = \<tau>' \<and> exp.Box S1 = M1 \<and> (u = v \<and> N1 = N1' \<or> u \<noteq> v \<and> N1 = [(u, v)] \<bullet> N1' \<and> u \<sharp> N1')" "u \<sharp> S1" "v \<sharp> M1"
    assume hN:"N1' \<leadsto> N2"
    from this have hN':"[(u,v)]\<bullet>N1'\<leadsto>[(u,v)]\<bullet>N2" by (auto simp:red.eqvt)
    assume hS1:"S1 \<in> Red A"
    assume hN1:"N1 \<in> repl C u (Red A)"
    from ih hN hN' hN1 hf show "LetBox \<tau>' M1 v N2 \<in> C" 
    proof (cases "u=v",subgoal_tac "N2 \<in> repl C u (Red A)",auto simp:using_pcr2 repl_in_pcr hC)
      assume heq:"M1 = exp.Box S1" "N1 = [(u, v)] \<bullet> N1'"
      from this hN' hN1 have hN2:"[(u,v)]\<bullet>N2 \<in> repl C u (Red A)" by (auto simp:using_pcr2 repl_in_pcr hC)
      assume hf': "u \<sharp> S1" "v \<sharp> S1" "u \<sharp> N1'"
      from this hN have hf'':"u \<sharp> N2" by (auto simp:red_ufresh)
      from this hf' have "LetBox \<tau>' M1 v N2 = LetBox \<tau>' M1 u ([(u,v)]\<bullet>N2)"
        by (simp add:exp.inject, auto simp:pt_swap_bij'' alpha pt_swap_bij' pt_fresh_left at_calc 
            at_uvar_inst pt_uvar_inst red_ufresh)
      from this heq hN' hN2 ih show "LetBox \<tau>' (exp.Box S1) v N2 \<in> C" by (simp)
    qed
  qed

(* Lemma 11 *)
lemma red_box: "M \<in> Red A \<Longrightarrow> Box (M) \<in> RedBox (Red A)" 
  apply(rule redboxI,clarsimp)
  apply(subgoal_tac "M \<in> SN")
  defer
  apply(simp add:using_cr1 red_cr)
  apply(subgoal_tac "N \<in> SN")
  defer
  apply(simp add:using_pcr1 repl_in_pcr)
  apply(simp add:SN_def)
  apply(auto simp: red_box_aux)
  done

lemma red_app :
  "M \<in> Red (FunTy \<sigma> \<tau>) \<Longrightarrow> N \<in> Red \<sigma> \<Longrightarrow> App M N \<in> Red \<tau>"
  by (simp add: RedFun_def)

lemma bedefault_red :
  "\<theta> \<in> Red_repl \<tau> \<Longrightarrow> bedefault \<theta> \<in> Red \<tau>"
  by (induct rule:Red_repl.induct,auto)

lemma belookup_red :
  "\<theta> \<in> Red_repl \<tau> \<Longrightarrow> belookup \<theta> l = Some M \<Longrightarrow> M \<in> Red (bty \<tau> l)"
  apply (induct rule:Red_repl.induct,auto)
  apply (case_tac "la = l",simp_all)
  done

lemma qfbranch_red :
  "\<theta> \<in> Red_repl \<tau> \<Longrightarrow> \<forall>M. M \<in> Red (bty \<tau> l) \<longrightarrow> f M \<in> Red \<tau> \<Longrightarrow> qfbranch \<theta> l f \<in> Red \<tau>"
  apply(simp add:qfbranch_def, cases "belookup \<theta> l", simp_all)
  apply(auto simp:bedefault_red)
  apply(auto simp:belookup_red)
  done

lemma red_repl_fold:
  shows "\<theta> \<in> Red_repl \<tau> \<Longrightarrow> qfold q \<theta> \<in> Red \<tau>"
  and "\<And>Mnil Lcons. Mnil \<in> Red \<tau> \<Longrightarrow> \<theta> \<in> Red_repl \<tau> \<Longrightarrow> trail_label_arity Lcons = 2 \<Longrightarrow> qfaux qs \<theta> Mnil Lcons \<in> Red \<tau>"
  apply (nominal_induct q and qs avoiding:\<theta> rule:trail_trail_repl.strong_inducts,simp_all)
  apply (rule qfbranch_red, rule assms,auto simp:bty_def)
  apply (rule qfbranch_red,rule assms,auto simp:bty_def bt_def)
  apply (rule red_app, rule red_app,auto)
  apply (rule qfbranch_red,rule assms,auto simp:bty_def bt_def)+
  apply (rule red_app,auto)
  apply (rule qfbranch_red,rule assms,auto simp:bty_def bt_def)
  apply (rule red_app, rule red_app,auto)
  apply (rule qfbranch_red,rule assms,auto simp:bty_def bt_def)
  apply (rule red_app, rule red_app,auto)
  apply (rule qfbranch_red,rule assms,auto simp:bty_def bt_def)+
  apply (subgoal_tac "App (App M Ma) (qfold trail \<theta>) \<in> Red \<tau>",auto)
  apply (rule red_app, rule red_app,auto)
  apply (rule qfbranch_red,rule assms,auto simp:bty_def bt_def)+
  apply (rule red_app, rule red_app,auto)
  apply (rule qfbranch_red,rule assms,auto simp:bty_def bt_def)
  apply (rule red_app, rule red_app,auto)
  done

lemma SN_ernil : "sn M \<Longrightarrow> sn_repl (ErNil M)"
  proof (induct M rule:sn.induct, rule sn_repl.intros)
    fix M \<theta>'
    assume "ErNil M \<leadsto> \<theta>'"
    from this have "\<exists>N. M \<leadsto> N \<and> \<theta>' = ErNil N"
      by (erule_tac red_repl_cases(1),simp add:exp_repl.inject)
    from this obtain N where "M \<leadsto> N" and "\<theta>' = ErNil N" by blast
    moreover assume "(\<And>N. M \<leadsto> N \<Longrightarrow> sn N)" " (\<And>N. M \<leadsto> N \<Longrightarrow> sn_repl (ErNil N))"
    ultimately show "sn_repl \<theta>'"
    by (simp)
  qed

lemma red_repl_cons_case : 
  assumes h:"ErCons l M \<theta> \<leadsto> \<theta>'"
  shows "(\<exists>M'. M \<leadsto> M' \<and> \<theta>' = ErCons l M' \<theta>) \<or> (\<exists>\<theta>''. \<theta> \<leadsto> \<theta>'' \<and> \<theta>' = ErCons l M \<theta>'')"
  by (rule red_repl_cases(2)[OF h],auto simp:exp_repl.inject)

lemma combine_SN_repl_aux:
  assumes a: "sn a"
  and b: "sn_repl b"
  and hyp: "\<And>x z.
    \<lbrakk>\<And>y. x \<leadsto> y \<Longrightarrow> sn y; \<And>y. x \<leadsto> y \<Longrightarrow> P y z;
     \<And>u. z \<leadsto> u \<Longrightarrow> sn_repl u; \<And>u. z \<leadsto> u \<Longrightarrow> P x u\<rbrakk> \<Longrightarrow> P x z"
  shows "P a b"
proof -
  from a
  have r: "\<And>b. sn_repl b \<Longrightarrow> P a b"
  proof (induct a rule: sn.induct)
    case (step_sn x)
    note SNI' = step_sn
    have "sn_repl b" by fact
    thus ?case
    proof (induct b rule: sn_repl.induct)
      case (step_sn_repl y)
      show ?case
        apply (rule hyp)
        apply (erule SNI')
        apply (erule SNI')
        apply (rule sn_repl.step_sn_repl)
        apply (erule step_sn_repl)+
        done
    qed
  qed
  from b show ?thesis by (rule r)
qed

lemma combine_SN_repl[consumes 2]:
  fixes a::exp and b::exp_repl and P::"exp \<Rightarrow> exp_repl \<Rightarrow> bool"
  assumes ha: "sn a"
  and     hb: "sn_repl b" 
  and     hc: "\<And>x z. \<lbrakk>\<And>y. x \<leadsto> y \<Longrightarrow> P y z; \<And>u. z \<leadsto> u \<Longrightarrow> P x u\<rbrakk> \<Longrightarrow> P x z"
  shows "P a b"
using ha hb hc
apply(rule_tac combine_SN_repl_aux)
apply(assumption)+
apply(blast)
done

lemma SN_red_repl:
  "\<theta> \<in> Red_repl \<tau> \<Longrightarrow> sn_repl \<theta>"
  apply(induct \<theta> rule: Red_repl.induct)
  apply(rule SN_ernil, subgoal_tac "M \<in> SN", simp add:SN_def)
  apply(auto simp:using_cr1 red_cr)
  proof-
    fix M l Ns \<theta>'
    assume "M \<in> Red (bty \<tau> l)"
    from this have "sn M" by (subgoal_tac "M \<in> SN",simp add:SN_def,auto simp:using_cr1 red_cr)
    moreover assume hns: "sn_repl Ns"
    ultimately show "sn_repl (ErCons l M Ns)"
    proof (induct M Ns rule:combine_SN_repl,rule_tac sn_repl.intros)
      fix R Rs \<theta>'
      assume ih:"(\<And>y. R \<leadsto> y \<Longrightarrow> sn_repl (ErCons l y Rs))" "(\<And>u. Rs \<leadsto> u \<Longrightarrow> sn_repl (ErCons l R u))"
      assume "ErCons l R Rs \<leadsto> \<theta>'"
      note red_repl_cons_case[OF this]
      from this show "sn_repl \<theta>'" by (auto simp:ih)
    qed
  qed  

lemma CR2_repl:
  shows "\<theta> \<leadsto> \<theta>' \<Longrightarrow> \<theta> \<in> Red_repl \<tau> \<Longrightarrow> \<theta>' \<in> Red_repl \<tau>"
  apply(nominal_induct \<theta> \<theta>' rule:red_red_repl.strong_inducts(2)[of _ _ "\<lambda>x1 x2 x3. True"],auto)
  apply(rule Red_repl.intros)
  apply(erule_tac Red_repl.cases,auto simp:exp_repl.inject)
  apply(auto simp: using_cr2 red_cr)
  apply(erule_tac Red_repl.cases,auto simp:exp_repl.inject)
  apply(auto simp: using_cr2 red_cr Red_repl.intros)
  apply(erule_tac Red_repl.cases,auto simp:exp_repl.inject)
  done

(* Lemma 12(1) *)
lemma red_repl: 
  assumes "\<theta> \<in> Red_repl \<tau>"
  shows "Insp \<theta> \<in> Red \<tau>"
  proof-
    have hsn:"sn_repl \<theta>" by (rule SN_red_repl,rule assms)
    from hsn assms show ?thesis
    apply(induct)
    apply(rule using_cr3)
    apply(simp add:NT_def,rule nt.intros)
    apply(rule red_cr)
    apply(auto)
    apply(erule red_cases,simp)
    apply(subgoal_tac "sfold (skel_of_trail q) Ms = qfold q Ms")
    apply(simp,rule red_repl_fold)
    apply(auto simp add:exp.inject)
    apply(simp add:qfold_def)
    apply(auto simp:CR2_repl)
    done
 qed

(* Lemma 12(2) *)
lemma red_repl2: 
  assumes "M \<in> Red (BoxTy \<sigma>)"
  assumes "\<theta> \<in> Red_repl \<tau>"
  shows "IInsp M \<theta> \<in> Red \<tau>"
  proof-
    have "Red (BoxTy \<sigma>) \<in> CR" by (rule red_cr)
    from this assms have "M \<in> SN" by (auto simp:using_cr1)
    from this have hsn1:"sn M" by (simp add: SN_def)
    have hsn2:"sn_repl \<theta>" by (rule SN_red_repl,rule assms)
    from hsn1 hsn2 assms show ?thesis
    apply(induct M \<theta> rule:combine_SN_repl)
    apply(rule using_cr3)
    apply(simp add:NT_def,rule nt.intros)
    apply(rule red_cr)
    apply(auto)
    apply(erule red_cases,simp)
    apply(subgoal_tac "sfold (skel_of_trail q) \<theta> = qfold q \<theta>")
    apply(simp,rule red_repl_fold)
    apply(auto simp add:exp.inject)
    apply(simp add:qfold_def)
    apply(auto simp:CR2_repl)
    apply(subgoal_tac "M' \<in> RedBox (Red \<sigma>)",auto)
    apply(rule using_cr2,auto)
    apply(subgoal_tac "Red (BoxTy \<sigma>) \<in> CR",simp)
    apply(rule red_cr)
    done
 qed

(* Theorem 13 *)
lemma main_property:
  shows "\<Delta>;\<Gamma> \<turnstile> M : \<tau> \<Longrightarrow>
    red_ctx \<Gamma> \<gamma> \<Longrightarrow>
         red_uctx \<Delta> \<delta> \<Longrightarrow>
         M<\<delta>,\<gamma>> \<in> Red \<tau>"
  and "\<Delta> \<turnstile> Ms : \<tau> \<Longrightarrow>
         red_uctx \<Delta> \<delta> \<Longrightarrow>
         Ms<\<delta>,[]> \<in> Red_repl \<tau>"
  apply(nominal_induct \<Delta> \<Gamma> M \<tau> and \<Delta> Ms \<tau> avoiding: \<gamma> \<delta> rule:wf_wf_repl.strong_inducts)
  apply(simp add: lookup_is_red)
  apply(simp add: ulookup_is_red)
  apply(simp)
  apply(rule red_lam,clarsimp)
  apply(subgoal_tac "M<\<delta>,(x,N)#\<gamma>> \<in> Red \<tau>'",simp only:subst1_merge)
  apply(simp add:red_ctx.intros)
  apply(simp)
  apply(rule redfunE,auto) 
  apply(rule red_box)
  apply(subgoal_tac "M<\<delta>,\<gamma>> = M<\<delta>,[]>",simp add:red_ctx.intros, erule box_empty)
  apply(rule redboxE,blast,simp add:subst2_fresh, rule red_cr)
  apply(simp add:repl_def,auto)
  apply(subgoal_tac "N<(u,UVar u)#\<delta>,\<gamma>> \<in> Red \<tau>'",simp only:uvar_id)
  apply(auto simp add:red_uctx.intros uvar_is_red)
  apply(subgoal_tac "N<(u,Na)#\<delta>,\<gamma>> \<in> Red \<tau>'",simp only:usubst1_merge)
  apply(auto simp add:red_uctx.intros uvar_is_red)
  apply(subgoal_tac "Ms<\<delta>,\<gamma>> = Ms<\<delta>,[]>",simp,auto simp:red_repl, erule box_empty_repl)
  apply(rule red_repl2,auto)
  apply(subgoal_tac "Ns<\<delta>,\<gamma>> = Ns<\<delta>,[]>",simp,erule box_empty_repl)
  apply(rule Red_repl.intros, subgoal_tac "red_ctx [] []",auto, rule red_ctx.intros)
  apply(rule Red_repl.intros, subgoal_tac "red_ctx [] []",auto, rule red_ctx.intros)
  done  

lemma id_subst_is_red:
  shows "valid \<Gamma> \<Longrightarrow> red_ctx \<Gamma> (id_subst \<Gamma>)"
  apply(induct \<Gamma> rule: valid.induct)
  apply(auto simp add: var_in_c red_cr red_ctx.intros)
  done

lemma id_usubst_is_red:
  shows "uvalid \<Delta> \<Longrightarrow> red_uctx \<Delta> (id_usubst \<Delta>)"
  apply(induct \<Delta> rule: uvalid.induct)
  apply(auto simp add: var_in_c red_cr red_uctx.intros)
  done

(* Corollary 14 *)
theorem wf_is_SN:
  "\<Delta>;\<Gamma> \<turnstile> M : \<tau> \<Longrightarrow> M \<in> SN"
  apply(subgoal_tac "M<id_usubst \<Delta>,id_subst \<Gamma>> \<in> Red \<tau>")
  apply(rule using_cr1)
  apply(simp add:id_subst_is_id id_usubst_is_id)
  apply(rule red_cr)
  apply(rule main_property,simp)
  apply(simp_all add: id_subst_is_red id_usubst_is_red wf_valid)
  done

end