theory Translation
imports "~~/src/HOL/Nominal/Nominal" "Modal" "CCau"
(*imports "Nominal/Nominal2"*)

begin

(* This module defines a translation from \<lambda>hc to \<lambda>hs: 
   this allows us to extend the SN result for \<lambda>hs to \<lambda>hc. *)

lemma trail_label_swap[simp] :
  fixes \<pi>::"var prm"
  fixes \<pi>'::"uvar prm"
  fixes l::trail_label
  shows "\<pi>\<bullet>l = l \<and> \<pi>'\<bullet>l = l"
  by (nominal_induct l rule:trail_label.strong_inducts,auto)

lemma code_swap_id[simp] :
  fixes a::var
  fixes u::uvar
  fixes \<sigma>::ty
  fixes M::code
  fixes \<theta>::code_repl
  shows "[(a,a)]\<bullet>\<sigma> = \<sigma> \<and> [(u,u)]\<bullet>\<sigma> = \<sigma>"
  and "[(a,a)]\<bullet>M = M \<and> [(u,u)]\<bullet>M = M"
  and "[(a,a)]\<bullet>\<theta> = \<theta> \<and> [(u,u)]\<bullet>\<theta> = \<theta>"
  by (nominal_induct \<sigma> and M and \<theta> rule:ty_code_code_repl.strong_inducts,auto simp:calc_atm)

lemma trail_swap_id[simp] :
  fixes a::var
  fixes u::uvar
  fixes q::trail
  fixes \<zeta>::trail_repl
  shows "[(a,a)]\<bullet>q = q \<and> [(u,u)]\<bullet>q = q"
  and "[(a,a)]\<bullet>\<zeta> = \<zeta> \<and> [(u,u)]\<bullet>\<zeta> = \<zeta>"
  by (nominal_induct q and \<zeta> rule:trail_trail_repl.strong_inducts,auto simp:calc_atm)

lemma trm_swap_id[simp] :
  fixes a::var
  fixes u::uvar
  fixes M::trm
  fixes \<theta>::trm_repl
  shows "[(a,a)]\<bullet>M = M \<and> [(u,u)]\<bullet>M = M"
  and "[(a,a)]\<bullet>\<theta> = \<theta> \<and> [(u,u)]\<bullet>\<theta> = \<theta>"
  by (nominal_induct M and \<theta> rule:trm_trm_repl.strong_inducts,auto simp:calc_atm)

(* Box-free contexts (expressions with holes) for \<lambda>hs *)
nominal_datatype punch_exp =
  PHole
| PLam "Common.ty" "var" "punch_exp"
| PAppL "punch_exp" "exp" | PAppR "exp" "punch_exp"
| PLetL "Common.ty" "punch_exp" "uvar" "exp" | PLetR "Common.ty" "exp" "uvar" "punch_exp"
| PInsp "punch_exp_repl"
| PIInspL "punch_exp" "exp_repl" | PIInspR "exp" "punch_exp_repl"
and punch_exp_repl = 
  PErNil "punch_exp"
| PErConsL "trail_label" "punch_exp" "exp_repl" | PErConsR "trail_label" "exp" "punch_exp_repl"

nominal_primrec pe_apply :: "punch_exp \<Rightarrow> exp \<Rightarrow> exp" ("_\<lceil>_\<rceil>" [100,60] 60) 
and ptr_apply :: "punch_exp_repl \<Rightarrow> exp \<Rightarrow> exp_repl" ("_\<lceil>_\<rceil>" [100,60] 60) where
  "PHole\<lceil>M\<rceil> = M"
| "(PLam \<sigma> a F)\<lceil>M\<rceil> = Modal.Lam \<sigma> a (F\<lceil>M\<rceil>)"
| "(PAppL F N)\<lceil>M\<rceil> = Modal.App (F\<lceil>M\<rceil>) N"
| "(PAppR N F)\<lceil>M\<rceil> = Modal.App N (F\<lceil>M\<rceil>)"
| "(PLetL \<sigma> F u N)\<lceil>M\<rceil> = Modal.LetBox \<sigma> (F\<lceil>M\<rceil>) u N"
| "(PLetR \<sigma> N u F)\<lceil>M\<rceil> = Modal.LetBox \<sigma> N u (F\<lceil>M\<rceil>)"
| "(PInsp Fs)\<lceil>M\<rceil> = Modal.Insp (Fs\<lceil>M\<rceil>)"
| "(PIInspL F \<theta>)\<lceil>M\<rceil> = Modal.IInsp (F\<lceil>M\<rceil>) \<theta>"
| "(PIInspR N Fs)\<lceil>M\<rceil> = Modal.IInsp N (Fs\<lceil>M\<rceil>)"
| "(PErNil F)\<lceil>M\<rceil> = Modal.ErNil (F\<lceil>M\<rceil>)"
| "(PErConsL l F \<theta>)\<lceil>M\<rceil> = Modal.ErCons l (F\<lceil>M\<rceil>) \<theta>"
| "(PErConsR l N Fs)\<lceil>M\<rceil> = Modal.ErCons l N (Fs\<lceil>M\<rceil>)"
by (rule TrueI)+

lemma pexp_apply_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  and F::punch_exp
  and Fs::punch_exp_repl
  and M::exp
  shows "pi\<bullet>(F\<lceil>M\<rceil>) = (pi\<bullet>F)\<lceil>pi\<bullet>M\<rceil>"
  and "pi'\<bullet>(F\<lceil>M\<rceil>) = (pi'\<bullet>F)\<lceil>pi'\<bullet>M\<rceil>"
  and "pi\<bullet>(Fs\<lceil>M\<rceil>) = (pi\<bullet>Fs)\<lceil>pi\<bullet>M\<rceil>"
  and "pi'\<bullet>(Fs\<lceil>M\<rceil>) = (pi'\<bullet>Fs)\<lceil>pi'\<bullet>M\<rceil>"
  by (nominal_induct F and Fs rule:punch_exp_punch_exp_repl.strong_inducts,auto)

(*
lemma pt_trail_apply_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  and F::punch_trm
  and Fs::punch_trm_repl
  and q::trail
  shows "pi\<bullet>(F\<lceil>q\<rceil>) = (pi\<bullet>F)\<lceil>pi\<bullet>q\<rceil>"
  and "pi'\<bullet>(F\<lceil>q\<rceil>) = (pi'\<bullet>F)\<lceil>pi'\<bullet>q\<rceil>"
  and "pi\<bullet>(Fs\<lceil>q\<rceil>) = (pi\<bullet>Fs)\<lceil>pi\<bullet>q\<rceil>"
  and "pi'\<bullet>(Fs\<lceil>q\<rceil>) = (pi'\<bullet>Fs)\<lceil>pi'\<bullet>q\<rceil>"
*)

(* The erasure map of Sect. 3 *)
nominal_primrec xlate_ty :: "CCau.ty \<Rightarrow> Common.ty"
and xlate_c :: "CCau.code \<Rightarrow> Common.code"
and xlate_crepl :: "CCau.code_repl \<Rightarrow> Common.code_repl"
where
  "xlate_ty PTy = Common.PTy"
| "xlate_ty (FunTy \<sigma> \<tau>) = Common.FunTy (xlate_ty \<sigma>) (xlate_ty \<tau>)"
| "xlate_ty (BoxTy _ \<sigma>) = Common.BoxTy (xlate_ty \<sigma>)"
| "xlate_c (Var a) = Common.Var a"
| "xlate_c (UVar u) = Common.UVar u"
| "a \<sharp> \<sigma> \<Longrightarrow> xlate_c (Lam \<sigma> a r) = Common.Lam (xlate_ty \<sigma>) a (xlate_c r)"
| "xlate_c (App r s) = Common.App (xlate_c r) (xlate_c s)"
| "u \<sharp> (\<sigma>,r) \<Longrightarrow> xlate_c (LetBox \<sigma> r u s) = Common.LetBox (xlate_ty \<sigma>) (xlate_c r) u (xlate_c s)"
| "xlate_c (Box r) = Common.Box (xlate_c r)"
| "xlate_c (OInsp \<theta>) = Common.OInsp (xlate_crepl \<theta>)"
| "xlate_c (IInsp r \<theta>) = Common.IInsp (xlate_c r) (xlate_crepl \<theta>)"
| "xlate_crepl (CrNil r) = Common.CrNil (xlate_c r)"
| "xlate_crepl (CrCons l r ss) = Common.CrCons l (xlate_c r) (xlate_crepl ss)"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(simp_all add: abs_fresh fresh_prod exp.inject fresh_ty lookup_eqvt ulookup_eqvt)+
  apply(fresh_guess)+
done

(*
lemma cp_ty :
  shows "cp TYPE (ty) TYPE(var) TYPE(uvar)"
  and "cp TYPE (ty) TYPE(uvar) TYPE(var)"
apply(simp add: cp_def)
apply(rule allI)+
apply(induct_tac x)
apply(simp_all)
done
*)

lemma pi_fresh_aux :
  fixes a::var
  fixes u::uvar
  fixes \<pi>::"var prm"
  fixes \<pi>':: "uvar prm"
  fixes \<sigma>::ty
  fixes r::code
  shows "a \<sharp> \<sigma> \<Longrightarrow> (\<pi>\<bullet>a) \<sharp> (\<pi>\<bullet>\<sigma>)"
  and "a \<sharp> r \<Longrightarrow> (\<pi>\<bullet>a) \<sharp> (\<pi>\<bullet>r)"
  and "a \<sharp> \<sigma> \<Longrightarrow> (\<pi>'\<bullet>a) \<sharp> (\<pi>'\<bullet>\<sigma>)"
  and "a \<sharp> r \<Longrightarrow> (\<pi>'\<bullet>a) \<sharp> (\<pi>'\<bullet>r)"
  and "u \<sharp> \<sigma> \<Longrightarrow> (\<pi>'\<bullet>u) \<sharp> (\<pi>'\<bullet>\<sigma>)"
  and "u \<sharp> r \<Longrightarrow> (\<pi>'\<bullet>u) \<sharp> (\<pi>'\<bullet>r)"
  and "u \<sharp> \<sigma> \<Longrightarrow> (\<pi>\<bullet>u) \<sharp> (\<pi>\<bullet>\<sigma>)"
  and "u \<sharp> r \<Longrightarrow> (\<pi>\<bullet>u) \<sharp> (\<pi>\<bullet>r)"
by (auto simp add:pt_fresh_bij1 pt_var_inst at_var_inst pt_uvar_inst at_uvar_inst 
    pt_fresh_bij_ineq cp_var_uvar_inst cp_uvar_var_inst)

lemma xlate_ty_eqvt[eqvt]:
  fixes pi::"var prm"
  and pi'::"uvar prm"
  and \<sigma>::CCau.ty
  and s::CCau.code
  and ss::CCau.code_repl
  shows "pi\<bullet>(xlate_ty \<sigma>) = (xlate_ty (pi\<bullet>\<sigma>))" "pi'\<bullet>(xlate_ty \<sigma>) = (xlate_ty (pi'\<bullet>\<sigma>))"
  and "pi\<bullet>(xlate_c s) = (xlate_c (pi\<bullet>s))" "pi'\<bullet>(xlate_c s) = (xlate_c (pi'\<bullet>s))"
  and "pi\<bullet>(xlate_crepl ss) = (xlate_crepl (pi\<bullet>ss))" "pi'\<bullet>(xlate_crepl ss) = (xlate_crepl (pi'\<bullet>ss))"
by (nominal_induct \<sigma> and s and ss rule:ty_code_code_repl.strong_inducts, auto simp add: pi_fresh_aux)

lemma xlate_ty_fresh:
  fixes a::var
  fixes u::uvar
  shows "a \<sharp> xlate_ty \<sigma>" "u \<sharp> xlate_ty \<sigma>"
  and "a \<sharp> r \<Longrightarrow> a \<sharp> xlate_c r" "u \<sharp> r \<Longrightarrow> u \<sharp> xlate_c r"
  and "a \<sharp> rs \<Longrightarrow> a \<sharp> xlate_crepl rs" "u \<sharp> rs \<Longrightarrow> u \<sharp> xlate_crepl rs"
by (nominal_induct \<sigma> and r and rs rule:ty_code_code_repl.strong_inducts, auto simp:abs_fresh)

(*
nominal_datatype trail = Refl "code"
                       | Trans "trail" "trail"
                       | Beta "ty" "\<guillemotleft>var\<guillemotright>code" "code"
                       | BetaBox "ty" "code" "\<guillemotleft>uvar\<guillemotright>code"
                       | ITI "trail" "code_repl"
                       | OTI "trail" "code_repl"
                       | QLam "ty" "\<guillemotleft>var\<guillemotright>trail"
                       | QApp "trail" "trail"
                       | QLet "ty" "trail" "\<guillemotleft>uvar\<guillemotright>trail"
                       | ITB "trail" "trail_repl"
                       | OTB "trail_repl"
and trail_repl = QrNil "trail" (* represents the default branch *)
               | QrCons "trail_label" "trail" "trail_repl"
*)
nominal_primrec xlate_q :: "CCau.trail \<Rightarrow> Common.trail"
and xlate_qrepl :: "CCau.trail_repl \<Rightarrow> Common.trail_repl"
where
  "xlate_q (Refl r) = Common.Refl (xlate_c r)"
| "xlate_q (Trans q1 q2) = Common.Trans (xlate_q q1) (xlate_q q2)"
| "a \<sharp> (\<sigma>,s) \<Longrightarrow> xlate_q (Beta \<sigma> a r s) = Common.Beta (xlate_ty \<sigma>) a (xlate_c r) (xlate_c s)"
| "u \<sharp> (\<sigma>,r) \<Longrightarrow> xlate_q (BetaBox \<sigma> r u s) = Common.BetaBox (xlate_ty \<sigma>) (xlate_c r) u (xlate_c s)"
| "xlate_q (ITI q \<theta>) = Common.ITI (xlate_q q) (xlate_crepl \<theta>)"
| "xlate_q (OTI q \<theta>) = Common.OTI (xlate_q q) (xlate_crepl \<theta>)"
| "a \<sharp> \<sigma> \<Longrightarrow> xlate_q (QLam \<sigma> a q) = Common.QLam (xlate_ty \<sigma>) a (xlate_q q)"
| "xlate_q (QApp q1 q2) = Common.QApp (xlate_q q1) (xlate_q q2)"
| "u \<sharp> (\<sigma>,q1) \<Longrightarrow> xlate_q (QLet \<sigma> q1 u q2) = Common.QLet (xlate_ty \<sigma>) (xlate_q q1) u (xlate_q q2)"
| "xlate_q (ITB q qr) = Common.ITB (xlate_q q) (xlate_qrepl qr)"
| "xlate_q (OTB qr) = Common.OTB (xlate_qrepl qr)"
| "xlate_qrepl (QrNil q) = Common.QrNil (xlate_q q)"
| "xlate_qrepl (QrCons l q qs) = Common.QrCons l (xlate_q q) (xlate_qrepl qs)"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(simp_all add: abs_fresh fresh_prod exp.inject fresh_ty lookup_eqvt ulookup_eqvt)+
  apply(fresh_guess)+
done

lemma xlate_q_fresh:
  fixes a::var
  fixes u::uvar
  shows "a \<sharp> q \<Longrightarrow> a \<sharp> xlate_q q" "u \<sharp> q \<Longrightarrow> u \<sharp> xlate_q q"
  and "a \<sharp> qs \<Longrightarrow> a \<sharp> xlate_qrepl qs" "u \<sharp> qs \<Longrightarrow> u \<sharp> xlate_qrepl qs"
by (nominal_induct q and qs rule:trail_trail_repl.strong_inducts, auto simp:abs_fresh xlate_ty_fresh)

nominal_primrec xlate :: "CCau.trm \<Rightarrow> Modal.exp" 
and xlate_repl :: "CCau.trm_repl \<Rightarrow> Modal.exp_repl" where
  "xlate (TVar a) = Modal.Var a"
| "xlate (TUVar a) = Modal.UVar a"
| "a \<sharp> \<sigma> \<Longrightarrow> xlate (TLam \<sigma> a M) = Modal.Lam (xlate_ty \<sigma>) a (xlate M)"
| "xlate (TApp M N) = Modal.App (xlate M) (xlate N)"
| "u \<sharp> (\<sigma>,M) \<Longrightarrow> xlate (TLetBox \<sigma> M u N) = Modal.LetBox (xlate_ty \<sigma>) (xlate M) u (xlate N)"
| "xlate (TBox q M) = Modal.Box (xlate M)"
| "xlate (TOInsp \<theta>) = Modal.Insp (xlate_repl \<theta>)"
| "xlate (TIInsp M \<theta>) = Modal.IInsp (xlate M) (xlate_repl \<theta>)"
| "xlate_repl (TrNil M) = Modal.ErNil (xlate M)"
| "xlate_repl (TrCons l M \<theta>) = Modal.ErCons l (xlate M) (xlate_repl \<theta>)"
  apply(finite_guess)+
  apply(rule TrueI)+
  apply(simp_all add: abs_fresh fresh_prod exp.inject fresh_ty lookup_eqvt ulookup_eqvt)+
  apply(fresh_guess)+
done

lemma xlate_eqvt[eqvt]:
  fixes pi::"var prm"
  fixes pi'::"uvar prm"
  shows "pi\<bullet>(xlate M) = xlate (pi\<bullet>M)"
  and "pi'\<bullet>(xlate M) = xlate (pi'\<bullet>M)"
  and "pi\<bullet>(xlate_repl \<theta>) = xlate_repl (pi\<bullet>\<theta>)"
  and "pi'\<bullet>(xlate_repl \<theta>) = xlate_repl (pi'\<bullet>\<theta>)"
by (nominal_induct M and \<theta> rule: trm_trm_repl.strong_inducts, auto simp:perm_bij fresh_bij xlate_ty_eqvt)

nominal_primrec xlate_pt :: "CCau.punch_trm \<Rightarrow> punch_exp" 
and xlate_pt_repl :: "CCau.punch_trm_repl \<Rightarrow> punch_exp_repl" where
  "xlate_pt PTHole = PHole"
| "xlate_pt (PTLam \<sigma> a F) = PLam (xlate_ty \<sigma>) a (xlate_pt F)"
| "xlate_pt (PTAppL F N) = PAppL (xlate_pt F) (xlate N)"
| "xlate_pt (PTAppR M F) = PAppR (xlate M) (xlate_pt F)"
| "xlate_pt (PTLetL \<sigma> F u N) = PLetL (xlate_ty \<sigma>) (xlate_pt F) u (xlate N)"
| "xlate_pt (PTLetR \<sigma> M u F) = PLetR (xlate_ty \<sigma>) (xlate M) u (xlate_pt F)"
| "xlate_pt (PTOInsp Fs) = PInsp (xlate_pt_repl Fs)"
| "xlate_pt (PTIInspL F \<theta>) = PIInspL (xlate_pt F) (xlate_repl \<theta>)"
| "xlate_pt (PTIInspR M Fs) = PIInspR (xlate M) (xlate_pt_repl Fs)"
| "xlate_pt_repl (PTrNil F) = PErNil (xlate_pt F)"
| "xlate_pt_repl (PTrConsL l F \<theta>) = PErConsL l (xlate_pt F) (xlate_repl \<theta>)"
| "xlate_pt_repl (PTrConsR l M Fs) = PErConsR l (xlate M) (xlate_pt_repl Fs)"
apply (auto)
done

lemma red_pe :
  fixes F :: punch_exp
  and Fs :: punch_exp_repl
  and M N :: exp
  assumes "M \<leadsto> N"
  shows "F\<lceil>M\<rceil> \<leadsto> F\<lceil>N\<rceil>" and "Fs\<lceil>M\<rceil> \<leadsto> Fs\<lceil>N\<rceil>"
  apply (nominal_induct F and Fs rule:punch_exp_punch_exp_repl.strong_inducts,auto simp:assms)
  apply (rule, auto)
  apply (rule, auto)
  apply (rule, auto)
  apply (rule red_letbox1_strong, auto)
  apply (rule red_letbox2_strong, auto)
  apply (rule, auto)+
  done

lemma xlate_fresh :
  fixes a :: var
  fixes u :: uvar
  fixes M :: trm
  fixes \<theta> :: trm_repl
  shows "(a \<sharp> M \<longrightarrow> a \<sharp> xlate M) \<and> (u \<sharp> M \<longrightarrow> u \<sharp> xlate M)" 
  and "(a \<sharp> \<theta> \<longrightarrow> a \<sharp> xlate_repl \<theta>) \<and> (u \<sharp> \<theta> \<longrightarrow> u \<sharp> xlate_repl \<theta>)"
by (nominal_induct M and \<theta> rule:trm_trm_repl.strong_inducts, auto simp: abs_fresh xlate_ty_fresh)

lemma pt_fresh :
  fixes a :: var and  u :: uvar
  fixes F :: punch_trm and Fs :: punch_trm_repl
  fixes M :: trm
  shows "a \<sharp> F \<Longrightarrow> a \<sharp> M \<Longrightarrow> a \<sharp> (F\<lceil>M\<rceil>)" and "u \<sharp> F \<Longrightarrow> u \<sharp> M \<Longrightarrow> u \<sharp> (F\<lceil>M\<rceil>)"
  and "a \<sharp> Fs \<Longrightarrow> a \<sharp> M \<Longrightarrow> a \<sharp> (Fs\<lceil>M\<rceil>)" and "u \<sharp> Fs \<Longrightarrow> u \<sharp> M \<Longrightarrow> u \<sharp> (Fs\<lceil>M\<rceil>)"
  apply (nominal_induct F and Fs rule:punch_trm_punch_trm_repl.strong_inducts,auto)
  apply (simp_all add:abs_fresh)
  done
  
lemma xlate_pt_fresh :
  fixes a :: var
  fixes u :: uvar
  fixes F :: punch_trm
  fixes Fs :: punch_trm_repl
  shows "a \<sharp> F \<Longrightarrow> a \<sharp> xlate_pt F" and "u \<sharp> F \<Longrightarrow> u \<sharp> xlate_pt F" 
  and "a \<sharp> Fs \<Longrightarrow> a \<sharp> xlate_pt_repl Fs" and "u \<sharp> Fs \<Longrightarrow> u \<sharp> xlate_pt_repl Fs" 
by (nominal_induct F and Fs rule:punch_trm_punch_trm_repl.strong_inducts, auto simp: abs_fresh xlate_fresh xlate_ty_fresh)

lemma trm_lam_alpha : "b \<sharp> M \<Longrightarrow> TLam \<tau> a M = TLam \<tau> b ([(b,a)]\<bullet>M)"
 by (simp add:trm.inject, rule sym, auto simp add:alpha abs_fresh)

lemma trm_letbox_alpha : "v \<sharp> (M,N) \<Longrightarrow> TLetBox \<tau> M u N = TLetBox \<tau> M v ([(v,u)]\<bullet>N)"
 by (simp add:trm.inject, rule sym, auto simp add:alpha abs_fresh)

lemma xlate_pt_to_pe[simp] :
  fixes F :: punch_trm 
  and Fs::punch_trm_repl
  and M :: trm
  shows "xlate (F\<lceil>M\<rceil>) = (xlate_pt F)\<lceil>xlate M\<rceil>"  
  and "xlate_repl (Fs\<lceil>M\<rceil>) = (xlate_pt_repl Fs)\<lceil>xlate M\<rceil>"
  apply (nominal_induct F and Fs rule:punch_trm_punch_trm_repl.strong_inducts,auto)
  apply (rename_tac ty a F)
  apply(subgoal_tac "\<exists> b::var. b \<sharp> (a,F,M,ty)",clarify)
  apply(subgoal_tac "b \<sharp> (F\<lceil>M\<rceil>)")
  apply(subgoal_tac "b \<noteq> a")
  apply(subgoal_tac "TLam ty a (F\<lceil>M\<rceil>) = TLam ty b ([(b,a)]\<bullet>(F\<lceil>M\<rceil>))")
  apply(simp, simp add:exp.inject alpha abs_fresh fresh_prod,rule)
  apply(subgoal_tac "xlate ([(b, a)] \<bullet> (F\<lceil>M\<rceil>)) = [(b,a)]\<bullet>(xlate (F\<lceil>M\<rceil>))",simp)
  apply(rule sym, rule xlate_eqvt(1))
  apply(subgoal_tac "xlate_pt F\<lceil>xlate M\<rceil> = xlate (F\<lceil>M\<rceil>)",simp add:xlate_fresh,simp)
  apply(simp add:trm_lam_alpha)
  apply(force simp:fresh_prod fresh_atm)
  apply(force simp add:pt_fresh fresh_prod)
  apply(rule exists_fresh,auto,finite_guess)
  apply (rename_tac ty F u N)
  apply(subgoal_tac "\<exists> v::uvar. v \<sharp> (F,M,u,N,ty)",clarify)
  apply(subgoal_tac "v \<sharp> (F\<lceil>M\<rceil>)")
  apply(subgoal_tac "TLetBox ty (F\<lceil>M\<rceil>) u N = TLetBox ty (F\<lceil>M\<rceil>) v ([(v,u)]\<bullet>N)")
  apply(simp, auto simp add:exp.inject alpha eqvts xlate_fresh trm_letbox_alpha)
  apply(rule exists_fresh,auto,finite_guess)
  apply (rename_tac ty N u F)
  apply(subgoal_tac "\<exists> v::uvar. v \<sharp> (F,M,u,N,ty)",clarify)
  apply(subgoal_tac "v \<sharp> (F\<lceil>M\<rceil>)")
  apply(subgoal_tac "v \<noteq> u")
  apply(subgoal_tac "TLetBox ty N u (F\<lceil>M\<rceil>) = TLetBox ty N v ([(v,u)]\<bullet> (F\<lceil>M\<rceil>))")
  apply(simp, simp add:exp.inject alpha abs_fresh fresh_prod,rule)
  apply(subgoal_tac "xlate ([(v, u)] \<bullet> (F\<lceil>M\<rceil>)) = [(v,u)]\<bullet>(xlate (F\<lceil>M\<rceil>))",simp)
  apply(rule sym, rule xlate_eqvt(2))
  apply(subgoal_tac "xlate_pt F\<lceil>xlate M\<rceil> = xlate (F\<lceil>M\<rceil>)",simp add:xlate_fresh,simp)
  apply(simp add:trm_letbox_alpha)
  apply(force simp:fresh_prod fresh_atm)
  apply(force simp add:pt_fresh fresh_prod)
  apply(rule exists_fresh,auto,finite_guess)
done

lemma xlate_ty_subst[simp] :
  fixes \<sigma>::ty
  fixes a::var
  fixes u::uvar
  fixes r::code
  shows "xlate_ty (\<sigma>[a\<mapsto>r]) = xlate_ty \<sigma>"
  and "xlate_ty (\<sigma>{u\<mapsto>r}) = xlate_ty \<sigma>"
by (nominal_induct \<sigma> rule: ty_code_code_repl.strong_inducts(1),auto)

lemma xlate_subst[simp] :
  fixes M N :: trm
  fixes \<theta> :: trm_repl
  shows "xlate (M[a\<mapsto>N]) = (xlate M)[a\<mapsto>xlate N] \<and> xlate (M{u\<mapsto>N,q,N'}) = (xlate M){u\<mapsto>xlate N}"
  and "xlate_repl (\<theta>[a\<mapsto>N]) = (xlate_repl \<theta>)[a\<mapsto>xlate N] \<and> xlate_repl (\<theta>{u\<mapsto>N,q,N'}) = (xlate_repl \<theta>){u\<mapsto>xlate N}"
  apply (nominal_induct M and \<theta> avoiding: a u N q N' rule: trm_trm_repl.strong_inducts)
  apply (auto simp:xlate_fresh)
  apply (subgoal_tac "var \<sharp> (ty[a\<mapsto>t2c N])",simp)
  apply (simp add:abs_fresh subst_fresh t2c_fresh)
  apply (subgoal_tac "var \<sharp> (ty{u\<mapsto>N'})",simp)
  apply (simp add:abs_fresh usubst_fresh)
  apply (subgoal_tac "uvar \<sharp> (ty[a\<mapsto>t2c N],trm1[a\<mapsto>N])",auto simp:fresh_prod)
  apply (auto simp add:abs_fresh subst_fresh t2c_fresh)
  apply (subgoal_tac "uvar \<sharp> (ty{u\<mapsto>N'},trm1{u\<mapsto>N,q,N'})")
  apply (auto simp:tlusubst_fresh abs_fresh fresh_prod usubst_fresh)
done

lemma btlookup_belookup_None :
  fixes N:: trm and \<theta> :: trm_repl
  shows "(\<lambda>x. True) N" and "btlookup \<theta> l = None \<Longrightarrow> belookup (xlate_repl \<theta>) l = None"
  by (nominal_induct N and \<theta> rule:trm_trm_repl.strong_inducts,auto)

lemma btlookup_belookup_Some :
  fixes M N:: trm and \<theta> :: trm_repl
  shows "(\<lambda>x. True) N" and "btlookup \<theta> l = Some M \<Longrightarrow> belookup (xlate_repl \<theta>) l = Some (xlate M)"
  by (nominal_induct N and \<theta> rule:trm_trm_repl.strong_inducts,auto)

lemma btdefault_bedefault : 
  fixes N::trm and \<theta>:: trm_repl
  shows "(\<lambda>x. True) N" and "xlate (btdefault \<theta>) = bedefault (xlate_repl \<theta>)"
  by (nominal_induct N and \<theta> rule:trm_trm_repl.strong_inducts,auto)

lemma qtfbranch_nil[simp] : 
  shows "qtfbranch (TrNil M) l f = M"
  by (simp add: qtfbranch_def)

lemma qfbranch_nil[simp] : 
  shows "qfbranch (ErNil M) l f = M"
  by (simp add: qfbranch_def)

lemma xlate_fold_0[simp]:
  fixes \<theta>::trm_repl
  shows "xlate (qtfbranch \<theta> l (\<lambda>M. M)) = qfbranch (xlate_repl \<theta>) l (\<lambda>M. M)"
  apply (simp add:qtfbranch_def qfbranch_def)
  apply (case_tac "btlookup \<theta> l",simp_all add:btlookup_belookup_None btlookup_belookup_Some btdefault_bedefault)
  done

lemma xlate_fold_nil[simp]:
  fixes \<theta>::trm_repl
  shows "xlate (qtfold trail \<theta>) = qfold (xlate_q trail) (xlate_repl \<theta>) \<Longrightarrow>
       xlate (qtfbranch \<theta> l (\<lambda>Ma. TApp (TApp Ma (qtfold trail \<theta>)) M)) =
       qfbranch (xlate_repl \<theta>) l (\<lambda>Ma. exp.App (exp.App Ma (qfold (xlate_q trail) (xlate_repl \<theta>))) (xlate M))"
  apply (simp add:qtfbranch_def qfbranch_def)
  apply (case_tac "btlookup \<theta> l",simp_all add:btlookup_belookup_None btlookup_belookup_Some btdefault_bedefault)
  done

lemma xlate_fold_1[simp]:
  fixes \<theta>::trm_repl
  shows "xlate (qtfold trail \<theta>) = qfold (xlate_q trail) (xlate_repl \<theta>) \<Longrightarrow>
   xlate (qtfbranch \<theta> l (\<lambda>M. TApp M (qtfold trail \<theta>))) =
   qfbranch (xlate_repl \<theta>) l (\<lambda>M. exp.App M (qfold (xlate_q trail) (xlate_repl \<theta>)))"
  apply (simp add:qtfbranch_def qfbranch_def)
  apply (case_tac "btlookup \<theta> l",simp_all add:btlookup_belookup_None btlookup_belookup_Some btdefault_bedefault)
  done

lemma xlate_fold_2[simp]:
  fixes \<theta>::trm_repl
  shows "xlate (qtfold trail1 \<theta>) = qfold (xlate_q trail1) (xlate_repl \<theta>) \<Longrightarrow>
   xlate (qtfold trail2 \<theta>) = qfold (xlate_q trail2) (xlate_repl \<theta>) \<Longrightarrow>
   xlate (qtfbranch \<theta> l (\<lambda>M. TApp (TApp M (qtfold trail1 \<theta>)) (qtfold trail2 \<theta>))) =
   qfbranch (xlate_repl \<theta>) l (\<lambda>M. exp.App (exp.App M (qfold (xlate_q trail1) (xlate_repl \<theta>))) 
     (qfold (xlate_q trail2) (xlate_repl \<theta>)))"
  apply (simp add:qtfbranch_def qfbranch_def)
  apply (case_tac "btlookup \<theta> l",simp_all add:btlookup_belookup_None btlookup_belookup_Some btdefault_bedefault)
  done

lemma xlate_fold_itb[simp]:
  fixes \<theta>::trm_repl
  shows "xlate (qtfold trail \<theta>) = qfold (xlate_q trail) (xlate_repl \<theta>) \<Longrightarrow>
       (\<And>M l. xlate (qtfaux trail_repl \<theta> M l) = qfaux (xlate_qrepl trail_repl) (xlate_repl \<theta>) (xlate M) l) \<Longrightarrow>
       xlate (qtfbranch \<theta> LITBc
               (\<lambda>M. qtfbranch \<theta> LITBn (\<lambda>N. qtfaux trail_repl \<theta> (TApp (TApp M N) (qtfold trail \<theta>)) LITBc))) =
       qfbranch (xlate_repl \<theta>) LITBc
        (\<lambda>M. qfbranch (xlate_repl \<theta>) LITBn
              (\<lambda>N. qfaux (xlate_qrepl trail_repl) (xlate_repl \<theta>) (exp.App (exp.App M N) (qfold (xlate_q trail) (xlate_repl \<theta>))) LITBc))"
  apply (simp add:qtfbranch_def qfbranch_def)
  apply (case_tac "btlookup \<theta> LITBc",simp_all add:btlookup_belookup_None btlookup_belookup_Some btdefault_bedefault)
  apply (case_tac "btlookup \<theta> LITBn",simp_all add:btlookup_belookup_None btlookup_belookup_Some btdefault_bedefault)
  done

lemma xlate_fold_otb[simp]:
  fixes \<theta>::trm_repl
  shows "(\<And>M l. xlate (qtfaux trail_repl \<theta> M l) = qfaux (xlate_qrepl trail_repl) (xlate_repl \<theta>) (xlate M) l) \<Longrightarrow>
       xlate (qtfbranch \<theta> LOTBn (\<lambda>M. qtfaux trail_repl \<theta> M LOTBc)) =
       qfbranch (xlate_repl \<theta>) LOTBn (\<lambda>M. qfaux (xlate_qrepl trail_repl) (xlate_repl \<theta>) M LOTBc)"
  apply (simp add:qtfbranch_def qfbranch_def)
  apply (case_tac "btlookup \<theta> LOTBn",simp_all add:btlookup_belookup_None btlookup_belookup_Some btdefault_bedefault)
  done

lemma xlate_fold[simp] :
  fixes q :: trail
  and \<theta> :: trm_repl
  and M :: trm
  shows "xlate (qtfold q \<theta>) = qfold (xlate_q q) (xlate_repl \<theta>)"
  and "\<And>M l. xlate (qtfaux \<zeta> \<theta> M l) = qfaux (xlate_qrepl \<zeta>) (xlate_repl \<theta>) (xlate M) l"
by (nominal_induct q and \<zeta> avoiding:\<theta> rule:trail_trail_repl.strong_inducts,simp_all add:xlate_ty_fresh xlate_q_fresh)

definition xlate_ctx_entry where
 "xlate_ctx_entry e = (fst e,xlate_ty (snd e))"

primrec xlate_ctx :: "CCau.ctx \<Rightarrow> Common.ctx" where
  "xlate_ctx [] = []"
| "xlate_ctx (e#\<Gamma>) = xlate_ctx_entry e#xlate_ctx \<Gamma>"

definition xlate_uctx_entry where
 "xlate_uctx_entry e = (fst e,xlate_ty (snd e))"

primrec xlate_uctx :: "CCau.uctx \<Rightarrow> Common.uctx" where
  "xlate_uctx [] = []"
| "xlate_uctx (e#\<Delta>) = xlate_uctx_entry e#xlate_uctx \<Delta>"

lemma hyp_xlate_ctx[simp] :
  fixes a :: var
  fixes u :: uvar
  fixes \<sigma> :: CCau.ty
  fixes \<Gamma> :: CCau.ctx
  fixes \<Delta> :: CCau.uctx
  shows "(a,\<sigma>) \<in> set \<Gamma> \<Longrightarrow> (a,xlate_ty \<sigma>) \<in> set (xlate_ctx \<Gamma>)"
  and "(u,\<sigma>) \<in> set \<Delta> \<Longrightarrow> (u,xlate_ty \<sigma>) \<in> set (xlate_uctx \<Delta>)"
apply (induct \<Gamma>,auto simp:xlate_ctx_entry_def)
apply (induct \<Delta>,auto simp:xlate_uctx_entry_def)
done

lemma xlate_ctx_fresh[simp] :
  fixes \<Gamma> \<Delta>
  fixes a::var
  fixes u::uvar
  shows "a \<sharp> \<Gamma> \<Longrightarrow> a \<sharp> xlate_ctx \<Gamma>"
  and "u \<sharp> \<Delta> \<Longrightarrow> u \<sharp> xlate_uctx \<Delta>"
apply (induct \<Gamma>,auto simp:xlate_ctx_entry_def fresh_list_cons fresh_list_nil fresh_prod xlate_ty_fresh)
apply (induct \<Delta>,auto simp:xlate_uctx_entry_def fresh_list_cons fresh_list_nil fresh_prod xlate_ty_fresh)
done

inductive_cases valid_cons[elim]:
"valid (e#\<Gamma>)"

lemma xlate_valid[simp] :
  fixes \<Gamma> \<Delta>
  fixes r::code
  fixes \<sigma>::ty
  shows "\<Delta>;\<Gamma> \<turnstile> r : \<sigma> \<Longrightarrow> valid (xlate_ctx \<Gamma>) \<and> uvalid (xlate_uctx \<Delta>)"
  and "\<Delta> \<turnstile> \<theta> : \<sigma> \<Longrightarrow> uvalid (xlate_uctx \<Delta>)"
  and "\<Delta> \<turnstile> \<Gamma> \<Longrightarrow> valid (xlate_ctx \<Gamma>) \<and> uvalid (xlate_uctx \<Delta>)"
  and "\<turnstile> \<Delta> \<Longrightarrow> uvalid (xlate_uctx \<Delta>)"
  and "\<Delta>;\<Gamma> \<turnstile> \<sigma> \<Longrightarrow> valid (xlate_ctx \<Gamma>) \<and> uvalid (xlate_uctx \<Delta>)"
apply (nominal_induct \<Delta> \<Gamma> r \<sigma> and \<Delta> \<theta> \<sigma> and \<Delta> \<Gamma> and \<Delta> and \<Delta> \<Gamma> \<sigma> 
       rule:wfc_wfc_repl_wf_ctx_wf_uctx_wf_ty.strong_inducts,auto simp:valid_cons)
apply (rule)
apply (simp add:xlate_ctx_entry_def,rule,simp)
apply (auto simp:xlate_ctx_fresh)
apply (rule)
apply (simp add:xlate_uctx_entry_def,rule,simp)
apply (auto simp:xlate_ctx_fresh)
done

lemma xlate_bt :
  shows "xlate_ty (CCau.bt \<tau> n) = Common.bt (xlate_ty \<tau>) n"
by (induct n,auto)
(*
xlate_ty (CCau.bt \<tau> (trail_label_arity l)) = Common.bt (xlate_ty \<tau>) (trail_label_arity l)
*)

lemma xlate_bty[simp] :
  shows "xlate_ty (CCau.bty \<tau> l) = Common.bty (xlate_ty \<tau>) l"
by (simp add:CCau.bty_def Common.bty_def xlate_bt)

(* Lemma 5 *)
theorem xlate_wf :
  fixes M :: trm
  fixes \<theta> :: trm_repl
  shows "\<Delta>;\<Gamma> \<turnstile> M : \<tau> ; M' \<Longrightarrow> xlate_uctx \<Delta>;xlate_ctx \<Gamma> \<turnstile> xlate M : xlate_ty \<tau>"
  and "\<Delta> \<turnstile> \<theta> : \<tau> ; \<theta>' \<Longrightarrow> xlate_uctx \<Delta> \<turnstile> xlate_repl \<theta> : xlate_ty \<tau>"
  apply (nominal_induct \<Delta> \<Gamma> M \<tau> M' and \<Delta> \<theta> \<tau> \<theta>' rule:CCau.wf_wf_repl.strong_inducts)
  apply (simp_all)
  apply (rule, auto simp add: xlate_ctx_entry_def xlate_uctx_entry_def xlate_ctx_fresh fresh_prod xlate_fresh)+
done

(* Lemma 7 *)
theorem xlate_red :
  fixes M N::trm
  shows "M \<leadsto> N \<Longrightarrow> xlate M \<leadsto> xlate N"
  apply (nominal_induct M N rule:CCau.red.strong_induct,simp_all)
  apply (rule, rule red_pe, rule, simp add:fresh_prod xlate_fresh xlate_ty_fresh)+
  apply (rule, rule red_pe, subgoal_tac "Insp (xlate_repl \<theta>) \<leadsto> xlate (qtfold q \<theta>)", simp add:qtfold_def)
  apply (simp add: xlate_fold qfold_def,rule)
  apply (rule, rule red_pe, subgoal_tac "Modal.IInsp (Modal.Box (xlate M)) (xlate_repl \<theta>) \<leadsto> xlate (qtfold q2 \<theta>)", simp add:qtfold_def)
  apply (simp add: xlate_fold qfold_def, rule)
  apply (rule, rule red_pe, auto)
done

(* strongly normalizing \<lambda>hs expressions *)
inductive tsn :: "trm \<Rightarrow> bool" 
where 
  tstep_sn : "(\<And> N. M \<leadsto> N \<Longrightarrow> tsn N) \<Longrightarrow> tsn M"
equivariance tsn
nominal_inductive tsn
done

definition TSN :: "trm set"
where "TSN \<equiv> {M . tsn M}"

lemma xlate_SN_aux:
  fixes s::exp
  assumes "s \<in> SN"
  shows "\<forall>M. xlate M = s \<longrightarrow> M \<in> TSN"
using assms
apply (simp add:TSN_def SN_def)
apply (induct rule:sn.induct)
apply (intro allI impI,rule)
apply (subgoal_tac "xlate Ma \<leadsto> xlate N",blast)
apply (auto simp:xlate_red)
done

lemma xlate_SN:
  fixes M::trm
  assumes "xlate M \<in> SN"
  shows "M \<in> TSN"
using assms
by (auto simp:xlate_SN_aux)

(* Corollary 15: \<lambda>hc is SN *)
theorem wf_is_TSN:
  fixes M N::trm
  assumes "\<Delta>;\<Gamma> \<turnstile> M : \<tau> ; M'"
  shows "M \<in> TSN"
using assms
apply (rule_tac xlate_SN)
apply (rule wf_is_SN)
apply (rule xlate_wf,simp)
done

end